# SCRIPT FOR SENDING PUSHOVER NOTIFICATIONS WITH  WITH A GIF FILE ATTACHMENT
#
# Version: 2.1
# Created by jaydeel at ipcamtalk.com
#
# Changelog:
# V2.1 - 2024-09-21  - BREAKING CHANGE: must replace old user-settings file
#                    - NEW: script now supports Telegram Bot notifications (alternative to Pushover); 
#                      added 3 new user settings to support this feature; 
#                      added 1 new column (TelegramMs') to the script's CSV log file;
#                      added push method ('pushover' or 'telegram') to BI logfile entry
#                    - NEW: added argument -PushService ('Pushover' or 'Telegram') to support on-demand service selection (case-insensitive)
#                    - NEW: added argument -Sound to support on-demand sounds (Pushover notifications only)
#                    - NEW: script can now make GIFs via BI Clipboard clips (or Alerts created via Clipboard clips)
#                    - NEW: added user setting '$append_performance_tag_flg' to hide default notification tag 'GifMakeTime: N msec'
#                    - NEW: added user setting '$experienced_user_flg' to hide output intended for first-time users
#                    - FIXED: script now handles DAT files with JPG images stored oldest to newest; effective BI 5.9.4.11+
#                    - FIXED: fixed bug where script did not try BVR extraction when DAT files contained fewer than required JPG images
#                    - FIXED: fixed bug preventing script from automatically using the BVR file when the DAT file not exist
#                      (thanks snz at ipcamtalk)
#                    - FIXED: fixed bug where script ignored argument -NoArchive
#                    - FIXED: fixed bug in BIJsonGetAlertList(); previously failed to exclude cloned camera entries
#                    - FIXED: fixed bug in BIJsonAlertListParse(); previously failed to exclude cloned camera entries
#                    - CHANGED: changed user setting '$append_DAT_tag_flg' to 'append_gif_method_tag_flg';
#                      used to hide default notification tags, like ' (viaDAT:n)' or ' (viaClip)' or ' (viaLive)'
#                    - CHANGED: reformatted notification tags for DAT-generated GIFs from '(DATn)' to '(viaDAT:n)'
#                    - CHANGED: made the following changes to the BI logfile entry to shorten the entry (127 characters max):
#                      replaced script basename (typically BI_gif_tool') to 'gif_tool'; saved 3 chars;
#                      changed duration ' sec' units '[320 H, 2 FPS, 6.0 sec, 724.5 KB]' to '[320 H, 2 FPS, 6.0 s, 724.5 KB]'; saved 2 chars;
#                      removed redundant ' ms' units '[0 ms, 8559 ms, 246 ms, 9021 ms]' to '[0, 8559, 246, 9021 ms]'; saved 9 chars;
#                      removed substring 'method=' (leaving just 'viaDat', 'viaClip', 'viaLive'); saved 7 characters
#                      removed entry 'rec='; saved at least 4 chars;
#                      removed substring 'source='; saved 7 chars;
#                      total characters saved = 3+2+9+7+4+7-8 for added push method substring = 24
#                    - IMPROVED: made function IsBiVersion() more efficient (gets BI version from registry instead of JSON call)
#                    - IMPROVED: -Live argument now supports the use of non-default folder naming patterns in 
#                      Camera settings > Record > Video
#                    - IMPROVED: updated functions: FindBytes() (V3), IsBiVersion() (V2)
# V2 - 2024-03-18    - BREAKING CHANGE: must replace old user-settings file 
#                    - BREAKING CHANGE: changed name of argument -DebRec to -DbRecNo; to clarify that script is 
#                      expecting a record ID number, not a string like "@xxxx"
#                    - BREAKING CHANGE: deprecated redundant argument -AlertDB (use -AlertRec instead)
#                    - NEW: script now supports multiple Blue Iris 'New' folders
#                    - NEW: added argument -Latest (use only w/ -Camera); makes GIF from the most recent alert
#                    - NEW: added argument -DatImageMsec to improve DAT-derived GIF duration estimates;
#                      should be the same as the value specified for setting 'Analyze one image each x msec'
#                      (BI > Camera settings >  Alert > AI Configuration)
#                    - NEW: argument -DevMode now exports each camera's JSON alertlist (to 'app_folder\_devmode')
#                    - NEW: added descriptive/explanatory '_README.txt' files to all programmatically 
#                      created folders
#                    - NEW: script's CSV log file automatically renamed for version changes
#                    - IMPROVED: changed method for making -Live GIFs (now using live stream instead of BVR)
#                    - IMPROVED: supports the use of non-default folder naming patterns in Camera settings > Record > Video
#                      e.g., %Y-%m-%d\&CAM\&CAM.%Y%m%d_%H%M%S
#                    - IMPROVED: args -TrimBegPct & -TrimEndPct now ALSO work on DAT-derived GIFs; this
#                      works by skipping the the extraction of an appropriate number of JPGs from the DAT file
#                    - IMPROVED: argument -TrimBegPct may be negative to include pre-trigger images
#                      (BVR-derived GIFs only)
#                    - IMPROVED: added 2 columns ('PushoverMs' & 'TotalMs') to the script's performance log
#                    - IMPROVED: updated -Help output to clarify usage options
#                    - IMPROVED: enhanced validation of the script's arguments
#  	                 - IMPROVED: script now provides feedback if commandline includes unsupported arguments 
#                    - CHANGED: deprecated argument -NukeDatFile; WHY? per Blue Iris Support a 
#                      DB/Regenerate action is always REQUIRED after deleting a DAT file. Obviously, 
#                      this must be avoided because it momentarily stops all recording DURING AN EVENT. 
#                      Instead, consider batch deleting no-longer-needed DAT files via my 'BI_DAT_tool.ps1' script
#                    - CHANGED: deprecated argument -NoArchiveDat; WHY? has no use after deprecating -NukeDatFile
#                    - CHANGED: deprecated user-file setting $dat_archive_lifetime_hours; WHY? has no use after deprecating -NukeDatFile
#                    - FIXED: fixed null source filespec for BVR when using argument -Live
#                    - FIXED: fixed untrapped error if Blue Iris service is not running
#                    - FIXED: fixed bug in function SendPushover()
#                    - IMPROVED: updated functions BIJsonCall(), BIJsonGetAlertList(), $FindBytes()
#                    - IMPROVED: function BIJsonAlertListParse() is now more tolerant of database issues
# V1.9  - 2023-09-07 - NEW: a repair is now attempted for situations where the JPG start- & end-of-image 
#                      byte markers appear to be unbalanced
#                    - IMPROVED: JPG validation function 'IsJpegFileValid()' now also tests for 
#                      valid end-of-image byte marker
#                    - CHANGED: the usage guide (shown via argument -Help) is no longer displayed by default
#                    - CHANGED: argument -NukeDatFile has been disabled until Blue Iris support provides 
#                      a safe method to delete DAT files
#                    - FIXED: fixed SendPushover() call; lacked grave accent (`) after -Camera param line; 
#                      caused malformed value
#                    - FIXED: fixed DAT-extracted JPG files that previously lacked the req'd JPEG 
#                      Interchange Format end-of-image byte marker ('FF D9')
# V1.8  - 2023-07-12 - IMPROVED: handling of warning if user appears to be trying to use 
#                      parameter -ImagesCnt with DAT files
#                    - FIXED: fixed argument -Priority; mis-typed as [String] instead of [Int]; 
#                      PS7 forgiving; PS5 not
# V1.7  - 2023-07-09 - NEW: added error trapping for non-existent database record IDs
#                    - IMPROVED: shortened message in BI logfile, and 'gif_source' entry in script's CSV log (removed path)
#                    - CHANGED: logged GIF duration values units from msec to seconds 
#                      (REQUIRES RENAMING OLD CSV FILE)
#                    - FIXED: fixed BI logfile message; missing source file
# V1.6  - 2023-06-30 - NEW: added warning if user appears to be trying to use parameter -ImagesCnt 
#                      with DAT files
#                    - IMPROVED: error trapping & feedback for DAT files with insufficient images
#                    - IMPROVED: error trapping when trying to get $gif_jpg_WxH from one of the JPG files
#                    - IMPROVED: ffmpeg issues (if any) are now displayed in script output to the console
#                    - FIXED: fixed handling of function ExtractDatImages() when its returned object is empty
# V1.5  - 2023-06-28 - IMPROVED: performance log (*.csv file) now reporting GIF's JPG WxH instead of  Alert's WxH
#                    - IMPROVED: function IsJpegFileValid() is now releasing resources ($img.Dispose()) 
#                      to reduce risk of memory leak
#                    - FIXED: fixed function CopyFileAsSequence()
# V1.4  - 2023-06-27 - NEW: added error trap when ffmpeg fails to successfully create a GIF file
#                    - IMPROVED: GIF size reduction algorithm now meets the Pushover API limit 
#                      (GIF can no longer have <1 image)
#                    - IMPROVED: Blue Iris logfile logging levels now reflect the script's output if it 
#                      contains 'ERR/' or 'WARNING:'
# V1.3  - 2023-06-27 - NEW: added argument -NoArchiveDat
#                    - FIXED: fixed issue that prevented the logging of the source file & alert record to 
#                      the CSV performance log file
# V1.2  - 2023-06-26 - NEW: added sleep status bar SleepWithStatusBar()
#                    - FIXED: fixed problem retrieving actively recording BVR filename (PowerShell v5)
# V1.1  - 2023-06-26 - CHANGED: modified default value for user setting '$gif_trim_end_cust_obj'
# V1    - 2023-06-26 - initial release
#
# REQUIRES:
#    CHOCOLATEY - Download and install the Chocolatey package manager: https://chocolatey.org/install 
#    FFMPEG     - Install the latest ffmpeg build using [choco install ffmpeg] ... https://www.gyan.dev/ffmpeg/builds/
#               - Home: https://www.ffmpeg.org/download.html          
#
# USAGE:      {this script name}.ps1 Msg [Priority AttachFile Quality Width, TargetKB Quiet]
# ARGUMENTS:  -AlertRec      = (string) alert record, typically via Blue Iris macro &ATTACH_DB
#             -AlertDB       = (string) [DEPRECATED] alert record, typ. via Blue Iris macro &ATTACH_DB
#             -AlertMemo     = (string) alert AI memo, typ. via Blue Iris macro &MEMO
#             -Live          = (switch) create GIF from specified camera's stream
#             -Camera        = (string) camera shortname (required with -Live argument)
#             -UI3Rec        = (string) hyphenated numeric values following UI3 URL argument '&rec=', 
#                                       like 'ui3.htm?rec=208953322-1181690'
#             -DbRecNo       = (string) Blue Iris database record number for clip, like '208953322' (not prefixed with '@')
#             -Seconds       = (integer) GIF duration; w/ -Live argument only; default is 10 seconds
#             -DatImageMsec  = (integer) camera-specific msec between DAT images (applies to DAT-derived GIFs only)
#             -BvrDelayMsec  = (integer) camera-specific msec to wait before extracting images (applies to BVR-derived GIFs only)
#             -Msg           = (string) notification message - may contain HTML
#             -Title         = (string) custom title for notification
#             -Priority      = (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
#             -Sound         = (string) custom sound for notification
#             -Latest        = (switch) uses DAT file from the most recent alert
#             -IgnoreDatFile = (switch) ignores 'DAT' file (if it exists); 
#                                       by default, a DAT file is always used if it contains N or more JPGs
#                                       N is specified in the user-settings file '$gif_DAT_reqd_images_cnt'; the default is 4
#             -NukeDatFile   = (switch) [DEPRECATED] removes DAT file from Blue Iris storage
#             -NoArchiveDat  = (switch) [DEPRECATED] disables automatic archiving of DAT files (if used to create GIF)
#             -ImagesCnt     = (integer) count of JPGs to use in the GIF animation
#             -Height        = (integer) GIF'S height dimension; units pixels
#             -FPS           = (integer) GIF frames per second
#             -TrimBegPct    = (integer) percentage of alert clip to trim from GIF animation (left side)
#             -TrimEndPct    = (integer) percentage of alert clip to trim from GIF animation (right side)
#             -PushService   = (string) 'Pushover' or 'Telegram'; overrides user settings file '$push_service'
#             -NoPush        = (switch) disables sending push service notification (enabled by default)
#                                       NOTE: also disables management of GIF size to under 2.5 MB
#             -NoBiLog       = (switch) disables logging to the blue Iris logfile (enabled by default)
#             -NoArchive     = (switch) disables automatic archiving of GIF files (enabled by default)
#             -Quiet         = (switch) disables ALL background messaging, and all Console messaging 
#                                       EXCEPT warnings and return object
#             -Demo          = (switch) disables sending notifications, and posting to the Blue logfile
#             -Help          = (switch) enable display of usage guide
#             -RtnHashtable  = (switch) returns Powershell hashtable instead of default JSON string (background only)
# RETURNS:    (hashtable) @{}
#
# EXAMPLES:
#    To send Pushover with GIF from Blue Iris 'On alert...' action
#       Parameters: "C:\ps_scripts\BI_gof_tool.ps1" -AlertRec `&ALERT_DB -Msg """<b>&CAM</b> &MEMO <a href='&WAN/alerts/&ALERT_DB?fulljpeg'>Hi-Res</a> <a href='&WAN/ui3.htm?rec=&ALERT_DB'>UI3</a>""" -Priority 2 -UseDatFile -ImagesCnt 10 -FPS 2 -Height 320 -TrimBegPct 0 -TrimEndPct 50
#

# MUST BE THE FIRST LINE IN A SCRIPT
param(
	[String]$AlertRec,
	#[String]$AlertDB,
	[String]$AlertMemo,
	[Switch]$Live,
	[String]$Camera,
	[Int]$Seconds,    # Default = 10
	[Int]$DatImageMsec,
	[Int]$BvrDelayMsec,
	[String]$UI3Rec,
	[String]$DbRecNo,
	[String]$Msg,
	[String]$Title,
	[Int]$Priority,
	[String]$Sound,
	[Switch]$Latest,
	[Switch]$IgnoreDatFile,
	[Switch]$NukeDatFile,  # Deprecated, forced to $false (argument kept for backward compatibility)
	[Int]$ImagesCnt,  # Default = 10
	[Int]$Height,     # Default = 320
	[Int]$FPS,        # Default = 2
	[Int]$TrimBegPct,
	[Int]$TrimEndPct,
	[String]$PushService,
	[Switch]$NoPush,
	[Switch]$NoBiLog,
	[Switch]$NoArchive,
	[Switch]$Quiet,
	[Switch]$Demo,
	[Switch]$Help,
	[Switch]$RtnHashtable,
	[Switch]$DevMode,
	[Switch]$Sim1,  # [OPTIONAL] remove ALL 'memo':'value' pairs from raw JSON 'alertlist' response
	[Switch]$Sim2,  # [OPTIONAL] remove 'memo':'value' pairs for 3 random alerts from JSON 'alertlist' response (per iteration)
	[Switch]$Sim3,  # [OPTIONAL] simulate BIJsonGetAlertList() response w/ all 'memo'='NA' params
	[Switch]$Sim4,  # [OPTIONAL] simulate BIJsonGetAlertList() response w/ 3 random 'memo'='NA' params (per iteration)
	[String]$Sim5   # [OPTIONAL] simulate BIJsonGetAlertList() response w/ zero date values; arg value is CamName
)

# NOTE: PowerShell recognizes parameters by the shortest unique match, meaning -title, -t, or -ti are the same parameter. 
# If you have 2 parameters like -title and -tag then the shortest unique match would be -ti and -ta respectively, 
# and using -t would throw an ambiguous parameter error.

# For debugging
# Write-Host ("PSBoundParameters = " + ($PSBoundParameters | ConvertTo-Json -Depth 100))
# Write-Host "AlertRec = $AlertRec"
# Write-Host "AlertMemo = $AlertMemo"
# Write-Host "Camera = $Camera"
# Write-Host "Live = $Live"
# Write-Host "DbRecNo = $DbRecNo"
# Write-Host "UI3Rec = $UI3Rec"
# Write-Host "Seconds = $Seconds"
# Write-Host "DatImageMsec = $DatImageMsec"
# Write-Host "Msg = $Msg"
# Write-Host "Title = $Title"
# Write-Host "Priority = $Priority"
# Write-Host "Sound = $Sound"
# Write-Host "IgnoreDatFile = $IgnoreDatFile"
# Write-Host "NukeDatFile = $NukeDatFile"
# if(!$PSBoundParameters.ContainsKey('ImagesCnt')) {
	# Write-Host "argument -ImagesCnt not used"
# }
# Write-Host "ImagesCnt = $ImagesCnt"
# Write-Host "Height = $Height"
# Write-Host "TrimBegPct = $TrimBegPct"
# Write-Host "TrimEndPct = $TrimEndPct"
# Write-Host "FPS = $FPS"
# Write-Host "Quiet = $Quiet"
# Write-Host "Demo = $Demo"

#Set-StrictMode  -Version 3
#Start-Transcript  # default destination is C:\Users\{user}l\Documents

# GATHER SCRIPT INFO 
$ScriptVersion = 'V2.1'
$ScriptName = $MyInvocation.MyCommand.Name
$ScriptBaseName = $ScriptName.Replace('.ps1', '')
$ScriptPath = Split-Path $MyInvocation.MyCommand.Path

# IMPORT USER SETTINGS...
# Note: The 'user settings' script is expected to be in same folder as this script
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	return "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
}

# ------------------------------------
# VALIDATE USER SETTINGS
if (1 -eq 1) {  # Use to collapse this code section while editing

	# ... Handle REQUIRED user settings
	$variable_test = Get-Variable -Name app_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$app_path"
		return $err_msg
	} else {
		if ($app_path.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$app_path is not set"
			return $err_msg
		}
	}
	$app_path = $app_path.TrimEnd('\')

	$variable_test = Get-Variable -Name bi_alerts_folder_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$bi_alerts_folder_path"
		return $err_msg
	} else {
		if ($bi_alerts_folder_path.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$bi_alerts_folder_path is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name bi_new_folder_paths -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$bi_new_folder_paths"
		return $err_msg
	} else {
		$bi_new_folder_paths_ary = $bi_new_folder_paths -Split ','
		if ($bi_new_folder_paths_ary[0].StartsWith('YOUR')) {
			$err_msg = "ERR/user-settings file setting `$bi_new_folder_paths is not set"
			return $err_msg
		}
	}
	$i = 0
	foreach ($item in $bi_new_folder_paths_ary) {
		$bi_new_folder_paths_ary[$i] = $item.Trim().TrimEnd('\')
		if (!(Test-Path -LiteralPath "$item")) {
			$err_msg = "ERR/user-setting `$bi_new_folder_paths contains a non-existent path: '$item'"
			return $err_msg
		}
		$i++
	}
	# For debugging
	#Write-Host ("bi_new_folder_paths_ary = " + ($bi_new_folder_paths_ary | ConvertTo-Json))
	if ($bi_new_folder_paths_ary.Count -gt 1) {
		if ($DevMode) {
			if ($FromConsole) {
				Write-Host "(dev mode) Blue Iris appears to be using MULTIPLE 'NEW' FOLDERS" -ForegroundColor Cyan
			}
		}
	}

	$variable_test = Get-Variable -Name bi_webserver -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/User-settings file is missing setting `$bi_webserver"
		return $err_msg
	} else {
		if ($bi_webserver.StartsWith('YOUR')) {  
			$err_msg = "ERR/user-setting file setting `$bi_webserver is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name bi_user_name -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/User-settings file is missing setting `$bi_user_name"
		return $err_msg
	} else {
		if ($bi_user_name.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$bi_user_name is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name bi_user_pw -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/User-settings file is missing setting `$bi_user_pw"
		return $err_msg
	} else {
		if ($bi_user_pw.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$bi_user_pw is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name push_service -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$push_service = "pushover"
	} else {
		if (!(@("pushover", "telegram") -contains $push_service.ToLower())) {  # -contains operator is case-insensitive by default
			$err_msg = "ERR/user-setting file setting `$push_service must be 'Pushover' or 'Telegram'"
			return $err_msg
		}
	}
	$push_service = $push_service.ToLower()
	
	if ($push_service -eq "pushover") {

		$variable_test = Get-Variable -Name po_app_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$po_app_token"
			return $err_msg
		} else {
			if ($po_app_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$po_app_token is not set"
				return $err_msg
			}
		}

		$variable_test = Get-Variable -Name po_user_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$po_user_token"
			return $err_msg
		} else {
			if ($po_user_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$po_user_token' is not set"
				return $err_msg
			}
		}

		$variable_test = Get-Variable -Name po_priority -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_priority = 0
		}

		$variable_test = Get-Variable -Name po_sound -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_sound = ""
		}

		$variable_test = Get-Variable -Name po_device -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_device = ""
		}

		$variable_test = Get-Variable -Name po_retry -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_retry = 30
		}

		$variable_test = Get-Variable -Name po_expire -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_expire = 300
		}

		$variable_test = Get-Variable -Name po_tags -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_tags = ""
		}

	}
	
	if ($push_service -eq "telegram") {

		$variable_test = Get-Variable -Name tg_chat_id -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$tg_chat_id"
			return $err_msg
		} else {
			if ($tg_chat_id.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$tg_chat_id is not set"
				return $err_msg
			}
		}

		$variable_test = Get-Variable -Name tg_api_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$tg_api_token"
			return $err_msg
		} else {
			if ($tg_api_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$tg_api_token' is not set"
				return $err_msg
			}
		}
	}

	# ... Handle OPTIONAL user settings
	
	$variable_test = Get-Variable -Name experienced_user_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$experienced_user_flg = $false
	}

	$variable_test = Get-Variable -Name bg_msgs_on -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$bg_msgs_on = $true
	}

	$variable_test = Get-Variable -Name gif_images_cnt -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_images_cnt = 10
	}
	#Write-Host "gif_trim_beg = $gif_trim_beg"

	$variable_test = Get-Variable -Name gif_trim_beg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_trim_beg = 0
	}
	#Write-Host "gif_trim_end = $gif_trim_end"

	$variable_test = Get-Variable -Name gif_trim_end -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_trim_end = 0
	}
	#Write-Host "gif_trim_end = $gif_trim_end"

	$variable_test = Get-Variable -Name gif_trim_end_cust_obj -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_trim_end_cust_obj = @{}
	}
	#Write-Host ("gif_trim_end_cust_obj = " + ($gif_trim_end_cust_obj | ConvertTo-Json))

	$variable_test = Get-Variable -Name gif_height -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_height = 320
	}

	$variable_test = Get-Variable -Name gif_fps -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_fps = 2
	}

	$variable_test = Get-Variable -Name gif_trim -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_trim = 0
	}

	$variable_test = Get-Variable -Name gif_archive_lifetime_hours -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_archive_lifetime_hours = 720
	}

	$variable_test = Get-Variable -Name gif_DAT_reqd_images_cnt -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_DAT_reqd_images_cnt = 4
	}

	$variable_test = Get-Variable -Name dat_image_interval_msec -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$dat_image_interval_msec = 500
	}

	$variable_test = Get-Variable -Name append_gif_method_tag_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$append_gif_method_tag_flg = $true
	}

	$variable_test = Get-Variable -Name append_performance_tag_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$append_performance_tag_flg = $true
	}

	$variable_test = Get-Variable -Name post_to_BI_logfile_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_flg = $false
	}

	$variable_test = Get-Variable -Name post_all_events -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_all_events = $false
	}
}
# ------------------------------------

# GATHER/VALIDATE SCRIPT INFO 
if (1 -eq 1) {  # Use to collapse this code section while editing
	#$ScriptName = $MyInvocation.MyCommand.Name
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')

	# For debugging
	#Write-Host "ScriptName = $ScriptName"
	#Write-Host "ScriptPath = $ScriptPath"
	#Write-Host "ScriptBaseName = $ScriptBaseName"
	#Write-Host "ScriptPath = $ScriptPath"

	# INITIALIZE VARS
	$PsVersion = $PSVersionTable.PSVersion.Major
	# NOTE: Scripts running from Windows Task Scheduler have a parent process named 'svchost'.
	#       https://stackoverflow.com/a/65894924
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	# NOTE: Scripts called by another script have a 'PSCommandPath'.
	#       Method 'PSCommandPath' returns null if script is executed via the terminal
	#       https://stackoverflow.com/a/23370618/13499984
	$CallingScript = $MyInvocation.PSCommandPath

	# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM BLUE IRIS)
	if ($ParentProcess -eq 'BlueIris') {
		$FromConsole = $false
	} else {
		$FromConsole = ($null -eq $CallingScript)
	}

	# Get the current username and domain (hostname)
	# NOTE: sometimes needed for msg.exe (e.g. from PowerShell ISE)
	# ... tip: from the console, you can type 'hostname' or 'whoami'
	$my_servername = (whoami).Split('\')[0]
	$server_username = (whoami).Split('\')[1]

	# For debugging
	#msg.exe $server_username /TIME:10 "entered script '$ScriptName'"
	#msg.exe $server_username /TIME:10 "PsVersion ($ScriptName) = $PsVersion"
	#msg.exe $server_username /TIME:10 "ParentProcess ($ScriptName) = $ParentProcess"
	#msg.exe $server_username /TIME:10 "FromConsole ($ScriptName) = $FromConsole"
	#msg.exe $server_username /TIME:5 "From script... $ScriptName`n-----`nPowerShell version is... $PsVersion`nParent process is... '$ParentProcess'`nCalling script name... $CallingScript"

	# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
	if ($bg_msgs_on -and !$FromConsole) {
		$ExecPolicy = Get-ExecutionPolicy
		$exec_policy_str = "PSExecutionPolicy = $ExecPolicy"
		if ($env:PSExecutionPolicyPreference) {
			$ExecPolicyPref = $env:PSExecutionPolicyPreference
			$exec_policy_str = "Session ExecutionPolicy = $ExecPolicyPref"
		}
		$msg_sec = 10
		$bg_msg = "Script '$ScriptName' is running in the background" + 
		"`n$exec_policy_str" + 
		"`nPsVersion = $PsVersion" + 
		"`nParentProcess = $ParentProcess" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"

		msg.exe $server_username /TIME:$msg_sec $bg_msg
	}

	# For testing - Simulate CLI mode
	#$FromConsole = $false

	# FEEDBACK UNEXPECTED COMMANDLINE ARGUMENTS
	#Write-Host ("params = " + ($PSBoundParameters | ConvertTo-Json))
	$cmdline_args_passed = [String[]]$PSBoundParameters.Keys
	#Write-Host "Extra Args w/ values = $Args"
	$cmdline_args_bad = ($Args | where{$_ -like "-*"}) | ForEach-Object{$_ -Replace '-', ''}
	if ($cmdline_args_bad) {
		$cmdline_args_bad = [String[]]$cmdline_args_bad
	} else {
		$cmdline_args_bad = @()
	}

	# For debugging
	#Write-Host "Cmdline Args Passed = $cmdline_args_passed"
	#Write-Host "Cmdline Args Bad = $cmdline_args_bad"

	if ($cmdline_args_bad.Count -gt 0) {
		if ($FromConsole) {
			Write-Host ("HEADS-UP: commandline contains unsupported argument(s): [ " + ($cmdline_args_bad -Join ', ') + " ]") -ForegroundColor Green
			return
		}
	}
}

# ===============================================

function PopupBackgroundMsg() {

	param(
		[String]$Message,   # [REQUIRED]
		[Int]$Seconds,      # [REQUIRED]
		[String]$UserName,  # [REQUIRED]
		[Boolean]$InBackground = $true
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	if ($InBackground) {
		msg.exe $UserName /TIME:$Seconds $Message
	}
}

function BIJsonCall() {
# PURPOSE:   To get a Blue Iris JSON interface response from the Blue Iris web server.
#            Intended to be called by other PowerShell scripts.
# ARGUMENTS: -JsonCmd        = (hashtable) like '@{cmd=xxx},@{cmd=xxx,camera=name},...'
#            -ForceFlg       = [OPTIONAL] (boolean) forces new login (gets a new authenticated session id)
#            -RetryNo        = [OPTIONAL] (int) if -gt 0, this call is a retry attempt
# RETURNS:   (object) PowerShell PSCustomObject
# GLOBALS:   Used global vars: $FromConsole, $Quiet 
#            $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
#            $script:Quiet       = (boolean) $true suppresses verbose output to the console
# REFS:      1) Blue Iris Help PDF: chapter 'Administration', section 'JSON Interface'
#            2) https://www.houselogix.com/docs/blue-iris/BlueIris/json.htm  (old!)
#            3) https://ipcamtalk.com/threads/json-interface.21491/
#            4) https://ipcamtalk.com/threads/json-help-please.40381/#post-372770
# NOTES:
# 1) Saves the BI API session id to a file for reuse (to avoid unnecessary repeated 
#    logins while the session is still active); the file stored in folder 
#    'c:\temp\$ScriptBaseName' (created if necessary)
# 2) In the event that the JSON call fails with reason 'Invalid session',
#    the script will automatically attempt to repeat the call up to $max_retries
#
# Version: 4 (8/29/2024) ) replaced *Flg args w/ global vars
# Created by jaydeel at ipcamtalk.com

	param (
		[Hashtable]$JsonCmd,
		[Boolean]$ForceFlg,
		[Int]$RetryNo
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered function: $fn_name() ..."	
	# Write-Host "`Entered function $fn_name()`n-----------------------------"
	# Write-Host ("$fn_name() JsonCmd = " + ($JsonCmd | ConvertTo-Json))
	# Write-Host "$fn_name() PsVersion = $PsVersion"	
	# Write-Host "$fn_name() FromConsole = $script:FromConsole"

	# For debugging
	#Write-Host "$fn_name() retry_cntr = $retry_cntr"	

	# System vars
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	$PsVersion = $PSVersionTable.PSVersion.Major
	# Get the current server domain (hostname) and username
	# Used for background messaging via msg.exe
	# ... tip: from the console, you can type 'hostname' or 'whoami'
	$server_name = (whoami).Split('\')[0]
	$server_username = (whoami).Split('\')[1]

	function encryptMD5($str) {
	# PURPOSE: Get MD5 hash for a string.
	# ARGS:    $str = (string) string to be encrypted
	# RETURNS: (string) MD5 encrypted string... like, '0753a0e20c9027ccab70f4291a21aa08'
	# REF:     https://infosecscout.com/get-md5-hash-in-powershell/
	# NOTES:   (from REF)
	#	Unfortunately, PowerShell provides no direct function to generate a hash from a string.
	#	However, PowerShell offers a cmdlet to generate MD5 hash for a file: Get-FileHash. 
	# 	The cmdlet can also be used to get the MD5 hash for a string, by opening a stream and hashing it.
	# UTILITY: https://www.md5hashgenerator.com/
	# ... Example: 'MD5string' returns '0753a0e20c9027ccab70f4291a21aa08'

		$stringAsStream = [System.IO.MemoryStream]::new()
		$writer = [System.IO.StreamWriter]::new($stringAsStream)
		$writer.write($str)
		$writer.Flush()
		$stringAsStream.Position = 0
		$hash_obj = Get-FileHash -InputStream $stringAsStream -Algorithm MD5
		
		# For debugging
		#$str = $hash_obj | Out-String
		#Write-Host "hash_obj = $str"

		return $hash_obj.Hash
	}
	
	# MAX RETRIES FOR JSON CMD
	# ... This function will recursively call itself if a JSON cmd's previous
	#     $json_response.result value is 'fail' and the reason is 'Invalid session'
	$max_retries = 3
	# ... Wait time before attempting a 'Invalid session' recursive call 
	$retry_delay_ms = 100
	
	if (!$RetryNo) {
		$retry_cntr = 0
	} else {
		$retry_cntr = $RetryNo
	}

	# Abort if we've exceeded the retries 'Invalid session' break variable
	if ($retry_cntr -ge $max_retries) {
		$err_msg ="ERR/JsonCmd '" + $JsonCmd.cmd + "' unable to create a valid session"
		return $err_msg
	}
		
	# Show progress bar as a heads-up visualization that retries are being attempted
	<# if ($retry_cntr -gt 0) {
		if ($script:FromConsole) {
			$WriteProgressParams = @{
				Activity         = "Retrying BI JSON Call..."
				Status           = "Progress: retry $retry_cntr of $max_retries max"
				PercentComplete  = (($retry_cntr / $max_retries) * 100)
			}
			Write-Progress @WriteProgressParams
			# Note: If loop is too fast, you might need a delay to get the progress bar to show
			Start-Sleep -Milliseconds 2000				
		}
	} #>

	# HANDLE FILES FOR SAVED SESSION IDS AND RETRY COUNTER
	# ... Perform existence test for the *.txt file used to save session ids
	#     Create its path if it does not exist
	#     NOTE: if a path is not specified, the default Windows location is 'C:\Users\<username>"
	$path = "c:\temp\$ScriptBaseName"
	if (!(Test-Path -LiteralPath $path)) {
		New-Item -ItemType Directory -Force -Path $path
	}
	# ... NOTE: we use a different filename when running the script via different processes;
	#     this avoids collisions with scripts running via:
	#     'scvhost' (Task Scheduler), 'WindowsTerminal', 'explorer' (Powershell ISE)
	$base_name = "$path\bi_session_id"
	$session_filespec = $base_name + "_" + $ParentProcess + ".txt"
	$found_saved_session_file_flg = (Test-Path -LiteralPath "$session_filespec")

	# PARSE COMMANDLINE ARGS
	if ($JsonCmd -eq $null) {
		$err_msg = "ERR/expecting an argument: JsonCmd (PowerShell hashtable)"
		return $err_msg
	} else {
		# For debugging
		#Write-Host ("requested JsonCmd type = " + ($JsonCmd.GetType()))
	}
	# ... Save original JsonCmd (in case we need to make a recursive call)
	$json_cmd_orig = $JsonCmd
	# ... Override reuse of saved session_id
	if ($ForceFlg) {
		$found_saved_session_file_flg = $false
	}

	# ========================
	# PERFORM API CALLS

	# Make the URL
	$uri = "http://" + $script:bi_webserver + "/json";

	# For debugging
	#Write-Host "$fn_name() uri = $uri"

	# Flag 'logout' JSON cmd requests
	# ... NOTE: All JSON 'logout' cmds should be performed by the calling script!
	if ($JsonCmd.cmd -eq 'logout') {
		$logging_out_flg = $true
	} else {
		$logging_out_flg = $false
	}
	
	# For debugging
	if ($logging_out_flg) {
		#Write-Host "$fn_name() detected JSON cmd to log out of BI API session"
	} else {
		#Write-Host "$fn_name() found_saved_session_file_flg... $found_saved_session_file_flg"
		#msg.exe * /TIME:5 "$fn_name() found_saved_session_file_flg... $found_saved_session_file_flg"
	}

	# SPECIAL HANDLING FOR 'geofence' JSON COMMAND REQUESTS
	$dev_uuid = $null  # Default
	$json_cmd_keys = $JsonCmd.Keys
	if ($json_cmd_keys -contains 'uuid') {
		if ($JsonCmd.uuid -ne $null) {
			$dev_uuid = $JsonCmd.uuid
		}
	}

	# For debugging
	if ($dev_uuid) {
		#Write-Host "$fn_name() dev_uuid = $dev_uuid"
	}

	# EXECUTE API CALL
	
	if (!$found_saved_session_file_flg) {
		
		# ================================================================
		# 1.1 PRELIMINARY LOGIN
		
		# Create, then encode the JSON 'login' command
		$parameters = @{
			cmd = 'login'
		}
		$json_params = $parameters | ConvertTo-Json

		# For debugging
		#Write-host "$fn_name() (1.1) login #1 json_params... `n$json_params"
		#msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 json_params... `n$json_params"

		# Execute the JSON request
		$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 jsonResponse... `n$jsonResponseObj"
		
		# Get the session id
		$session = $jsonResponseObj.session
		
		# ================================================================
		# 1.2 AUTHENTICATE THE LOGIN

		# Create MD5 hash-encoded credentials string
		$response_str = $bi_user_name + ":" + $session + ":" + $bi_user_pw
		$response_hash = encryptMD5 $response_str

		# Create, then encode a modified JSON 'login' command
		if ($dev_uuid) {
			$parameters = @{
				cmd = 'login'
				session = $session
				response = $response_hash
				uuid = $dev_uuid
			}
		} else {
			$parameters = @{
				cmd = 'login'
				session = $session
				response = $response_hash
			}
		}
		$json_params = $parameters | ConvertTo-Json

		# For debugging
		#Write-Host "$fn_name() (1.2) authorizing json login... `n$json_params"
		#msg.exe $server_username /TIME:15 "$fn_name() (1.2) authorizing json login ... `n$json_params"
		
		# Execute the JSON request
		$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#Write-Host ("$fn_name() (1.2) jsonResponse... `n" + ($jsonResponseObj  | ConvertTo-Json))
		#msg.exe $server_username /TIME:15 "$fn_name() (1.2) jsonResponse... `n$jsonResponseObj"
		
		# Handle failure
		if ($jsonResponseObj.result -eq 'fail') {
			$err_msg = "ERR/unable to authenticate Blue Iris JSON interface`n"
			return $err_msg
		}
		
		# Get the authenticated session id
		$auth_session_id = $jsonResponseObj.session

		# For debugging
		#msg.exe $server_username /TIME:5 "$fn_name() auth_session_id... $auth_session_id"
	
		# Save the authenticated session id
		# NOTE: the session id is always saved, whether used or not
		$auth_session_id | Set-Content "$session_filespec" -NoNewline

	} else {
	
		# Retrieve the saved session id
		# KEY POINT: 
		#   Force typing the output of 'Get-Content' to [String] was essential when running script in
		#   in Powershell ISE. Otherwise the output was similar to an object.
		#   *** But in actuality it is more complicated ***
		#   Per https://stackoverflow.com/a/55881978/13499984
		#      When 'Get-Content' is assigned to a variable, its output is an array of .NET [String] objects, 
		#      each representing a line from the input file. However, Get-Content also decorates these string 
		#      objects with additional properties that provide helpful metadata about where the strings came from.
		$saved_session_id = [String](Get-Content -Path "$session_filespec")

		$auth_session_id = $saved_session_id

		# For debugging
		if (!$logging_out_flg) {
			#Write-Host ("$fn_name() saved_session_id type = " + $saved_session_id.GetType())
			#Write-Host "$fn_name() saved_session_id... $saved_session_id"
			#msg.exe * /TIME:5 "$fn_name() saved_session_id... $saved_session_id"
		}
		
	}

	# ================================================================
	# 1.3 EXECUTE THE REQUESTED JSON COMMAND

	if ($dev_uuid -eq $null) {
		
		# In the event that this JSON call fails with reason 'Invalid session',
		# the script will automatically attempt up to $max_retries
	
		# Modify, then encode the requested JSON login command
		$JsonCmd.session = $auth_session_id
		$json_params = $JsonCmd | ConvertTo-Json

		# Execute the JSON request
		$params = @{
			Uri                = $uri
			#Headers            = @{"Content-Type"  = "application/json"}
			Method             = 'POST'
			#Body               = $JsonCmd | ConvertTo-Json
			Body               = $json_params
			ContentType        = 'application/json'
		}
		
		$json_response_obj = Invoke-RestMethod @params

		# Alternate syntaxes (pick one)
		#$json_response_obj = Invoke-RestMethod -Uri $uri -Method Post -Body ($JsonCmd | ConvertTo-Json) #>
		#$json_response_obj = ($JsonCmd | ConvertTo-Json) | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#if ($JsonCmd.cmd -match 'update') {
			#Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "'")
			#Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "' ...`n" + $json_params)
			#Write-Host ("$fn_name() json_response_obj = `n" + ($json_response_obj | ConvertTo-Json))
		#}

		if ($json_response_obj.result -eq 'fail') {
			
			# For debugging
			#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response.result = fail")
			#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response = `n" + ($json_response_obj | ConvertTo-Json))

			# Try to determine reason
			if ($json_response_obj.data) {

				$data_obj = $json_response_obj.data

				# For debugging
				#Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data = " + ($data_obj | ConvertTo-Json))
				
				if ($data_obj.reason) {
					$reason = $data_obj.reason
				
					# For debugging
					#Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data.reason = $reason")

					if ($reason -match 'Invalid session') {
						# Update retry counter
						$retry_cntr++
						
						# RECURSIVE CALL
						#Start-Sleep -Milliseconds 500
						$json_response_obj = `
							BIJsonCall `
							-JsonCmd $JsonCmd `
							-ForceFlg $true `
							-RetryNo $retry_cntr
						
					} elseif ($reason -match 'Access denied') {
						$err_msg ="ERR/JsonCmd '" + $JsonCmd.cmd + "' requires an Administrator account"
						return $err_msg
						
					} else {
						$err_msg = "ERR/JsonCmd '" + $JsonCmd.cmd + "' failed with reason '" + $reason + "'"
						return $err_msg
					}
					
				} else {
					$err_msg = "ERR/JsonCmd '" + $JsonCmd.cmd + "' failed without a reason"
					return $err_msg
				}
			}
		} 
	}

	# For debugging
	if ($json_response_obj) {
		#Write-Host ("$fn_name() (1.3) jsonResponse type = " + $json_response_obj.GetType())
		#Write-Host ("$fn_name() (1.3) jsonResponse... `n" + ($json_response_obj  | ConvertTo-Json))
		#msg.exe $server_username /TIME:5 "$fn_name() (1.3) jsonResponse... $json_response_obj"
	} 

	# Retain for error testing
	#$json_response_obj = $null
	#$json_response_obj = ''

	if ($json_response_obj -eq $null) {
		$err_msg ="ERR/null response from Blue Iris JSON interface"
		return $err_msg
	} 
	if ($json_response_obj -eq '') {
		$err_msg = "ERR/empty response from Blue Iris JSON interface"
		return $err_msg
	}
	if ($json_response_obj.result -ne $null) {

		# For debugging
		#Write-Host ("$fn_name() (1.3) json_response_obj.result... " + ($json_response_obj.result))
		
		if ($json_response_obj.result -eq 'fail') {
			
			# Show the contents of the $json_response_obj 'data' key

			$json_data_obj = $json_response_obj.data
			
			# For debugging
			#Write-Host ("$fn_name() (1.3) json_response_obj = " + ($json_response_obj | ConvertTo-Json))
			#Write-Host ("$fn_name() (1.3) json_response_obj.data = " + ($json_response_obj.data))

			# 'Minify' the $json_data_obj.
			if ($PsVersion -ge 7) {
				# Avoids the following WARNING...
				# "Resulting JSON is truncated as serialization has exceeded the set depth of N"
				$jdo_str = $json_data_obj | ConvertTo-Json -Depth 16
			} else {
				$jdo_str = $json_data_obj | ConvertTo-Json
			}
				
			# For debugging
			#Write-Host ("jdo_str = '" + $jdo_str + "'")
			
			if ($jdo_str) {				
				if ($jdo_str -eq 'null') {
					$err_msg = "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is null (DB record not found)"
					return $err_msg
				} else {
					$err_msg = "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is 'result:fail'"
					return $err_msg
				}
			}
		}
	}

	return $json_response_obj

}

function BIJsonClipstats() {
# PURPOSE:  To get blue iris json 'clipstats' cmd properties
# PARAMS:   $Clips          = [REQUIRED] (string) comma-delimited QUOTED DB locators, like "@926894649","@926871636"
#                              NOTE: extensions accepted, but not required, e.g., "@926894649.bvr",
#           Props           = [OPTIONAL] (string) comma-delimited JSON properties, like 'name,push,count'
# RETURNS:  1. (ordered hashtable) if multiple clips and/or properties requested
#           2. (string) if a single clip and a single property requested
#           3  'NA' if the clip is specified and the property does not exists
# GLOBALS:  Used global vars: $FromConsole, $Quiet, $DevMode, $DevModePath 
#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
#           $script:Quiet       = (boolean) $true suppresses verbose output to the console
# USAGE:    BIJsonClipstats -Clips list -Props list -Quiet
# USAGE EXAMPLE
#		$response_obj = BIJsonClipstats -Clips -Props 'camera,path,file,msec,alertmsec,res'
#
#       From the response, set these variables:
#		$clipstats_result = $response_obj[0] 
#                         like, {"result": "success"}, or {"result": "fail"}
#		$clipstats_data = $response_obj[1]
#                         like {"@152866668.bvr": "NA"}
#		                  or   {"@152866668.bvr": {
#			                       "camera": "DW1",
#			                       "path": "@152720932.bvr",
#			                       "file": "DW1.20230530_060000.1965063.17-0.jpg",
#			                       "msec": 3599166,
#			                       "alertmsec": 27426,
#			                       "res": "2688x1520"
#			                       "dummy": "NA"
#		                        }
#
# Version: 3 (9/4/2024) replaced *Flg args w/ global vars; now avoids error if argument -Props contains duplicate items
# Created by jaydeel at ipcamtalk.com

	# MUST BE THE FIRST LINE IN A SCRIPT
	param(
		[String]$Clips,
		[String]$Props
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "$fn_name() Clips = '$Clips'"
	#Write-Host "$fn_name() Props = '$Props'"
	
	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$Clips) {
		$err_msg = "ERR/$fn_name() expecting an argument (-Clips)"
		return $err_msg
	}
	$my_clips_ary = $Clips.split(',')
	
	# HANDLE OPTIONAL: COMMANDLINE ARGS
	$my_props_ary = @()
	if ($Props) {
		# Ensure that requested properties are unique
		$my_props_ary = $Props.split(',') | Select-Object -Unique
	}

	# For debugging
	#Write-Host ("$fn_name() my_clips_ary = " + $my_clips_ary)
	#Write-Host ("$fn_name() my_clips_ary cnt = " + $my_clips_ary.length)
	#Write-Host ("$fn_name() my_props_ary = " + $my_props_ary)

	# INITIALIZE
	$rtn_obj = [ordered]@{}
	<# $rtn_obj = [ordered]@{
		success = $null
		data = $null
	} #>
	$x = ""

	if ($my_clips_ary.length -gt 0) {
		$i = 1;
		foreach ($clip in $my_clips_ary) {
			
			if ($script:FromConsole -and !$script:Quiet) {
				if ($i -eq 1) {
					#Write-Host "$fn_name() processing clip '$clip'"
				} else {
					#Write-Host "`n$fn_name() processing clip '$clip'"
				}
			}

			# GET JSON DATA FROM THE BLUE IRIS WEB SERVER
			# ... REQUIRES COMPANION SCRIPT 'BI_json_call.ps1'
			#     WHICH RETURNS A POWERSHELL PSCustomObject
			$jsonCmd = @{
				cmd = 'clipstats'
				path = $clip
			}
			
			#$json_response_obj = BIJsonCall -JsonCmd $jsonCmd -QuietFlg $script:Quiet
			$json_response_obj = `
				BIJsonCall `
				-JsonCmd $jsonCmd

			# For debugging
			#Write-Host ("$fn_name() json_response_obj type = " + $json_response_obj.GetType())
			#Write-Host ("$fn_name() json_response_obj** = " + ($json_response_obj | ConvertTo-Json))
			#Write-Host ("json_response_obj.result** = " + ($json_response_obj.result))

			# If GET failed, process next clip
			if ($json_response_obj -match 'ERR/') {
				if ($script:FromConsole -and !$script:Quiet) {
					Write-Host $json_response_obj
				}
				$rtn_obj[$clip] = 'NA'
				break
			}
			if ($json_response_obj.result -eq "fail") {

				# For debugging
				#Write-Host ("$fn_name() json_response_obj.result type = " + (($json_response_obj.result).GetType()))
				#Write-Host ("$fn_name() son_response_obj.result = " + ($json_response_obj.result | ConvertTo-Json))

				if ($script:FromConsole -and !$script:Quiet) {
					Write-Host "$fn_name() json_response_obj.result = fail"
					Write-Host ($json_response_obj | ConvertTo-Json)
				}
				$rtn_obj[$clip] = 'NA'
				break
			}

			# Extract data from the JSON command response
			if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {

				$data_obj = $json_response_obj.data

				# For debugging
				#Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
				#Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))

				# $data_obj is a PSCustomObject like
				# @{
				#	camera=DW1;
				#	path=@746553655.bvr;
				#	file=DW1.20230428_220000.1572042.17-0.jpg; --OR-- 2023-04-28\\DW2\\DW1.20230428_220000.1572042.17-0.jpg
				#	offset=1572009;
				#	date=1682744400;
				#	color=16711808;
				#	flags=134217985;
				#	zones=0;
				#	memo=;
				#	plate=;
				#	res=2688x1520;
				#	msec=2814017;
				#	alertmsec=24520:
				#	camprofile=7;
				#	camschedule=1;
				#	camsmart=2;
				#	camzones=1;
				#	campreset=30;
				#	filesize=46m55s (419.5M);
				#	filetype=bvr H264 New
				# }

				# Convert PSCustomObject into a hashtable
				# REF: https://tinyurl.com/ddatr635
				$hashtable = @{}
				foreach ($property in $data_obj.psobject.properties.name) {
					$hashtable[$property] = $data_obj.$property
				}
				$data_obj_keys = $hashtable.Keys

				# For debugging
				# Write-Host ("$fn_name() hashtable = " + ($hashtable | ConvertTo-Json))
				# Write-Host "$fn_name() data_obj_keys = $data_obj_keys"

				$new_obj = [ordered]@{}		
				if ($my_props_ary.length -gt 0) {
					
					# LIST REQUESTED PROPERTIES
					foreach ($clipProp in $my_props_ary) {
						if ($data_obj.$clipProp -or $data_obj.$clipProp -eq 0) {
							$new_obj.Add($clipProp, $data_obj.$clipProp)
							$x = $data_obj.$clipProp

							# For debugging
							#Write-Host ($clipProp + " = "  + $data_obj.$clipProp)

						} else {
							$new_obj.Add($clipProp, "NA")
							$x = "NA"

							# For debugging
							#Write-Host ($clipProp + " = NA")
						}
					}
					$rtn_obj[$clip] = $new_obj
					
				} else {
					
					# LIST ALL PROPERTIES
					foreach ($clip_prop in $data_obj_keys) {
						if ($data_obj.$clip_prop -or $data_obj.$clip_prop -eq 0) {
							if ($script:FromConsole -and !$script:Quiet) {
								#Write-Host ($clip_prop + " = "  + $data_obj.$clip_prop)
							}							
							$new_obj.Add($clip_prop, $data_obj.$clip_prop)
						} else {
							if ($script:FromConsole -and !$script:Quiet) {
								#Write-Host ($clip_prop + " = NA")
							}							
							$new_obj.Add($clip_prop, "NA")
						}
					}
					$rtn_obj[$clip] = $new_obj
				}
				$i++
			}
		}

		# LOG OUT OF API SESSION
		$jsonCmd = @{cmd = 'logout'}
		BIJsonCall `
		-JsonCmd $jsonCmd
			
	} else {
		<# $data_obj_keys = @()
		foreach ($item in $data_obj) {
			$data_obj_keys += $($item | Get-Member -MemberType *Property).Name
		}
		foreach ($key in $data_obj_keys) {
			if ($script:FromConsole -and !$script:Quiet) {
				#Write-Host ($key + " = "  + $data_obj.$key)					
			}
			$rtn_obj.Add($key, $data_obj.$key)
		} #>
	}

	# Handle single clip, single property request
	if ($my_clips_ary.length -eq 1 -and $my_props_ary.length -eq 1) {
		$rtn_obj = $x
	}
								
	<# if ($script:FromConsole -and !$script:Quiet) {
		Write-Host "Clipstats:"
		Write-Host "----------"
		Write-Host ($rtn_obj | ConvertTo-Json)
	} #>

	return $rtn_obj
}

function BIJsonAlertListParse() {
# PURPOSE:  To get and parse the Blue Iris JSON interface cmd 'alertlist' response for a SINGLE CAMERA
#           To retrieve all or, if requested, a subset of 'alertlist' properties for each Alert
#           NOTE: returns non-existent requested 'alertlist' properties as 'alert_prop':'NA'
# PARAMS:   -CamName         = [REQUIRED] camera shortname
#           -Pattern         = process Alerts with descriptor substring; must be used w/ -CamName
#           -AlertProps      = (array) like 'memo,offset', comma delimited
#             			        options: camera, clip, color, date, file, filesize, flags memo, 
#                                     newalerts, newalerttime, offset, path, res, zones
#           -MemoTag         = (string) process only those protected Alerts with descriptor substring
#           -DatabaseView    = (string) default is 'alerts'
#                              options: all, new, stored, alerts, aux1, aux2, aux3, aux4, aux5, aux6, aux7,
#                                       flagged, export, archive, people, vehicles, confirmed, or canceled                             
#                              Note: When view='flagged', clip items are included too, but they don't have msec metadata.
#           -AlertPath        = (string) process single Alert with alertlist [path] like '@123456789'
#           -ExcludeClonesFlg = (boolean) exclude clones of cam names in -Cameras
#           -SilentFlg        = (boolean) true suppresses all output to the console, including warnings
# RETURNS:  (object) PowerShell hashtable, @('cam':camName,'alert_prop':value}
# REQUIRES: Functions: BIJsonCall(), IsBiVersion(), GetBiVersion()
# GLOBALS:  Uses global vars: $BiVersion, $FromConsole, $Quiet, $DevMode, $DevModePath 
#           $script:BiVersion   = (string) Blue Iris software current version
#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
#           $script:Quiet       = (boolean) $true suppresses verbose output to the console
#           $script:DevMode     = (boolean) $true 
#           $script:DevModePath = (string) 
#
# Created by jaydeel at ipcamtalk.com
# Version: 5 (8/29/2024) improved BI version checking; replaced *Flg args w/ global vars
# Version: 6 (9/9/2024) fixed bug where alertlist response unexpectedly contains cloned camera entries

	param(
		[String]$CamName,           # [REQUIRED] camera shortname
		[String]$Pattern,           # must be used with -CamName
		[String[]]$AlertProps,      # list of BI JSON 'alertlist' cmd properties (parameters)
		[String]$MemoTag,           # substring in Alert's memo
		[String]$DatabaseView,      # default is 'alerts'
		[String]$AlertPath,         # like '@123456789' (JSON alertlist [path] value)
		[Boolean]$ExcludeClonesFlg, # $true excludes $CamName's cloned camera entries 
		[Boolean]$SilentFlg         # suppresses all output to the console, including warnings
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered $fn_name() ..."
	# Write-Host "$fn_name() CamName = $CamName"
	# Write-Host "$fn_name() Pattern = $Pattern"
	# Write-Host ("$fn_name() AlertProps type = " + $AlertProps.GetType())
	# Write-Host ("$fn_name() AlertProps = " + ($AlertProps | ConvertTo-Json))
	# Write-Host "$fn_name() MemoTag = $MemoTag"
	# Write-Host "$fn_name() DatabaseView = $DatabaseView"
	# Write-Host "$fn_name() AlertPath = $AlertPath"
	# Write-Host "$fn_name() ExcludeClonesFlg = $ExcludeClonesFlg"
	# Write-Host "$fn_name() SilentFlg = $SilentFlg"
	# Write-Host ("$fn_name() global FromConsole = " + $script:FromConsole)
	# Write-Host ("$fn_name() global Quiet = " + $script:Quiet)"
	# Write-Host ("$fn_name() global DevMode = " + $script:DevMode)
	# Write-Host ("$fn_name() global DevModePath = " + $script:DevModePath)

	#Set-StrictMode  -Version 3
	# PARSE ARGS
	if ($Pattern -and !$CamName) {
		$err_msg = "ERR/argument -Pattern must be used with -CamName [$fn_name]"
		return $err_msg
	}
	if (!$CamName) {
		$err_msg = "ERR/expecting at least 1 argument (-CamName) [$fn_name]"
		return $err_msg
	}

	# If searching only a single AlertMemo, we can avoid GETTING all alerts 
	# via the JSON 'alertlist' cmd parameter 'search'
	$search_param_val = $null
	if ($MemoTag) {
		$search_param_val = $MemoTag
	}
	#Write-Host "$fn_name() search_param_val = $search_param_val"
	
	if (!$DatabaseView) {
		$DatabaseView = 'all'
	}
	if (!$AlertPath) {
		$AlertPath = "*"
	}

	# Initialize master collector for all alerts in database 
	$al_props_obj = @{}
	$al_props_obj.Add( 'cameras', @() )
	$al_props_obj.Add( 'paths', @() )
	$al_props_obj.Add( 'clips', @() )
	$al_props_obj.Add( 'files', @() )
	$al_props_obj.Add( 'memos', @() )
	$al_props_obj.Add( 'offsets', @() )
	$al_props_obj.Add( 'flagss', @() )  # double 's' intentional!
	$al_props_obj.Add( 'ress', @() )    # double 's' intentional!
	$al_props_obj.Add( 'zoness', @() )  # double 's' intentional!
	$al_props_obj.Add( 'dates', @() )
	$al_props_obj.Add( 'colors', @() )
	$al_props_obj.Add( 'filesizes', @() )
	
	# For debugging
	#Write-Host ("$fn_name() al_props_obj type = " + $al_props_obj.GetType())
	#Write-Host ("$fn_name() al_props = " + ($al_props_obj | ConvertTo-Json))

	# ---------------------------------------------------------------------------------------
	# KEY POINT: Prior to version 5.7.9.3 BI's 'alertlist' json cmd returned only 1000 items; 
	#            As of version 5.7.9.3 BI's 'alertlist' json cmd returns 2048 items; 
	#            so we must re-send the json cmd until we have retrieved all alerts.
	#            REF: https://ipcamtalk.com/threads/blue-iris-ui3.23528/post-788744
	
	$alertlist_2048_flg = `
		IsBiVersion `
		-CompareToVersion '5.7.9.3' `
		-CurrentVersion $script:BiVersion

	if ($alertlist_2048_flg) {
		$json_alertlist_max = 2048;
		#$json_cliplist_max = 4096;
	} else {
		$json_alertlist_max = 1000;  # Default (tested against a BI 5.6.3.7 server by bp2008@ipcamtalk)
		#$json_cliplist_max = 1200;   # Default (tested against a BI 5.6.3.7 server by bp2008@ipcamtalk)
	}
	# ---------------------------------------------------------------------------------------

	# Initialize WHILE LOOP variables
	[Array]$paths_ary = @('x')	# primary break loop variable... starts with a dummy value
	$oldest_date = 0  # oldest date found 
	$newest_date = 0  # newest date found 
	$skipped_next_json_call_flg = $false
	
	# PREP FOR SPECIAL HANDLING FOR THE 'MEMO' PARAMETER
	# WHY? Per Blue Iris support...
	#     Prior to version 5.8.2.4, the alertlist 'memo' field was included only if it was not blank.
	#     Starting with version 5.8.3.4 it is now always included (REF: 12/18/2023 email from Ken at blueirissupport.com)
	$alertlist_memo_always_flg = `
		IsBiVersion `
		-CompareToVersion '5.8.2.4' `
		-CurrentVersion $BiVersion
	
	# Initialize WHILE LOOP break variables
	$prev_loop_alerts_cnt = 0
	$max_tries = 10
	$iter = 1;

	# WHILE LOOP - Get JSON 'alertlist' data in batches until the cmd returns no new data
	while ($paths_ary.Count -gt 0) {
		
		if ($script:DevMode -and $script:FromConsole) {
			if ($iter -eq 1) {
				Write-Host ""
				Write-Host "(dev mode) ----- $CamName" -ForegroundColor Cyan
			} else {
				Write-Host "(dev mode) -----" -ForegroundColor Cyan
			}
			Write-Host "(dev mode) JSON alertlist iter #$iter" -ForegroundColor Cyan
		}

		# Initialize collector for this JSON 'alertlist' call
		$tmp_obj = @{}
		$tmp_obj.Add( 'cameras', @() )
		$tmp_obj.Add( 'paths', @() )
		$tmp_obj.Add( 'clips', @() )
		$tmp_obj.Add( 'files', @() )
		$tmp_obj.Add( 'memos', @() )
		$tmp_obj.Add( 'offsets', @() )
		$tmp_obj.Add( 'flagss', @() )  # double 's' intentional!
		$tmp_obj.Add( 'ress', @() )    # double 's' intentional!
		$tmp_obj.Add( 'zoness', @() )  # double 's' intentional!
		$tmp_obj.Add( 'dates', @() )
		$tmp_obj.Add( 'colors', @() )
		$tmp_obj.Add( 'filesizes', @() )
		#$tmp_obj.Add( 'AAAA', @() )
		
		# For debugging
		#Write-Host ("$fn_name() tmp_obj type = " + $tmp_obj.GetType())
		#Write-Host ("$fn_name() tmp_obj = " + ($tmp_obj | ConvertTo-Json))
		#Write-Host ("$fn_name() tmp_obj.paths type  = " + ($tmp_obj.paths).GetType())

		# GET JSON DATA FROM THE BLUE IRIS WEB SERVER

		# First call params - Default
		$jsonCmd = @{
			cmd = 'alertlist'
			camera = $CamName
			search = $search_param_val
			view = $DatabaseView
		}
		
		# Subsequent call params
		if ($iter -gt 1) {
			
			# For debugging
			#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
			#Write-Host "$fn_name() oldest_date = $oldest_date"
			
			if ($oldest_date -gt 0) {  #Just in case
				# Params for the next set of 'alertlist' data
				$jsonCmd = @{
					cmd = 'alertlist'
					camera = $CamName
					search = $search_param_val
					view = $DatabaseView
					enddate = $oldest_date   # oldest time found (so far); per Ken @ Blue Iris Support 
				}
			}
		}
		
		# For debugging
		# Write-Host ("$fn_name() jsonCmd = " + ($jsonCmd | ConvertTo-Json))

		# FIRST JSON CALL
		# Returns a powershell PSCustomObject
		$json_response_obj = `
			BIJsonCall `
			-JsonCmd $jsonCmd 

		# For debugging
		if ($iter -gt 1) {
			# Write-Host ("$fn_name() jsonCmd response type = " + $json_response_obj.GetType())
			# Write-Host ("$fn_name() jsonCmd response = " + ($json_response_obj | ConvertTo-Json))
		}

		if ($json_response_obj -match 'success') {

			# Test for $json_response_obj.data object
			if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {

				$data_obj = $json_response_obj.data

				if ($data_obj) {
					
					# For debugging
					# Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
					# Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))
					# Write-Host ("$fn_name() data_obj cnt = " + ($data_obj.Count))
					# Write-Host ("$fn_name() data_obj.camera = " + $data_obj.camera + "'")
					# Write-Host ("$fn_name() data_obj.path type = " + ($data_obj.path).GetType())
					# Write-Host ("$fn_name() data_obj.memo type = " + ($data_obj.memo).GetType())

					# KEY POINT: $paths_ary is the WHILE LOOP's primary break variable
					try {
						# Get array of all 'path' values in the $data_obj
						$paths_ary = [String[]]$data_obj.path  					
					}
					catch {
						$err_msg = "ERR/JSON 'alertlist' response for '$CamName' contains no 'path' entries (database corrupt?)"
						return $err_msg
					}
					
					# Handle cases where alertlist response ALSO contains entries from $CamName's clones
					# ... the goal is to remove those entries from $data_obj
					if ($ExcludeClonesFlg) {
						# Get array of all UNIQUE 'camera' values in the $data_obj
						$obj_cams_ary = $data_obj.camera | Select-Object -Unique
						# ... alertlist contains cloned cam names if the this UNIQUE array contains more that one element 
						if ($obj_cams_ary.length -gt 1) {
							# Remove cloned camera entries from object
							$data_obj = $data_obj | Where-Object { $_.camera -eq $CamName }
							# Refresh $paths_ary
							$paths_ary = [String[]]$data_obj.path  					
						}
					}
					
					# Get the alertlist dates array
					# KEY POINT: using try:catch as a redundant quality test
					try {
						# Get array of all 'date' values in the $data_obj
						$dates_ary = [int[]]$data_obj.date  					
					}
					catch {
						$err_msg = "ERR/JSON 'alertlist' response for '$CamName' contains no 'date' entries (database corrupt?)"
						return $err_msg
					}					
	
					# PERFORM SPECIAL HANDLING FOR THE 'MEMO' PARAMETER
					# WHY? Per Blue Iris support...
					#     Prior to version 5.8.2.4, the alertlist 'memo' field was included only if it was not blank.
					#     Starting with version 5.8.3.4 it is now always included (REF: 12/18/2023 email from Ken at blueirissupport.com)
					$memo_fix_cntr = 0
					if ($alertlist_memo_always_flg) {
						$memos_ary = [String[]]$data_obj.memo
						#Write-Host ("$fn_name() memos_ary** = " + ($memos_ary | ConvertTo-Json))
					} else {
						# For debugging
						#Write-Host "$fn_name() Special handling of Alert memos" 
						
						try {
							# KEY POINT: This code line fails if ANY Alert in the JSON 'alertlist' response lacks a 'memo'
							$memos_ary = [String[]]$data_obj.memo  
						} 
						catch {
							# REPAIR THE 'BAD' ALERTS
							
							# For debugging
							Write-Warning "$fn_name() Found alerts w/ empty memos; preparing to change values to 'NA'..." 
							
							$i = 0
							foreach ($item in $data_obj) {
								try {
									$x = $item.memo
								}
								Catch {
									$data_obj[$i] | Add-Member -MemberType NoteProperty -Name "memo" -Value "NA"
									$memo_fix_cntr++
								}
								$i++
							}
							$memos_ary = [String[]]$data_obj.memo
							
							if ($data_obj['memos'].Count -ne $tmp_obj['paths'].Count) {
								$err_msg = "ERR/unable to fix missing 'memo':'value' pairs in the JSON 'alertlist' response"
								return $err_msg
							}
							
							if ($script:FromConsole -and !$SilentFlg) {
								Write-Warning "repaired $memo_fix_cntr Alerts in the JSON 'alertlist' response that lacked a 'memo' key; now 'memo':'NA'"
							}
						}
					}

					# For debugging
					if ($memo_fix_cntr -gt 0) {
						Write-Host ("$fn_name() memos_ary cnt AFTER repair = " + $memos_ary.Count)
						Write-Host ("$fn_name() memos_ary AFTER repair = " + ($memos_ary | ConvertTo-Json))
					}

					# For debugging
					#Write-Host ("$fn_name() memos_ary = " + ($memos_ary | ConvertTo-Json))

					# Initialize vars
					$this_loop_alerts_cnt = $paths_ary.Count
					$this_loop_dates_cnt = $dates_ary.Count
					
					# Shows $dates_ary endpoints
					if ($script:DevMode -and $script:FromConsole) {
						Write-Host ("(dev mode) fist date = dates_ary[0] = " + $dates_ary[0]) -ForegroundColor Cyan 
						Write-Host ("(dev mode) last date = dates_ary[" + ($this_loop_dates_cnt-1) + "] = " + $dates_ary[$this_loop_dates_cnt-1]) -ForegroundColor Cyan
					}

					# Break loop if this iteration is a single result with same oldest date
					# THIS IN AN EDGE CASE where item #2048 has 'date' X, and item #2049 also has 'date' X
					if ($this_loop_alerts_cnt -eq 1) {
						if ($dates_ary[$this_loop_dates_cnt-1] -eq $oldest_date) {
							
							if ($script:DevMode -and $script:FromConsole) {
								Write-Host ("(dev mode) JSON alertlist this_loop_alerts_cnt = " + $this_loop_alerts_cnt) -ForegroundColor Cyan
								Write-Host ""
							}
							break
						}
					}

					# Break loop if item's 'path' and 'date' parameter counts do not match
					# SOMETHING'S WRONG
					if ($this_loop_alerts_cnt -ne $this_loop_alerts_cnt) {
						if ($script:DevMode -and $script:FromConsole) {
							Write-Host "(dev mode) terminating 'alertlist' call; response is unexpected ('path' & 'date' counts unequal)" -ForegroundColor Cyan
						}
						break
					}

					# Break loop if JSON 'alertlist' call returns the same number of alerts as the 
					# previous result (<2048 items)
					if ($this_loop_alerts_cnt -lt $json_alertlist_max -and ($this_loop_alerts_cnt -eq $prev_loop_alerts_cnt)) {
						if ($al_props_obj['paths'] -contains $paths_ary[0]) {
							if ($script:DevMode -and $script:FromConsole) {
								Write-Host "(dev mode) ignoring JSON alertlist response w/ same alert count as previous response" -ForegroundColor Cyan
							}
							break
						}
					}
					
					$prev_loop_alerts_cnt = $this_loop_alerts_cnt
					
					if ($script:DevMode -and $script:FromConsole) {
						Write-Host "(dev mode) JSON alertlist this_loop_alerts_cnt = $this_loop_alerts_cnt" -ForegroundColor Cyan
					}

					# For iterations > 1, remove the first alert in the response
					# THIS IN AN EDGE CASE where item #2048 has 'date' X, and item #2049 also has 'date' X
					if ($iter -gt 1) {
						
						if ($this_loop_alerts_cnt -ge 1 -and $this_loop_alerts_cnt -le $json_alertlist_max) {
							
							# REMOVE THE FIRST OBJECT ELEMENT 
							$data_obj = $data_obj[1 .. ($data_obj.Count - 1)]

							# SPECIAL HANDLING FOR THE 'MEMO' PARAMETER
							$memos_ary = $memos_ary[1 .. ($memos_ary.Count - 1)]
							
							# KEY POINT: 
							# Note that we do not update var $paths_ary at this point
							# Why? Because it is the while loop's break variable and we still have work to do
							
							# For debugging
							#Write-Host ("$fn_name() data_obj cnt AFTER = " + ($data_obj.Count))
						}
					}

					# For debugging
					#if ($iter -eq 2) {
						#$data_obj_keys += $($data_obj[0] | Get-Member -MemberType *Property).Name
						#Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
						#Write-Host ("$fn_name() data_obj keys = " + ($data_obj_keys | ConvertTo-Json))
						#Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))
						#Write-Host ("$fn_name() data_obj.date type = " + ($data_obj.date).GetType())
					#}

					# Get all dates for this set of Alerts
					# ... This SHOULD BE an array of UTC timestamps, ordered largest to smallest
					# ... NOTE: this array is also used for QUALITY testing
					[Int[]]$dates_ary = $data_obj.date
					$dates_cnt = $dates_ary.Count

					# HANDLE SIMULATIONS
					if ($script:Sim5) {  #simulate response w/ zero-date (timestamp) values; arg value is CamName
						if ($Cameras -contains $script:Sim5) {
							# Set first date to zero
							#$dates_ary[0] = 0
							#$data_obj[0].date = 0

							# Set last date to zero
							#$dates_ary[$dates_cnt-1] = 0
							#$data_obj[$dates_cnt-1].date = 0
						
							# Set 2nd date to zero
							$dates_ary[1] = 0
							$data_obj[1].date = 0

							# Set 2nd to last date to zero
							$dates_ary[$dates_cnt-2] = 0						
							$data_obj[$dates_cnt-2].date = 0
						
							# For debugging
							Write-Host ("SIM5 fist date = dates_ary[0] = " + $dates_ary[0])
							Write-Host ("SIM5 last date = dates_ary[" + ($dates_cnt-1) + "] = " + $dates_ary[$dates_cnt-1])
						}
					}
					
					# Get the oldest/newest dates for this set of Alerts 
					# ... The oldest date is needed for the next JSON 'alertlist' call, if necessary
					$newest_date = $dates_ary[0]
					$oldest_date = $dates_ary[$dates_cnt-1]
					$max_date = ($dates_ary | Measure-Object -Maximum).Maximum
					$min_date = ($dates_ary | Measure-Object -Minimum).Minimum
					$dates_sum = ($dates_ary | Measure-Object -Sum).Sum
					
					# For debugging
					#Write-Host ("$fn_name() dates_ary type = " + $dates_ary.GetType())
					#Write-Host "$fn_name() dates_cnt = $dates_cnt"
					#Write-Host "$fn_name() dates_ary = $dates_ary"
					#Write-Host ("$fn_name() newest_date type = " + $newest_date.GetType())
					#Write-Host "$fn_name() newest_date = $newest_date"
					#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
					#Write-Host "$fn_name() oldest_date = $oldest_date"
					#Write-Host "$fn_name()dates_sum = $dates_sum"
					#Write-Host ("$fn_name() data_obj.date = " + ($data_obj.date | ConvertTo-Json))

					# Check for more QUALITY issues...
					# ... 1. Check if all dates are zero (FATAL)
					if ($dates_sum -eq 0 ) {
						$skipped_next_json_call_flg = $true
						if ($script:FromConsole -and !$SilentFlg) {
							Write-Warning "JSON 'alertlist' response for '$CamName' contains only zero 'date' entries (database corrupt?); find w/ -DevMode"
						}
						break
					}
					# ... 2. Check if the first array element is the max date (IT SHOULD BE)
					$first_date_is_not_max = ($newest_date -lt $max_date)
					# ... 3. Check that the first and last array elements are NOT zero (THEY SHOULD BE)
					$first_date_is_zero = ($newest_date -eq 0)
					$last_date_is_zero = ($oldest_date -eq 0)
					$zero_dates_cntr = 0  # Initialize counter
					# ... 4. Check if any OTHER alerts have dates with zero values
					#        Make hashtable like @("date1":count, "date2":count ...}
					#        if zero-dates are found, $ht will have and entry like {..., "0": 23, ...}
					# ... REF: https://stackoverflow.com/a/12640991
					$dates_ary | Group | % { $ht = @{} } { $ht[$_.Name] = $_.Count } { $ht } | Out-Null
					
					# For debugging
					#Write-Host ("ht = " + ($ht | ConvertTo-Json))
					#Write-Host ("ht['0'] = " + $ht['0'])
					
					if ($ht['0']) {
						$zero_dates_cntr += $ht['0']
						$dates_between_zero_cntr = $ht['0']
						
						# For debugging
						#Write-Host ("ht['0'] type = " + ($ht['0']).GetType())
						
					} else {
						$dates_between_zero_cntr = 0
					}
					if ($first_date_is_zero -and $dates_between_zero_cntr -gt 0) {
						$dates_between_zero_cntr--
					}
					if ($last_date_is_zero -and $dates_between_zero_cntr -gt 0) {
						$dates_between_zero_cntr--
					}
					if ($script:FromConsole -and !$SilentFlg) {
						if ($first_date_is_zero) {
							if ($last_date_is_zero) {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' begins and ends with an Alert with zero-date values"
							} else {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' begins with an Alert with a zero-date value"
							}
						} else {
							if ($first_date_is_not_max) {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' does not have the newest date in the first entry (as expected)"
							}
							if ($last_date_is_zero) {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' ends with an Alert with a zero-date value"
							}
						}
						if ($dates_between_zero_cntr -gt 0) {
							Write-Warning "JSON 'alertlist' response #$iter for '$CamName' has $zero_dates_cntr zero-date values; find w/ -DevMode"
						}
					}
					
					# For debugging
					#Write-Host ("$fn_name() newest_date type = " + $newest_date.GetType())
					#Write-Host "$fn_name() newest_date = $newest_date"
					#Write-Host ("$fn_name() newest_date for iter #" + $iter + " ($this_loop_alerts_cnt alerts) = " + $newest_date)
					#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
					#Write-Host "$fn_name() oldest_date = $oldest_date"

					# Load collectors
					$tmp_obj['cameras']   = $data_obj.camera  # an array (adds a column of values to the hashtable object)
					$tmp_obj['paths']     = $data_obj.path
					$tmp_obj['clips']     = $data_obj.clip
					$tmp_obj['files']     = $data_obj.file
					#$tmp_obj['memos']     = $data_obj.memo   # We handle '$data_obj.memo' differently 
					$tmp_obj['memos']     = $memos_ary
					$tmp_obj['offsets']   = $data_obj.offset
					$tmp_obj['flagss']    = $data_obj.flags
					$tmp_obj['ress']      = $data_obj.res
					$tmp_obj['zoness']    = $data_obj.zones
					$tmp_obj['dates']     = $data_obj.date	# Note: $date_ary[] is reverse-ordered.
					$tmp_obj['colors']    = $data_obj.color
					$tmp_obj['filesizes'] = $data_obj.filesize

					# For debugging
					#Write-Host ("$fn_name() tmp_obj['memos'] = " + ($tmp_obj['memos'] | ConvertTo-Json))
	
					# HANDLE SIMULATIONS
					if ($script:Sim1) {  # remove ALL 'memo':'value' pairs from raw JSON 'alertlist' response
						foreach ($item in $data_obj) {
							$item.PSObject.Properties.Remove('memo')
						}
						if ($script:FromConsole) {
							Write-Host "(test mode) Simulation #1: removed ALL 'memo':'value' pairs from JSON 'alertlist' response" -Foregroundcolor Green
							#Write-Host ("(test mode) Simulation #1: data_obj AFTER = " + ($data_obj | ConvertTo-Json)) -Foregroundcolor Green
						}
					}
					if ($script:Sim2) {  # remove 'memo':'value' pairs for 3 random alerts from JSON 'alertlist' response
						try {
							$tmp_obj['memos'] = [Array]$data_obj.memo
							$total_cnt = $tmp_obj['paths'].Count
							$chgd_memos_cnt = [Math]::min(3,$alerts_cnt-1)
							$chgd_idxs_ary = @()
							$chgd_alerts_obj = @()
							$i = 0
							while ($i -lt $chgd_memos_cnt) {
								$idx = Get-Random -Maximum $total_cnt
								# Only protected alerts
								#Write-Host ("flags = " + ($data_obj[$idx]).flags)
								$flags_ary = `
									MakeBitArray `
									-DecVal ($data_obj[$idx]).flags `
									-NoBits 32
								# Flags: 2^0=flagged, 2^1=protected, 2^6=archive, 2^9=export, 2^28 confirmed
								if ($flags_ary[2] -eq 1) {
									if (!($chgd_idxs_ary -contains $idx)) {
										$data_obj[$idx].PSObject.Properties.Remove('memo')
										$chgd_alerts_obj += $data_obj[$idx]
										$chgd_idxs_ary += $idx
										$i++
									}
								}
							}
							if ($script:FromConsole) {
								Write-Host "(test mode) Simulation #2: removed JSON alertlist 'memo' entry from $chgd_memos_cnt random protected Alerts" -Foregroundcolor Green
								Write-Host ("(test mode) Simulation #2: changed alertlist entries = " + ($chgd_alerts_obj | ConvertTo-Json)) -Foregroundcolor Green
							}
						} catch {
							if ($script:FromConsole) {
								Write-Host "(test mode) Simulation #2: unable to run simulation; no Alerts have a 'memo':'value' pair"
							}
						}
					}
					if ($script:Sim3) {  # simulate response w/ all 'memo'='NA' params
						$tmp_obj['memos'] = 1..$tmp_obj['paths'].Count | % { 'NA' }
						if ($script:FromConsole) {
							Write-Host "(test mode) Simulation #3: set 'memo'='NA' for all Alerts in $fn_name() response" -Foregroundcolor Green
							Write-Host ("(test mode) Simulation #3: tmp_obj['memos'] modified = " + ($tmp_obj['memos'] | ConvertTo-Json)) -Foregroundcolor Green
						}
					}
					if ($script:Sim4) {  # simulate response w/ 3 random 'memo'='NA' params
						$memos_cnt = $tmp_obj['memos'].Count
						if ($memos_cnt -gt 0) {
							#Write-Host "memos_cnt = $memos_cnt"
							$memos_cnt = $memos_cnt
							$chgd_memos_cnt = [Math]::min(3,$memos_cnt-1)
							$chgd_idxs_ary = @()
							$chgd_alerts_ary = @()
							$i = 0
							while ($i -lt $chgd_memos_cnt) {
								$idx = Get-Random -Maximum $memos_cnt
								#Write-Host "idx = $idx"
								# Only protected alerts
								#Write-Host ("flags = " + ($data_obj[$idx]).flags)
								$flags_ary = `
									MakeBitArray `
									-DecVal ($data_obj[$idx]).flags `
									-NoBits 32

								# Flags: 2^0=flagged, 2^1=protected, 2^6=archive, 2^9=export, 2^28 confirmed
								if ($flags_ary[2] -eq 1) {
									if (!($chgd_idxs_ary -contains $idx)) {
										if ($tmp_obj['memos'][$idx] -ne 'NA') {
											$tmp_obj['memos'][$idx] = 'NA'
											$chgd_alerts_ary += (($data_obj[$idx]).file).Replace('.jpg','')
											$chgd_idxs_ary += $idx
											$i++
										}
									}
								}
							}
							if ($script:FromConsole) {
								Write-Host "(test mode) Simulation #4: set 'memo'='NA' for $chgd_memos_cnt random Alerts in $fn_name() response" -Foregroundcolor Green
								#Write-Host ("(test mode) Simulation #4: tmp_obj['memos'] modified = " + ($tmp_obj['memos'] | ConvertTo-Json)) -Foregroundcolor Green
								Write-Host ("(test mode) Simulation #4: modified AlertIDs = " + ($chgd_alerts_ary | ConvertTo-Json)) -Foregroundcolor Green
							}
						}
					}

					# For debugging
					
					#Write-Host ("$fn_name() tmp_obj['paths'] cnt = " + $tmp_obj['paths'].Count)
					#Write-Host ("$fn_name() tmp_obj['paths'] = " + ($tmp_obj['paths'] | ConvertTo-Json))
					#Write-Host ("$fn_name() tmp_obj['flagss'] cnt = " + $tmp_obj['flagss'].Count)
					#Write-Host ("$fn_name() tmp_obj['flagss'] = " + ($tmp_obj['flagss'] | ConvertTo-Json))
					#Write-Host ("$fn_name() tmp_obj['memos'] dnt = " + $tmp_obj['memos'].Count)
					#Write-Host ("$fn_name() tmp_obj['memos'] = " + ($tmp_obj['memos'] | ConvertTo-Json))

					# Merge results with the master accumulator
					$al_props_obj['cameras'] = $al_props_obj['cameras'] + $tmp_obj['cameras']
					$al_props_obj['paths']   = $al_props_obj['paths']   + $tmp_obj['paths']
					$al_props_obj['clips']   = $al_props_obj['clips']   + $tmp_obj['clips']
					$al_props_obj['files']   = $al_props_obj['files']   + $tmp_obj['files']
					$al_props_obj['memos']   = $al_props_obj['memos']   + $tmp_obj['memos']
					$al_props_obj['offsets'] = $al_props_obj['offsets'] + $tmp_obj['offsets']	
					$al_props_obj['flagss']  = $al_props_obj['flagss']  + $tmp_obj['flagss']	
					$al_props_obj['ress']    = $al_props_obj['ress']    + $tmp_obj['ress']	
					$al_props_obj['zoness']  = $al_props_obj['zoness']  + $tmp_obj['zoness']	
					$al_props_obj['dates']   = $al_props_obj['dates']   + $tmp_obj['dates']	
					$al_props_obj['colors']  = $al_props_obj['colors']  + $tmp_obj['colors']	
					$al_props_obj['filesizes']   = $al_props_obj['filesizes']   + $tmp_obj['filesizes']	
					
					# Stop requesting data if the last date is zero
					if ($oldest_date -eq 0) {
						$skipped_next_json_call_flg = $true
						if ($script:FromConsole -and !$SilentFlg) {
							Write-Warning "skipping next JSON 'alertlist' call because this call's oldest date is zero"
						}
						break
					}
				
				} else {
					break
				}
			} else {
				break
			}
		} else {
			break
		}

		# Avoid infinite loop
		if ($iter -gt $max_tries) {
			break
		}
		
		$iter++
	}
	
	# For debugging
	#Write-Host "Broke out of While loop"

	# TEST FOR MORE QUALITY ISSUES (indicators of a possibly corrupt database)
	# ... Test if 'clip' values contain '@-1.jpg' entries (WARNING)
	[String[]]$clips_ary = $al_props_obj['clips']
	if ($clips_ary -contains '@-1.jpg') {
		$clips_cnt = $clips_ary.Count
		$bad_clip_cntr = 0
		foreach($item in $clips_ary) {
			if ($item -eq '@-1.jpg') {
				$bad_clip_cntr++
			}
		}
		if ($script:FromConsole -and !$SilentFlg) {
			Write-Warning "JSON 'alertlist' response for '$CamName' has $bad_clip_cntr/$clips_cnt 'clip'='@-1.jpg' entries; find w/ -DevMode"
		}
	}

	# For debugging
	#Write-Host ("$fn_name() al_props_obj type = " + ($al_props_obj.GetType()))
	#Write-Host ("$fn_name() al_props_obj keys = " + $al_props_obj.Keys)
	#Write-Host ("$fn_name() al_props_obj = " + ($al_props_obj | ConvertTo-Json))
	#Write-Host ("$fn_name() al_props_obj['cameras'] = " + ($al_props_obj['cameras'] | ConvertTo-Json))
	#Write-Host ("$fn_name() al_props_obj['memos'] = " + ($al_props_obj['memos'] | ConvertTo-Json))
	#Write-Host ("$fn_name() al_props_obj['flagss'] = " + ($al_props_obj['flagss'] | ConvertTo-Json))
	#Write-Host ("$fn_name() al_props_obj.paths = " + $al_props_obj.paths)
	
	# DEV MODE - Save each camera's 'alertlist' data
	if ($script:DevMode -and $script:FromConsole) {
		if (!$script:Sim1 -and !$script:Sim2 -and !$script:Sim3 -and !$script:Sim4) {
			
			$al_objects_ary = @()  # Initialize accumulator; an array of hashtables
			$keys = [Array]$al_props_obj.Keys
			$all_alerts_cnt = $al_props_obj['paths'].Count
			$alertlist_filespec = "$DevModePath\alertlist.$CamName.json"
			
			# For debugging
			#Write-Host ("$fn_name() keys = " + ($keys | ConvertTo-Json))
			#Write-Host "$fn_name() all_alerts_cnt = $all_alerts_cnt"
			#Write-Host "$fn_name() alertlist_filespec = '$alertlist_filespec'"
			
			if ($all_alerts_cnt -gt 0) {

				# Possible keys in $al_props_obj
				$keys1 = @('cameras','memos','paths','clips','files','filesizes','dates','offsets')
				$keys2 = @('flagss','zoness')

				for ($i=0; $i -lt $all_alerts_cnt; $i++) {
					$ht = @{}  # Initialize hashtable
					# Modify $al_props_obj key for use in hashtable 
					
					#Write-Host ""
					
					foreach ($key in $keys) {
						if ($keys1 -contains $key) {
							$key_new = $key.TrimEnd('s')
						} else {  # $keys_2 -contains $key
							$key_new = $key.Replace('ss','s')
						}
						
						# For debugging
						#Write-Host "key = $key"
						#Write-Host ("al_props_obj[$key] type = " + ($al_props_obj[$key]).GetType())
						#Write-Host ("al_props_obj[$key] = " + ($al_props_obj[$key] | ConvertTo-Json))
						#Write-Host ("al_props_obj[$key][0] = " + $al_props_obj[$key][0])
						#Write-Host ("al_props_obj[$key][" + ($all_alerts_cnt-1) + "] = " + $al_props_obj[$key][$all_alerts_cnt-1])
						#Write-Host ("al_props_obj[$key][$i] = " + $al_props_obj[$key][$i])
						
						$ht.Add($key_new, $al_props_obj[$key][$i])	
						#exit
						
					}
					$al_objects_ary += $ht
				}

				# For debugging
				#Write-Host ("al_list_ary = " + ($al_list_ary | ConvertTo-Json))

				$alertlist_saved_flg = $false
				if (!$Cameras -and !$Pattern) {
					$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"
					$alertlist_saved_flg = $true
				} else {
					if ($Cameras) {
						$match_str = ($Cameras -Join '|')
						if ($match_str -match $CamName) {
							$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"	
							$alertlist_saved_flg = $true
						}
					}				
					if ($Pattern) {
						if ($Pattern -match $CamName) {
							$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"	
							$alertlist_saved_flg = $true
						}
					}				
				}
				if ($alertlist_saved_flg) {		
					if ($script:FromConsole) {
						Write-Host "(dev mode) 'alertlist' response saved to disk ($all_alerts_cnt items) --> '$alertlist_filespec'" -ForegroundColor Cyan
						#Write-Host ""
					}
				}
				
				# Top open this *.json file in Excel...
				# 1. Data > Get Data > From File > From JSON
				# 2. Navigate to *json file
				# 3. Click 'Import' 
				# 4. Click 'To Table'
				# 5. Click 'OK'
				# 6. Click icon by 'Column1'
				# 7. Click 'OK'
				# 8. Click 'Close & Load'
				
			}  else {
				if ($script:FromConsole -and !$SilentFlg) {
					Write-Warning "no alerts; unable to save alertlist to disk ($alertlist_filespec)"
				}
			}
		} else {
			if ($script:FromConsole) {
				Write-Host "(dev mode) HEADS-UP: cam 'alertlist' *NOT* saved to disk when using arguments -SimX" -ForegroundColor Cyan
			}
		}
	}

	# Get all alerts files, paths
	$all_alerts_paths = $al_props_obj.paths
	$all_alerts_files = $al_props_obj.files
	$all_alerts_cnt = $all_alerts_paths.Count
	
	$all_alerts_zones = $al_props_obj.zoness

	# For debugging
	#Write-Host ("$fn_name() AlertPath = " + $AlertPath)
	#Write-Host ("$fn_name() all_alerts_paths cnt = " + $all_alerts_paths.Count)
	#Write-Host ("$fn_name() all_alerts_paths = " + ($all_alerts_paths | ConvertTo-Json))
	#Write-Host ("$fn_name() all_alerts_files = " + ($all_alerts_files | ConvertTo-Json))
	#Write-Host ("$fn_name() all_alerts_zones type = " + $all_alerts_zones.GetType())
	#Write-Host ("$fn_name() all_alerts_zones cnt = " + $all_alerts_zones.Count)
	#Write-Host ("$fn_name() all_alerts_zones = " + ($all_alerts_zones | ConvertTo-Json))
	#Write-Host "$fn_name() Found $all_alerts_cnt total items`n"

	# Initialize collectors
	$rtn_obj = [ordered]@{}
	$new_obj = [ordered]@{}
	if ($AlertProps.Count -gt 0) {
		foreach ($alert_prop in $AlertProps) {
			$new_obj.Add($alert_prop, @())
		}
	} else {
		foreach ($al_props_key in $al_props_obj_keys) {
			$new_obj.Add($al_props_key, @())
		}
	}
	
	# For debugging
	#Write-Host ("$fn_name() new_obj = " + ($new_obj | ConvertTo-Json))

	# WALK THROUGH ALERTS LIST
	for ($i=0; $i -lt $all_alerts_cnt; $i++) { 
	
		# For debugging
		#Write-Host "`nalert idx = $i..."
			
		if ($AlertPath -ne "*") {
			if (!($all_alerts_paths[$i] -match $AlertPath)) {
				Continue
			} else {
				# For debugging
				#Write-Host "Found match for AlertPath ($AlertPath)"
			}
		}
		
		if ($Pattern.Length -gt 0) {
			if (!($all_alerts_files[$i] -match $Pattern)) {
				Continue
			} else {
				# For debugging
				#Write-Host ("Pattern '$Pattern' is a match for '" + $all_alerts_files[$i] + "'")
			}
		}

		if ($AlertProps.Count -gt 0) {
			
			# GET REQUESTED PROPERTIES

			foreach ($alert_prop in $AlertProps) {

				$key = $alert_prop + "s"  # $al_props_obj's keys are plural
				
				[Array]$arr = $al_props_obj.$key
				
				# For debugging
				#if ($key -eq 'memos') {
					#Write-Host ("$fn_name() arr[$key] = " + ($arr | ConvertTo-Json))
				#}
				
				if (!($null -eq $arr)) {  # KEY POINT: $arr can be a single element array @(0)
					
					$val = $arr[$i]
					
					# For debugging
					#if ($arr[$i] -eq '') {
						#Write-Host ($key + " = '" + $arr[$i] + "'")
					#}

					# Append value to collector
					if ($arr[$i] -or $arr[$i] -eq 0) {
						$new_obj.$alert_prop += $arr[$i]				
					} else {
						$new_obj.$alert_prop += ''
					}
					
					# For debugging
					#if ($arr[$i] -eq '') {
						#Write-Host ("new_obj.$alert_prop = " + ($new_obj.$alert_prop | ConvertTo-Json))
					#}

					# For debugging
					#Write-Host ("arr = " + $arr)
					#Write-Host ("item #" + ($i+1) + " $alert_prop = " + $arr[$i])

				} else {
					if ($i -eq 0) {
						if ($script:FromConsole) {
							Write-Host ("WARNING/nonexistent AlertProp '$alert_prop', expecting: [" + ($AlertProps -Join ',') + "]")
						}
					}
				}
			}
			
		} else {
			
			# GET ALL PROPERTIES
			
			# Get all property names
			if ($i -lt 2) {
				
				# NOTE: $al_props_obj's keys are plural
				$al_props_obj_keys = $al_props_obj.Keys

				# For debugging
				#Write-Host ("al_props_obj_keys = " + $al_props_obj_keys)
				
			}
			
			foreach ($al_props_key in $al_props_obj_keys) {

				# For debugging
				#Write-host ("new_obj.$al_props_key type = " + ($new_obj.$al_props_key).GetType())

				# Get all values for the property (an array)
				$vals = $al_props_obj.$al_props_key

				# Remove the terminal 's' from $al_props_key
				$new_obj_key = $al_props_key.Substring(0,$al_props_key.Length-1) 
				
				if ($vals) {
					$new_obj.$new_obj_key += $vals[$i]					
				} else {
					#$new_obj.$new_obj_key += $null
					$new_obj.$new_obj_key += ''
				}
				
				#For debugging
				#Write-Host ("item #" + ($i+1) + " $new_obj_key = " + $vals[$i])
			}
		}
			
		# For debugging
		#Write-host ("new_obj = " + ($new_obj | ConvertTo-Json))
	}
	
	$rtn_obj[$CamName] = $new_obj

	# For debugging
	#Write-Host ("rtn_obj cnt = " + $rtn_obj.count)
	#Write-Host "rtn_obj keys = "$rtn_obj.Keys
	#Write-Host ("rtn_obj keys = " + ($rtn_obj | ConvertTo-Json))
	#Write-Host ("rtn_obj = " + ($rtn_obj | ConvertTo-Json))s
	#Write-Host ("rtn_obj.FE = " + ($rtn_obj.FE | ConvertTo-Json))
	#Write-Host ("rtn_obj.DW1.BPS = '" + $rtn_obj.DW1.BPS + "'")
	#Write-Host ("rtn_obj.GT.path = '" + $rtn_obj.GT.path + "'")
	#Write-Host ("rtn_obj.GT.memo = '" + $rtn_obj.GT.memo + "'")
	#Write-Host ("rtn_obj.GT.flags = '" + $rtn_obj.GT.flags + "'")

	return $rtn_obj
}

function BIJsonGetAlertList() {
# PURPOSE:  To get properties from the Blue Iris JSON interface cmd 'alertlist'
# USAGE:    {this_script_name}.ps1 prop cam(s) prop(s)
# PARAMS:   -Cameras        = (string) filter results for Alerts with specified camera shortnames(comma-delimited),  
#                                      like 'DW1','DW2,'FD'
#           -Patterns       = (string) filter results for Alerts with specified descriptor substrings
#           -DatabaseView   = (string) default is 'alerts'
#                             options: all, new, stored, alerts, aux1, aux2, aux3, aux4, aux5, aux6, aux7,
#                                      flagged, export, archive, people, vehicles, confirmed, or canceled                             
#                             Note: When view='flagged', clip items are included too, but they don't have msec metadata.
#           -MemoTag        = (string) filter results for Alerts with specified memo substrings
#           -AlertProps     = (string) return only these JSON 'alertlist' cmd params (comma-delimited )
#                                      like 'enabled,profile,motion'
#           -ExcludeClones  = (switch) exclude clones of cam names in -Cameras
# RETURNS:  (hashtable) @{ Cam1: @{ prop1:[values], ... , propN:[values] }, Cam2: @{ prop1:[values], ... , propN:[values] }, ... }
# REQUIRES: Function: 'BIJsonAlertListParse()'
# GLOBALS:  Used global vars: $FromConsole, $Quiet, $DevMode, $DevModePath 
#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
#           $script:Quiet       = (boolean) $true suppresses verbose output to the console
#
# Created by jaydeel at ipcamtalk.com
# Version: 8 (8/29/2024) replaced *Flg args w/ global vars
# Version: 9 (9/9/2024) fixed bug where alertlist response unexpectedly contained cloned camera entries

	param(
		[String[]]$Cameras,
		[String[]]$Patterns,
		[String]$DatabaseView,      # default is 'alerts'
		[String]$MemoTag,
		[String[]]$AlertProps,
		[Switch]$ExcludeClones
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered function $fn_name() ..."
	# Write-Host "$fn_name() Cameras = $Cameras"
	# Write-Host "$fn_name() Patterns = $Patterns"
	# Write-Host "$fn_name() DatabaseView = $DatabaseView"
	# Write-Host "$fn_name() MemoTag = $MemoTag"
	# Write-Host ("$fn_name() AlertProps = " + ($AlertProps | ConvertTo-Json))
	# Write-Host "$fn_name() ExcludeClones = $ExcludeClones"
	# Write-Host ("$fn_name() FromConsole = " + $script:FromConsole)
	# Write-Host "$fn_name() Sim1 = $script:Sim1"
	# Write-Host "$fn_name() Sim3 = $script:Sim3"

	# HANDLE ARGS
	if ($Cameras) {
		$my_cams_ary = $Cameras -Split ','
	} else {
		$my_cams_ary = @()
	}
	$my_cams_cnt = $my_cams_ary.Count

	if ($Patterns) {
		$my_patterns_ary = $Patterns -Split ','
	} else {
		$my_patterns_ary = @('')
	}
	$my_patterns_cnt = $my_patterns_ary.Count

	if ($DatabaseView) {
		$db_view = $DatabaseView
	} else {
		$db_view = 'alerts'
	}
	
	if ($AlertProps.Count -eq 0) {
		$my_props_ary = @('camera', 'path', 'clip', 'file', 'memo', 'offset', 'flags', 'res', 'zones', 'date', 'color', 'filesize')
	} else {
		$my_props_ary = $AlertProps -Split ','
	}
	$my_props_cnt = $my_props_ary.Count

	# For debugging
	#Write-Host ("$fn_name() my_cams_ary type = " + $my_cams_ary.GetType())
	#Write-Host ("$fn_name() my_cams_ary = " + $my_cams_ary)
	#Write-Host ("$fn_name() my_cams_ary cnt = " + $my_cams_ary.Count)
	#Write-Host "my_patterns_ary = $my_patterns_ary"
	#Write-Host "my_flags_ary = $my_flags_ary"
	#Write-Host ("$fn_name() my_props_ary cnt = " + $my_props_ary.Count)
	#Write-Host ("$fn_name() my_props_ary = " + ($my_props_ary | ConvertTo-Json))

	# INITIALIZE
	$rtn_obj = @{}
	$progress_bar_cntr = 0
	#$cntr = 1
	$patterns_found_ary = @()
	$i = 0
	foreach ($my_cam in $my_cams_ary) {

		# For debugging
		# Write-Host ("$fn_name() my_cam[$i] = " + $my_cam)
		
		# Show progress bar
		if ($script:FromConsole) {
			$progress_bar_cntr++
			$WriteProgressParams = @{
				Activity         = "Parsing alertlist data..."
				Status           = "Progress:  cam '$my_cam' ($progress_bar_cntr of $my_cams_cnt requested cams)"
				PercentComplete  = (($progress_bar_cntr / $my_cams_cnt) * 100)
			}
			# Note: If loop is too fast, you might need a delay to get the progress bar to update
			#Start-Sleep -Milliseconds 150				
			Write-Progress @WriteProgressParams
		}
		
		$pattern_cntr = 0
		foreach ($pattern in $my_patterns_ary) {
			
			# For debugging
			#Write-Host "`n----- processing pattern '$pattern' -----"
			
			# Skip pattern search if already found
			if ($patterns_found_ary -contains $pattern) {
				continue
			}

			# GET JSON DATA FROM THE BLUE IRIS WEB SERVER
			# ... returns an ordered hashtable
			$response_obj = `
				BIJsonAlertListParse `
				-CamName $my_cam `
				-Pattern $pattern `
				-AlertProps $my_props_ary `
				-MemoTag $MemoTag `
				-DatabaseView $db_view `
				-ExcludeClonesFlg $ExcludeClones `
				-SilentFlg ($pattern_cntr -gt 0)

			# For debugging
			#Write-Host ("$fn_name() response_obj type = " + $response_obj.GetType())
			#Write-Host ("$fn_name() response_obj = " + ($response_obj | ConvertTo-Json))
			
			if ($response_obj.GetType() -match 'string') {
				if ($response_obj.StartsWith('ERR/')) {
					return $response_obj
				} else {
					$err_msg = "ERR/unexpected response from AlertListParse(); $response_obj"
					return $err_msg
				}
			}

			# $response_obj is an ordered hashtable, like
			# @{
			# 	DW1: @{
			#		"file": [
			#           filename1,
			#           ...,
			#           filenameN,
			#       ],
			#       "flags": [
			#           int1,
			#           ...,
			#           int2,
			#       ],
			#       "memo": [
			#           string1,
			#           ...,
			#           stringN,
			#       ],
			#	}
			# }

			# INITIALIZE
			$x = ''

			# EXTRACT PROPERTY VALUE(S)
			# ... Note: null/blank $response_obj.ram observed on occasion
			if ($response_obj) {	

				$response_obj_keys = $response_obj.Keys

				# For debugging
				#Write-Host ("$fn_name() response_obj type = " + $response_obj.GetType())
				#Write-Host ("$fn_name() response_obj cnt = " + $response_obj.Count)
				#Write-Host ("$fn_name() response_obj keys = " + $response_obj_keys)
				#Write-Host ("$fn_name() response_obj = " + ($response_obj | ConvertTo-Json))

				if ($response_obj_keys -contains $my_cam) {
					
					$cam = $my_cam
					$item_obj = $response_obj[$cam]
					$item_obj_keys = $item_obj.Keys

					# $item_obj (and $new_obj below) are System.Collections.Specialized.OrderedDictionary 
					# that should look like...
					<# 
					  "camera": ["DW1"],
					  "path": ["@224803430.bvr"],
					  "clip": ["@224669140.bvr"],
					  "file": ["DW1.20240128_020000.3152644.5-0.jpg"],
					  "memo": ["LD/C;vehicle:93%"],
					  "offset": ["3157680"],
					  "flags": [402980864],
					  "res": ["2688x1520"],
					  "zones": [65],
					  "date": [1705777428],
					  "color": [16711808],
					  "filesize": ["22 sec (523K)"]
					}
					#>					

					# For debugging
					#Write-Host ("$fn_name() item_obj type = " + $item_obj.GetType())
					#Write-Host ("$fn_name() item_obj = " + ($item_obj | ConvertTo-Json))
					#Write-Host ("$fn_name() item_obj_keys = " + ($item_obj_keys | ConvertTo-Json))
					#Write-Host ("$fn_name() item_obj = " + ($item_obj | ConvertTo-Json))
					#rite-Host ("$fn_name() item_obj[0] type = " +  ($item_obj[$item_obj_keys[0]]).GetType())
					#Write-Host ("$fn_name() item_obj[0] = " +  $item_obj[$item_obj_keys[0]])
					
					# Test for empty result
					if (!$item_obj[$item_obj_keys[0]]) {  # looking for $item_obj['camera'] = @()
						continue
					} else {
						if ($patterns_ary.Count -gt 1) {
							$patterns_found_ary += $pattern
						}
					}

					$new_obj = [ordered]@{}	
					
					if ($my_props_ary.Length -gt 0) {
						$props_ary = $my_props_ary   # RETURN REQUESTED PROPERTIES
					} else {
						$props_ary = $item_obj_keys  # RETURN ALL PROPERTIES			
					}

					# RETURN PROPERTIES

					foreach ($cam_prop in $props_ary) {
						
						$item_obj_ary = $item_obj.$cam_prop  # expect $item_obj.path = @(value)

						# For debugging
						#Write-Host "$fn_name() Processing property '$cam_prop' ..."
						#Write-Host ("$fn_name() item_obj.$cam_prop type = " + ($item_obj.$cam_prop).GetType())
						#Write-Host ("$fn_name() item_obj.$cam_prop ary = " + ($item_obj_ary | ConvertTo-Json))
						#Write-Host ("$fn_name() cam_prop = $cam_prop, val = '" + $item_obj_ary[0] + "'")
						
						# KEY POINT: $null must put on the left side of operator!!
						# REF: https://stackoverflow.com/a/49411823
						if ($null -eq $item_obj_ary) {
							$new_obj.Add($cam_prop, @('NA'))
							$x = 'NA'  # For single camera, single property requests								
						} else {
							if ($null -eq $item_obj_ary[0]) {
								$new_obj.Add($cam_prop, @('NA'))
								$x = 'NA'  # for single camera, single property requests
							} else {
								$new_obj.Add($cam_prop, $item_obj_ary)
								$x = $item_obj_ary[0]  # for single camera, single property requests
							}
						}
						
						# For debugging
						#Write-Host ("$fn_name() " + $cam_prop + " = '"  + $x + "'")	
					}
					
					# For debugging
					#Write-Host ("$fn_name() new_obj = " + $new_obj)
					#Write-Host ("$fn_name() new_obj = " + ($new_obj | ConvertTo-Json))

					if ($my_patterns_cnt -lt 1) {
						$rtn_obj[$cam] = $new_obj
					} else {
						# Handling multiple -Patterns is a bit more complex
						foreach ($item_obj_key in $item_obj_keys) {
							if ($pattern_cntr -eq 0) {
								# First time...
								if (!$rtn_obj[$cam]) {
									# Initialize the $rtn_obj's $cam keys 
									$rtn_obj.Add($cam, @{})
								}
								# Add & initialize the $rtn_obj[$cam]'s key's prop-value pairs
								$rtn_obj[$cam].Add($item_obj_key, @{})
								$rtn_obj[$cam][$item_obj_key] = @()
							}
							$rtn_obj[$cam][$item_obj_key] += $new_obj[$item_obj_key]
						}
					}

				} else {
					if ($script:FromConsole) {
						$err_msg = "ERR/nonexistent camera"
						return $err_msg
					}
					
					$rtn_obj[$cam] = 'NA'
				}
				
			} else {
				if ($script:FromConsole) {
					Write-Warning "JSON query returned null/blank`n"
				}
			}

			# Handle single camera, single property request
			if ($rtn_obj -ne 'NA') {
				if ($my_cams_ary.Length -eq 1 -and $my_props_ary.Length -eq 1) {
					$rtn_obj = $x
				}
			}
			
			$pattern_cntr++
		}
		$i++
	}

	# LOG OUT OF API SESSION
	$jsonCmd = @{cmd = 'logout'}
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd
		
	if ($script:FromConsole) {
		if ($json_response_obj.result -eq 'fail') {
			Write-Warning "BI API: log out fail; connection not terminated"
		} else {
			if (!$script:Quiet) {
				#Write-Host "BI API: log out successful; connection terminated"	
			}
		}
	}
	
	# For debugging
	#Write-Host ("$fn_name() rtn_obj = " + ($rtn_obj | ConvertTo-Json))

	return $rtn_obj
}

function BIJsonCamListParse() {
# PURPOSE:   Get Blue Iris web server JSON 'camlist' command response (for all or subset of BI cameras)
#            Intended to be called by other PowerShell scripts.
# ARGUMENTS: -CamProps = [OPTIONAL] (list) like 'bps,bps2' ; comma delimited
# RETURNS:  (object) PowerShell hashtable, @('cam':camName,'camProp':value}
#		                                   @{camName:{[camProp1:value1],[camProp2:value2],...},...}"
# REQUIRES:  function BIJsonCall()
#
# Description of JSON cmd 'camlist' from the Blue Iris Help PDF:
# 	"returns a list of cameras on the system ordered by group. Cameras not belonging to any
#	group are shown beneath the "all cameras" group. Disabled cameras are placed at the end of
#	the list. An array of objects is returned in data (note the [] surrounding a JSON array), each
#	describing a camera or a camera group."
#
# Version: 1
# Created by jaydeel at ipcamtalk.com

	# MUST BE THE FIRST LINE IN A SCRIPT
	param(
		[string]$CamProps,  # [OPTIONAL] list
		[boolean]$QuietFlg,
		[boolean]$FromConsoleFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered $fn_name() ..."
	# Write-Host "$fn_name() CamProps = $CamProps"
	# Write-Host "$fn_name() QuietFlg = $QuietFlg"

	#Set-StrictMode  -Version 3
	#Start-Transcript  # default destination is C:\Users\{user}l\Documents

	# PARSE COMMANDLINE ARGS
	if ($CamProps) {
		$cam_props_ary = $CamProps -Split ','
		#msg.exe * /TIME:5 "camProps... $cam_props_ary"
	} else {
		$cam_props_ary = @()
	}
	#Write-Host "cam_props_ary = $cam_props_ary"

	# PERFORM API CALLS
	$jsonCmd = @{
		cmd = 'camlist'
	}
	# Returns a powershell PSCustomObject
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd `
		-QuietFlg $QuietFlg `
		-FromConsoleFlg $FromConsoleFlg

	# For debugging
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj type = " + $json_response_obj.GetType())
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj = " + ($json_response_obj | ConvertTo-Json))
	#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response... $json_response_obj"

	# Abort if GET failed
	$fail_msg = "ERR/JSON request failed"
	if ($json_response_obj -match 'ERR/') {
		if ($fromConsole) {
			# For debugging
			#Write-Host "$fn_name() BIJsonCall() json_response_obj = $json_response_obj`n"
		} else {
			#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj = $json_response_obj"
		}
		return $fail_msg
	}
	if ($json_response_obj.result -eq 'fail') {
		$x  = $json_response_obj.result
		if ($fromConsole) {
			# For debugging
			#Write-Host "$fn_name() BIJsonCall() json_response_obj.result = $x`n"
		} else {
			#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj.result = $x"
		}
		return $fail_msg
	}

	# Extract data from the JSON command response
	$data_obj = $json_response_obj.data

	# For error testing
	#$data_obj = $null

	# Extract property value
	# ... Note: null/blank $data_obj.ram observed on occasion
	if ($data_obj) {
		$data_cnt = $data_obj.Count
		
		# For debugging
		#Write-Host ("data_obj cnt = " + $data_cnt)
		#Write-Host ("data_obj[0] = `n" + $data_obj[0])
		#Write-Host ("data_obj[" + ($data_cnt - 1) +"] = `n" + $data_obj[$data_cnt-1])

		$rtn_obj = [ordered]@{}

		for ($i=0; $i -lt $data_obj.Length; $i++) { 
			$item_obj = $data_obj[$i]
			$option_disp = $item_obj.optionDisplay;
			$option_val  = $item_obj.optionValue;
			
			# For debugging
			#Write-Host "$i : option_val = $option_val, option_disp = $option_disp"
			#Write-Host ("    option_val starts w/ @ = " + $option_val.StartsWith("@"))
			#Write-Host ("    option_disp starts w/ + = " + $option_disp.StartsWith("+"))
			
			if ($option_val -ne "Index" -and !$option_disp.StartsWith("+") -and !$option_val.StartsWith("@")) {
				$new_obj = [ordered]@{}
				
				if ($cam_props_ary.Length -gt 0) {
					# GET REQUESTED PROPERTIES
					
					foreach ($cam_prop in $cam_props_ary) {
						#Write-Host ("cam_prop = $cam_prop, val = " + $item_obj.$cam_prop)
						if ($item_obj.$cam_prop -or $item_obj.$cam_prop -eq 0) {
							$x = $item_obj.$cam_prop
							$new_obj.Add($cam_prop, $x)
							#$new_obj.$cam_prop = $x
						} else {
							$new_obj.Add($cam_prop, $null)
							#$new_obj.$cam_prop = $null
						}
					}
				} else {
					# GET ALL PROPERTIES
					
					# Get all property names in the JSON response (a PowerShell PSCustomObject)
					# ... NOTE: Extracting 'keys' from the PSCustomObject was bit tricky
					#           REF: https://stackoverflow.com/a/18780001/13499984
					$item_obj_keys = @()
					foreach ($item in $item_obj) {
						$item_obj_keys += $($item | Get-Member -MemberType *Property).Name
					}

					# For debugging
					#Write-Host ("item_obj_keys keys = " + $item_obj_keys)
					
					foreach ($cam_prop in $item_obj_keys) {
						#Write-Host ("cam_prop = $cam_prop, val = " + $item_obj.$cam_prop)
						if ($item_obj.$cam_prop -or $item_obj.$cam_prop -eq 0) {
							$x = $item_obj.$cam_prop
							$new_obj.Add($cam_prop, $x)
						} else {
							$new_obj.Add($cam_prop, $null)
						}
					}
				}
				$rtn_obj[$option_val] = $new_obj
			}
		}

		# @@@ DEV - Retain for testing
		#Write-Host ("rtn_obj cnt = " + $rtn_obj.count)
		#Write-Host "rtn_obj[DW1] = "$rtn_obj.DW1
		#Write-Host "rtn_obj[DW1]['BPS'] = "$rtn_obj['DW1']['BPS']
		#Write-Host "rtn_obj keys = "$rtn_obj.Keys
		
	} else {
		#Write-Host "sorry - JSON query returned null/blank`n"
	}

	return $rtn_obj
}

function Convert-HexToVersion {
# PURPOSE: To convert a hexadecimal number like 5090704 or 5091014 to a string like 5.9.7.4 or 5.9.10.14
# Source: ChatGPT (modified by jaydeel)

    param (
        [Parameter(Mandatory=$true)]
        [String]$Hex
    )
 
	# For debugging
	# Write-Host "Entered $fn_name()..."
	# Write-Host "$fn_name() Hex = '$Hex'"
	
    # Ensure the hex number is a string (if given as a number)
    $HexString = $Hex.ToString()

	# Ensure the hex string has an even number of characters by prepending a '0' if needed
    if ($HexString.Length % 2 -ne 0) {
        $HexString = "0" + $HexString
    } 

    # Split the hex string into pairs of characters
	# ... inserts a period after every two characters in the string
    $pairs = for ($i = 0; $i -lt $hex.Length; $i += 2) {
        $HexString.Substring($i, 2)
    }
	
	# For debugging
	# Write-Host ("$fn_name() pairs = " + ($pairs | ConvertTo-Json))

    # Convert each pair to its decimal equivalent and join them with '.'
    return ($pairs | ForEach-Object { [convert]::ToInt32($_, 16) }) -join '.'
}

<# function GetBiVersionOLD() {
# PURPOSE:  Get Blue Iris application's current software version
# ARGS:     None
# RETURNS:  (string) like "5.1.2.9"
# REQUIRES: BIJsonCall()
# GLOBALS:  Uses global vars: $FromConsole
#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered $fn_name()..."
	# Write-Host "$fn_name() FromConsole = $script:FromConsole"
	
	$rtn_str = ""  # Assume failure

	# Start timer
	# $timer = [Diagnostics.Stopwatch]::StartNew()

	# GET CURRENT VERSION
	$jsonCmd = @{
		cmd = 'login'
	}
	# Returns a powershell PSCustomObject
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd

	# For debugging
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj type = " + $json_response_obj.GetType())
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj = " + ($json_response_obj | ConvertTo-Json))

	# Abort if GET failed
	if ($json_response_obj -match 'ERR/') {
		if ($script:FromConsole) {
			Write-Host "$fn_name() ERR/JSON request failed"
			Write-Host "$fn_name() BIJsonCall() json_response_obj = $json_response_obj`n"
		}
		return $rtn_str
	}
	if ($json_response_obj.result -eq 'fail') {
		$x  = $json_response_obj.result
		if ($script:FromConsole) {
			Write-Host "$fn_name() ERR/JSON request failed ('result' = fail)"
			Write-Host "$fn_name() BIJsonCall() json_response_obj.result = $x`n"
		}
		return $rtn_str
	}

	# Extract data from the JSON command response
	if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {

		$data_obj = $json_response_obj.data

		# Get the current version
		$rtn_str = $data_obj.version
		
	}

	# Stop timer
	# $timer.stop()
	# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
	# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
	# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

	# Write-Host "$fn_name() = '$rtn_str'"

	return $rtn_str
} #>

function GetBiVersion() {
# PURPOSE:  Get Blue Iris application's current software version
# ARGS:     None
# RETURNS:  (string) like "5.1.2.9"
# REQUIRES: Convert-HexToVersion()
# GLOBALS:  Uses global vars: $FromConsole
#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
#
# Version: 1
# Created by jaydeel at ipcamtalk.com
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered $fn_name()..."
	# Write-Host "$fn_name() FromConsole = $script:FromConsole"

	# Start timer
	# $timer = [Diagnostics.Stopwatch]::StartNew()
	
	$rtn_str = ""  # Assume failure

	# GET CURRENT VERSION
	
	# Get current setting's value
	$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris'
	$key_name = 'version'
	$key_val_cur = `
		RegGetValue `
		-KeyPath $key_path `
		-KeyName $key_name

	# Abort if GET failed
	if ($key_val_cur -match 'ERR/') {
		if ($script:FromConsole) {
			Write-Host "$fn_name() ERR/REG request failed"
			Write-Host "$fn_name() RegGetVall() key_val_cur = $key_val_cur"
		}
		return $rtn_str
	}

	# Get the current version string
	$bi_version_hex = $key_val_cur.ToString("X")  # like hex 5090704
	$rtn_str = Convert-HexToVersion -Hex $bi_version_hex  # like '5.9.7.4'

	# Stop timer
	# $timer.stop()
	# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
	# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
	# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

	return $rtn_str
}

<# function IsBiVersionOLD() {
# PURPOSE:   To confirm if the CURRENT Blue Iris version exceeds a specified UPDATE version
# ARGUMENTS: -CompareToVersion = [REQUIRED] (string) compare version, '5.7.2.3'
#            -CurrentVersion   = [OPTIONAL] (string) current version
#            Note: supports version strings 'X.000.000.000' to 'X.999.999.999'
#            Note: Using arg -CurrentVersion improves performance if function is called many times via a loop
# REQUIRES:  function GetBiVersionOLD()
# 
# Version: 2 (9/20/2024)
# Created by jaydeel at ipcamtalk.com
	
	param (
		[String]$CompareToVersion,
		[String]$CurrentVersion 
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# Write-Host "Entered $fn_name()..."
	# Write-Host "$fn_name() CompareToVersion = $CompareToVersion"
	# Write-Host "$fn_name() CurrentVersion = $CurrentVersion"
	
	$rtn_flg = $false  # Assume failure

	if (!$CurrentVersion) {

		# GET CURRENT VERSION
		$bi_version = GetBiVersionOLD

		# Abort if GET failed
		if ($bi_version -match 'ERR/') {
			if ($script:FromConsoleFlg) {
				Write-Host "$fn_name() ERR/JSON request failed"
			}
			return $false
		}
			
	} else {
		$bi_version = $CurrentVersion
	}
	
	# Make comparator integer arrays
	[int[]]$my_ver_ary = $CompareToVersion.Split('.')
	[int[]]$bi_ver_ary = $bi_version.Split('.')

	# Make 'prime.subprime' version comparators, like 5.8 vs 5.7
	$my_ver_prime_dot = $my_ver_ary[0] + $my_ver_ary[1]/1000
	$bi_ver_prime_dot = $bi_ver_ary[0] + $bi_ver_ary[1]/1000

	# For debugging
	#Write-Host "bi_version = $bi_version"
	#Write-Host ("my_ver_ary = " + ($my_ver_ary | ConvertTo-Json))
	#Write-Host ("bi_version_ary = " + ($bi_ver_ary | ConvertTo-Json))
	#Write-Host "my_ver_prime_dot = $my_ver_prime_dot"
	#Write-Host "bi_ver_prime_dot = $bi_ver_prime_dot"
	
	if ($my_ver_prime_dot -eq $bi_ver_prime_dot) {

		# Make sub-subprime version comparators, 5.001 for 5.8.5.1, 999.020 for 5.8.999.20
				
		$mult1 = [Math]::Pow(10, 0)
		$mult2 = [Math]::Pow(10, -3)

		if ($my_ver_ary[2] -eq 0 -and $my_ver_ary[3] -eq 0) {
			$my_ver_sub = 0.000
		} else {
			$my_ver_sub = $my_ver_ary[2]*$mult1 + $my_ver_ary[3]*$mult2
		}

		if ($bi_ver_ary[2] -eq 0 -and $bi_ver_ary[3] -eq 0) {
			$bi_ver_sub = 0.000
		} else {
			$bi_ver_sub = $bi_ver_ary[2]*$mult1 + $bi_ver_ary[3]*$mult2
		}

		# For debugging
		#Write-Host ("my_ver_sub = " + ("{0:n3}" -f $my_ver_sub))
		#Write-Host ("bi_ver_sub = " + ("{0:n3}" -f $bi_ver_sub))

		$rtn_flg = ($my_ver_sub -ge $bi_ver_sub) 

	} else {
		$rtn_flg = ($my_ver_prime_dot -ge $bi_ver_prime_dot)
	}

	return $rtn_flg
} #>

function IsBiVersion() {
# PURPOSE:   To confirm if the CURRENT Blue Iris version exceeds a specified UPDATE version
# ARGUMENTS: -CompareToVersion = [REQUIRED] (string) compare version, '5.7.2.3'
#            -CurrentVersion   = [OPTIONAL] (string) current version
#            Note: supports version strings 'X.000.000.000' to 'X.999.999.999'
#            Note: Using arg -CurrentVersion improves performance if function is called many times via a loop
# RETURNS:   boolean; $true current BI version exceeds specified version
# REQUIRES:  function GetBiVersion()
#
# VERSION:   2
# By jaydeel at ipcamtalk.com
	
	param (
		[String]$CompareToVersion,
		[String]$CurrentVersion
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# Write-Host "Entered $fn_name()..."
	# Write-Host "$fn_name() CompareToVersion = $CompareToVersion"
	# Write-Host "$fn_name() CurrentVersion = $CurrentVersion"
	# Write-Host "$fn_name() FromConsole = $script:FromConsole"

	# Start timer
	# $timer = [Diagnostics.Stopwatch]::StartNew()
	
	$rtn_flg = $false  # Assume failure

	if (!$CurrentVersion) {
		
		# GET CURRENT VERSION
		$bi_version = GetBiVersion

		# Abort if GET failed
		if ($bi_version -match 'ERR/') {
			if ($script:FromConsole) {
				Write-Host "$fn_name() ERR/REG request failed"
			}
			return $false
		}
		
	} else {
		$bi_version = $CurrentVersion
	}
	
	# Make comparator integer arrays
	[int[]]$my_ver_ary = $CompareToVersion.Split('.')
	[int[]]$bi_ver_ary = $bi_version.Split('.')

	# Make 'prime.subprime' version comparators, like 5.8 vs 5.7
	$my_ver_prime_dot = $my_ver_ary[0] + $my_ver_ary[1]/1000
	$bi_ver_prime_dot = $bi_ver_ary[0] + $bi_ver_ary[1]/1000

	# For debugging
	# Write-Host "bi_version_hex (registry) = $bi_version_hex"
	# Write-Host "bi_version = $bi_version"
	# Write-Host ("my_ver_ary = " + ($my_ver_ary | ConvertTo-Json))
	# Write-Host ("bi_version_ary = " + ($bi_ver_ary | ConvertTo-Json))
	# Write-Host "my_ver_prime_dot = $my_ver_prime_dot"
	# Write-Host "bi_ver_prime_dot = $bi_ver_prime_dot"
	
	if ($my_ver_prime_dot -eq $bi_ver_prime_dot) {

		# Make sub-subprime version comparators, 5.001 for 5.8.5.1, 999.020 for 5.8.999.20
				
		$mult1 = [Math]::Pow(10, 0)
		$mult2 = [Math]::Pow(10, -3)

		if ($my_ver_ary[2] -eq 0 -and $my_ver_ary[3] -eq 0) {
			$my_ver_sub = 0.000
		} else {
			$my_ver_sub = $my_ver_ary[2]*$mult1 + $my_ver_ary[3]*$mult2
		}

		if ($bi_ver_ary[2] -eq 0 -and $bi_ver_ary[3] -eq 0) {
			$bi_ver_sub = 0.000
		} else {
			$bi_ver_sub = $bi_ver_ary[2]*$mult1 + $bi_ver_ary[3]*$mult2
		}

		# For debugging
		# Write-Host ("my_ver_sub = " + ("{0:n3}" -f $my_ver_sub))
		# Write-Host ("bi_ver_sub = " + ("{0:n3}" -f $bi_ver_sub))

		$rtn_flg = ($bi_ver_sub -ge $my_ver_sub) 


	} else {
		$rtn_flg = ($bi_ver_prime_dot -ge $my_ver_prime_dot)
	}

	# Stop timer
	# $timer.stop()
	# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
	# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
	# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

	return $rtn_flg
}

function FindBytes() {
# PURPOSE:  Find all starting indices in $Bytes where $Search was found
# ARGS:     see param lines below
# RETURNS:  (array) integers, or empty array @()
# Source:   https://stackoverflow.com/a/62511302/13499984
# Version   4 (5/22/2024) Added arg -Reverse; 3 (5/5/2024) added comments to clarify usage

	param (
		[Byte[]]$Bytes,    # [REQUIRED]
		[Byte[]]$Search,   # [REQUIRED]
		[Int]$StartIndex,  # start searching array $Bytes at index X
		[Int]$StopIndex,   # stop searching array $Bytes at index Y
		[Switch]$Reverse,  # search backwards
		[Switch]$All       # returns all matches, otherwise returns FIRST match only
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "entered function $fn_name()..."
	# Write-Host ("$fn_name() Bytes = " + ($Bytes -Join ' '))
	# Write-Host ("$fn_name() Search = " + ($Search -Join ' '))
	# Write-Host "$fn_name() All = $All"
	# Write-Host "$fn_name() StartIndex = $StartIndex"
	# Write-Host "$fn_name() Reverse = $Reverse"
	
	[Int[]]$rtn_ary = @()
	
	if ($Reverse -and ($StopIndex -gt $StartIndex)) {
		if ($script:FromConsole) {
			Write-Warning "$fn_name() param -Reverse requires -StopIndex <= -StartIndex"
		}
		return $rtn_ary
	}
		
    if (!$Reverse) {
		
		if ($StopIndex) {
			$start_index = $StartIndex
			$stop_index  = $StopIndex
		} else {
			$start_index = $StartIndex
			$stop_index  = $Bytes.Length
		}

		# For debugging
		# Write-Host "$fn_name() start_index = $start_index"
		# Write-Host "$fn_name() stop_index = $stop_index"
	
		#for ($idx = $StartIndex; $idx -le $Bytes.Length - $Search.Length ; $idx++) {
		for ($idx = $start_index; $idx -le $stop_index - $Search.Length ; $idx++) {
			for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
			if ($i -ge $Search.Length) { 
				$rtn_ary += $idx
				if (!$All) { 
					return $rtn_ary
				}
			}
		}
		
	} else {
		
		if ($StopIndex) {
			$start_index = $StartIndex
			$stop_index  = $StopIndex
		} else {
			$start_index = $StartIndex
			$stop_index  = 0
		}

		# For debugging
		# Write-Host "$fn_name() start_index = $start_index"
		# Write-Host "$fn_name() stop_index = $stop_index"
		
		#for ($idx = $StartIndex - $Search.Length ; $idx -ge 0 ; $idx--) {
		for ($idx = $start_index - $Search.Length ; $idx -ge $stop_index ; $idx--) {
			for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
			if ($i -ge $Search.Length) { 
				$rtn_ary += $idx
				if (!$All) {
					return $rtn_ary
				}
			}
		}
		
	}
	
	# KEY POINT: By default, Powershell functions will return $null if attempting to return an empty array.
	# e.g., return @()
	# To avoid this, the options are:
	# 1) return ,$rtn_ary  ... where the comma is the Powershell 'array construction operator'
	# 2) write-output $arr -NoEnumerate 
	# REF: https://stackoverflow.com/a/18477004

	return ,$rtn_ary
}

function FindBytes_OLD() {
# PURPOSE:  Find all starting indices in $Bytes where $Search was found
# ARGS:     see param lines below
# RETURNS:  (array) integers, or empty array @()
# Source:   https://stackoverflow.com/a/62511302/13499984
# Version   4 (5/22/2024) Added arg -Reverse; 3 (5/5/2024) added comments to clarify usage

	param (
		[Byte[]]$Bytes,    # [REQUIRED]
		[Byte[]]$Search,   # [REQUIRED]
		[Int]$StartIndex,  # start searching array $Bytes at index N (rare use case)
		[Switch]$Reverse,  # search backwards
		[Switch]$All       # returns all matches, otherwise returns FIRST match only
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "entered function $fn_name()..."
	# Write-Host ("$fn_name() Bytes = " + ($Bytes -Join ' '))
	# Write-Host ("$fn_name() Search = " + ($Search -Join ' '))
	# Write-Host "$fn_name() All = $All"
	# Write-Host "$fn_name() StartIndex = $StartIndex"
	# Write-Host "$fn_name() Reverse = $Reverse"
	
	[Int[]]$rtn_ary = @()
	
    if (!$Reverse) {
		
		for ($idx = $StartIndex; $idx -le $Bytes.Length - $Search.Length ; $idx++) {
			for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
			if ($i -ge $Search.Length) { 
				$rtn_ary += $idx
				if (!$All) { 
					return $rtn_ary
				}
			}
		}
		
	} else {
		
		for ($idx = $StartIndex - $Search.Length ; $idx -ge 0 ; $idx--) {
			for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
			if ($i -ge $Search.Length) { 
				$rtn_ary += $idx
				if (!$All) {
					return $rtn_ary
				}
			}
		}
		
	}
	
	# KEY POINT: By default, Powershell functions will return $null if attempting to return an empty array.
	# e.g., return @()
	# To avoid this, the options are:
	# 1) return ,$rtn_ary  ... where the comma is the Powershell 'array construction operator'
	# 2) write-output $arr -NoEnumerate 
	# REF: https://stackoverflow.com/a/18477004

	return ,$rtn_ary
}

function RepairDatFile() {
# PURPOSE: Try to repair a DAT file with unexpected structure [unbalanced JPG beg/end markers]
# ARGS:    see param()
# NOTE:    Repair is attempted only when the count of 'end markers' <> the count of 'beg markers'
# RETURNS  [array] @([Boolean]rtn_flg, [array]SOI_mrks, [array]EOI_mrks)

	param (
		[byte[]]$Bytes,
		[array]$BegOfImageMarks,
		[array]$EndOfImageMarks
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	#Write-Host "Entered $fn_name()..."
	#Write-Host ("$fn_name() BegOfImageMarks type = " + $BegOfImageMarks.GetType())
	#Write-Host ("$fn_name() BegOfImageMarks cnt = " + $BegOfImageMarks.Count)
	#Write-Host "$fn_name() BegOfImageMarks = $BegOfImageMarks"
	#Write-Host ("$fn_name() EndOfImageMarks type = " + $EndOfImageMarks.GetType())
	#Write-Host ("$fn_name() EndOfImageMarks cnt = " + $EndOfImageMarks.Count)
	#Write-Host "$fn_name() EndOfImageMarks = $EndOfImageMarks"
	
	# Initialize vars
	$rtn_flg = $true  # Assume success
	$SOI_mrks_cnt = $BegOfImageMarks.Count
	$EOI_mrks_cnt = $EndOfImageMarks.Count		

	# Must convert arrays to ArrayLists to use method 'RemoveAt()' (later)
	$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$BegOfImageMarks
	$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EndOfImageMarks

	# For debugging
	#Write-Host ("$fn_name() SOI_mrks_lst type = " + $SOI_mrks_lst.GetType())
	#Write-Host ("$fn_name() SOI_mrks_lst cnt = " + $SOI_mrks_lst.Count)
	#Write-Host "$fn_name() SOI_mrks_lst = $SOI_mrks_lst"
	#Write-Host ("$fn_name() EOI_mrks_lst type = " + $EOI_mrks_lst.GetType())
	#Write-Host ("$fn_name() EOI_mrks_lst cnt = " + $EOI_mrks_lst.Count)
	
	if ($SOI_mrks_cnt -gt $EOI_mrks_cnt) {

		for ($i=0; $i -lt $EOI_mrks_cnt; $i++) {
			$indices_delta = [Int]$EndOfImageMarks[$i] - [Int]$BegOfImageMarks[$i]
			
			# For debugging
			#Write-Host ("- end index = [$i] " + $EndOfImageMarks[$i])
			#Write-Host "- indices_delta = $indices_delta"

			if ($indices_delta -lt 0) {
				
				$SOI_mrks_lst_new = $SOI_mrks_lst
				$SOI_mrks_lst_new.RemoveAt($i)

				# For debugging
				#Write-Host "starting recursive call..."
				#Write-Host "- SOI_mrks_lst_new = $SOI_mrks_lst_new"

				# Recursive call
				$arr1 = [array]$SOI_mrks_lst_new
				$arr2 = [array]$EOI_mrks_lst
				$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
					RepairDatFile `
					-Bytes $bytes `
					-BegOfImageMarks $arr1 `
					-EndOfImageMarks $arr2

				# For debugging
				#Write-Host "recursive result_flg = $result_flg"

				if ($result_flg) {
					$rtn_flg = $true
					$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$SOI_mrk_idxs_ary
					$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EOI_mrk_idxs_ary
					break
				} else {
					$rtn_flg = $false
				}				
			}
		}					
	}
	
	if ($EOI_mrks_cnt -ge $SOI_mrks_cnt) {

		for ($i=0; $i -lt $SOI_mrks_cnt; $i++) {
			$indices_delta = [Int]$EndOfImageMarks[$i] - [Int]$BegOfImageMarks[$i]
			
			# For debugging
			#Write-Host ("- beg index = [$i] " + $BegOfImageMarks[$i])
			#Write-Host "- indices_delta = $indices_delta"
			
			if ($indices_delta -lt 0) {
				
				$EOI_mrks_lst_new = $EOI_mrks_lst
				$EOI_mrks_lst_new.RemoveAt($i)

				# For debugging
				#Write-Host "starting recursive call..."
				#Write-Host "- EOI_mrks_lst_new = $EOI_mrks_lst_new"

				# Recursive call
				$arr1 = [array]$SOI_mrks_lst
				$arr2 = [array]$EOI_mrks_lst_new
				$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
					RepairDatFile `
					-Bytes $bytes `
					-BegOfImageMarks $arr1 `
					-EndOfImageMarks $arr2
					
				# For debugging
				#Write-Host "recursive result_flg = $result_flg"

				if ($result_flg) {
					$rtn_flg = $true
					$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$SOI_mrk_idxs_ary
					$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EOI_mrk_idxs_ary
					break
				} else {
					$rtn_flg = $false
				}				
			}
		}
	}

	return @($rtn_flg, [array]$SOI_mrks_lst, [array]$EOI_mrks_lst)

}

function IsJpegFileValid {
# PURPOSE: To confirm that a JPG file is a valid JPG image
# NOTES:
#    Using 2 methods:
#    1. Test image type
#       REFS:  http://learningpcs.blogspot.com/2011/07/powershell-validate-jpeg-files.html
#              https://www.powershellgallery.com/packages/PScribo/0.9.1/Content/Src%5CPrivate%5CGet-ImageMimeType.ps1
#    2. Per the JPEG Interchange Format... any JPEG file or stream that is written according to the 
#       JPEG Interchange Format means that they must start with a SOI (Start-of-Image) marker (the two bytes 0xFF, 0xD8) 
#       and end with the EOI (End-of-Image) marker (two bytes, 0xFF, 0xD9). 
#       These two markers will not be found anywhere else in a JPEG file/stream.
#       REFS: https://stackoverflow.com/a/46809226
#             https://www.codeproject.com/Questions/421855/Verifying-a-file-is-JPEG
# DISCUSSION:
#    Originally I used only Method #1 in this function. However, I learned later that Method #1 would still return valid
#    if the image bytes did not terminate with the EOI (End-of-Image) marker (0xFF, 0xD9). These JPG files still opened 
#    just fine in every windows viewer I tried (Photos, Paint, Paint3D, PhotoScape X, ACDSee...), HOWEVER, if they were 
#    opened in a web browser (tested Chrome, Edge) the images had a white band across the bottom. 
#    Additionally, web apps like https://app.roboflow.com/ would not import the JPGs.
#    So clearly, Method 1 is not sufficient. For redundancy, I've opted to use both methods. 

	param(
		[String]$FilePath
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#msg.exe $script:server_username /TIME:15 "entered function $fn_name()..."
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() FilePath = '$FilePath'"
	#msg.exe $script:server_username /TIME:30 "$fn_name() FilePath = '$FilePath'"
	
	# Initialize flags
	$method1_flg = $false  # Validate image file Type
	$method2_flg = $false  # Validate bytes per the 'JPEG Interchange Format' (see Notes above)
	
	# ===========================
	# METHOD 1 - Validate image file Type

	# WARNING: The following code line IS ESSENTIAL when executing from Task Scheduler (or Blue Iris)
	#          (however works just fine from Powershell ISE)
	#          REF: https://tinyurl.com/2p929nhz
	add-type -AssemblyName System.Drawing
	
	try {
		$img = [System.Drawing.Image]::FromFile("$FilePath");
		$method1_flg = $img.RawFormat.Equals([System.Drawing.Imaging.ImageFormat]::Jpeg)
		$img.Dispose()  # Release resources

	} catch [OutOfMemoryException] {
		$method1_flg = $false
	}

	# ===========================
	# METHOD 2 - Validate bytes per the 'JPEG Interchange Format' (see Notes above)
	
	# Import the file to a byte[] array
	$ImageBytes = [System.IO.File]::ReadAllBytes($FilePath)
	
	# Capture the start and end byte pairs.
	#$bytes_rng1 = $ImageBytes[0],$ImageBytes[1]
	#$bytes_rng2 = $ImageBytes[$ImageBytes.Count-2],$ImageBytes[$ImageBytes.Count-1]
	$bytes_rng1 = $ImageBytes[0..1]
	$bytes_rng2 = $ImageBytes[($ImageBytes.Count-2)..($ImageBytes.Count-1)]

	# Define the required JPG start/end markers
	$markers1 = @(255,216) # hexArray @("FF", "D8")
	$markers2 = @(255,217) # hexArray @("FF", "D9")
	<# $hexArray = "FF", "D8"
	$markers1 = @()
	foreach ($hex in $hexArray) {
		$markers1 += [Int32]"0x$hex"
	}
	$hexArray = "FF", "D9"
	$markers2 = @()
	foreach ($hex in $hexArray) {
		$markers2 += [Int32]"0x$hex"
	} #>

	# Validate the start and end byte pairs.
	# ... Note: Method Compare-Object returns null if the arrays are the identical
	$x1 = Compare-Object $bytes_rng1 $markers1
	$x2 = Compare-Object $bytes_rng2 $markers2
	if ($null -eq $x1 -and $null -eq $x2) {
		$method2_flg = $true
	} else {
		$method2_flg = $false
	}
	
	# For testing - Simulate Method 2 (Validate bytes per the 'JPEG Interchange Format')
	#$method2_flg = $true

	# For debugging
	#Write-Host "markers1 = $markers1, markers2 = $markers2"
	#Write-Host "bytes_rng1 = $bytes_rng1, bytes_rng2 = $bytes_rng2"
	#Write-Host "method1_flg = $method1_flg, method2_flg = $method2_flg"
	#msg.exe $script:server_username /TIME:15 "$fn_name() method1_flg = $method1_flg, method2_flg = $method2_flg"
	
	if ($method1_flg -and $method2_flg) {
		return $true
	} else {
		return $false
	}
}

<# function IsJsonStringValid() {
# PURPOSE:   To test for a valid JSON string
# PARAMS:    -JsonStr = [string] to be tested
# RETURNS:   (boolean)
# REFS:      https://stackoverflow.com/a/17170800/13499984
#            https://jsonlint.com/  <-- Handy on-line JSON string test utility

	param(
		[string]$JsonStr
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host ("$fn_name() JsonStr type = " + $JsonStr.GetType())
	#Write-Host "$fn_name() JsonStr  = $JsonStr"
	#Write-Host ("JsonStr = " + ($JsonStr | ConvertTo-Json))

	# Test for non-alphanumeric characters
	#If (!($JsonStr -match '^[a-z0-9.-]{1,10}$')) {
		#return $false
	#}

	# For debugging (this was how I found the null terminator issue that caused grief in PS5)
	#$json_filespec = "$script:dest_path\temp\tmp.json"
	#Write-Host "json_filespec = '$json_filespec'"
	#$JsonStr | Out-File -FilePath "$json_filespec" -Encoding UTF8
	#$JsonStr = Get-Content -Raw -Path $json_filespec

	$PsVersion = $PSVersionTable.PSVersion.Major

	# Validity test
	# ... Note: I unsuccessfully tried PS6 '$JsonStr | Test-Json' (see commented code below); 
	#           it output a non-fatal error for json strings with binary characters
	if ($JsonStr) {
		if ($PsVersion -ge 6) {
			try {
				#$json_test = ($JsonStr | Test-Json)
				$powershellRepresentation = ConvertFrom-Json $JsonStr -ErrorAction Stop
			}
			catch {
				Write-Warning "Cannot parse JSON string"
				return $false
			}

			if (!$json_test) {
				return $false
			}
			
		} else {
			try {
				$powershellRepresentation = ConvertFrom-Json $JsonStr -ErrorAction Stop
			} catch {
				Write-Warning "Cannot parse JSON string"
				return $false
			}
		}
		try {
			$powershellRepresentation = ConvertFrom-Json $JsonStr -ErrorAction Stop
		} catch {
			#Write-Warning "Cannot parse JSON string"
			return $false
		}
	} else {
		return $false
	}
	
	return $true
} #>

function ExtractDatImages() {
# PURPOSE:  Extract all JPG files from a Blue Iris *.DAT file, and save them to disk
#           with filenames like 'img01.jpg', 'img02.jpg', ...
# ARGS:     see param lines below
# RETURNS:  1. (hashtable) @{ source=filespec ; datjpgs=N ; data=@(jpgfilelist) }
#           2. or (string) if issue/fail (NOTE: string must begin with 'ISSUE:', or 'FAIL:')
# GLOBALS:  Uses global vars: $BiVersion, $FromConsole, $Quiet
#           $script:BiVersion   = (string) Blue Iris software current version
#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
#           $script:Quiet       = (boolean) $true suppresses verbose output to the console
#
# Version   3 (8/29/2024) faster getting JPG image beg/end markers; replaced *Flg args w/ global vars
# Coded by: jaydeel @ ipcamtalk.com
#
# hastable return object is like...
# {
#   "source": "D:\\BlueIris\\Alerts\\DW1.20240306_080000.1950096.5-0.dat",
#   "jpgstotal": 13,
#   "jpgsextracted": 
#      [
#         "D:\\BI_gif_tool\\DW1\\img05.jpg",
#         "D:\\BI_gif_tool\\DW1\\img04.jpg",
#         "D:\\BI_gif_tool\\DW1\\img03.jpg",
#         "D:\\BI_gif_tool\\DW1\\img02.jpg",
#         "D:\\BI_gif_tool\\DW1\\img01.jpg"
#      ]
# }
# ATTRIBUTIONS:
#    Per Ken @blueirissupport... "The DAT file is custom serialized format. 
#    To extract JPGs: search the byte content for blocks that start with 'FF D8 FF E0' and go through 'FF D9'. 
#    To extract the JSON string: after the bitmap there are 2 QWORDs (8 bytes each), then the STRING for the JSON 
#    (comprised of a DWORD length then the characters)."
#
#    Per the JPEG Interchange Format... any JPEG file or stream is written according to the JPEG Interchange 
#    Format must start with a SOI (Start-of-Image) marker (the two bytes 0xFF, 0xD8) and end with the 
#    EOI (End-of-Image) marker (two bytes, 0xFF, 0xD9). 
#    These two markers will not be found anywhere else in a JPEG file/stream.
#    REF: https://stackoverflow.com/a/46809226
#
# KEY POINT: 
#    JPG images in a BI *.DAT file are stored in reverse chronological order
	
    param(
		[String]$AlertFile,          # Alert file, like 'GT.20230610_070000.41437.17-0.jpg' (via JSON 'clipstats')
		[String]$AlertPath,          # folder path for Blue Iris 'DAT' files (full path) (via user-setting $bi_alerts_folder_path)
		[Boolean]$LatestFlg,         # use DAT file from the most recent alert (that has a DAT file)
		[String]$DestPath,           # folder for the extracted JPG files
		[Int]$DatReqdImages,         # number of images that must be found before extracting any images	
		[Int]$ImagesCnt,             # number of images to be extracted from the BVR file, if the DAT file determined to be 'bad'	
		[Boolean]$ImagesCntUsedFlg,  # true indicates that the param -ImagesCnt was used on the command line
		[Int]$ImagesTrimBegPct,      # percent of images to skip at front
		[Int]$ImagesTrimEndPct      # percent of images to skip at end
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	$PsVersion = $PSVersionTable.PSVersion.Major

	# For debugging
	# Write-Host "Entered function: $fn_name()"	
	# Write-Host "$fn_name() AlertFile = '$AlertFile'"
	# Write-Host "$fn_name() AlertPath = '$AlertPath'"
	# Write-Host "$fn_name() LatestFlg = $LatestFlg"
	# Write-Host "$fn_name() DestPath = '$DestPath'"
	# Write-Host ("$fn_name() DatReqdImages type = " + $DatReqdImages.GetType())
	# Write-Host "$fn_name() DatReqdImages = $DatReqdImages"
	# Write-Host "$fn_name() ImagesCntUsedFlg = $ImagesCntUsedFlg"
	# Write-Host "$fn_name() ImagesTrimBegPct = $ImagesTrimBegPct"
	# Write-Host "$fn_name() ImagesTrimEndPct = $ImagesTrimEndPct"
	# Write-Host ("$fn_name() BiVersion = " + $script:BiVersion)
	# Write-Host ("$fn_name() Quiet = " + $script:Quiet)
	# Write-Host ("$fn_name() FromConsole = " + $script:FromConsole)
	# Write-Host "$fn_name() PsVersion = $PsVersion"	
	
	# HANDLE ARGS
	$DestPath = $DestPath.TrimEnd('\')
	
	# Initialize
	$dat_file_found_flg = $false  # Default
	$rtn_obj = [ordered]@{}
	# ... extract camname
	$str = Split-Path $AlertFile -leaf
	$arr = $str.Split(".")
	$cam = $arr[0]	
	
	# For debugging
	#Write-Host "$fn_name() str = '$str'"
	#Write-Host ("$fn_name() arr = " + ($arr | ConvertTo-Json))
	#Write-Host "$fn_name() cam = '$cam'"
	
	# TEST FOR DESTINATION FOLDER FOR JPGs
	$jpg_path =  $DestPath + "\" + $cam
	if (!(Test-Path -LiteralPath "$jpg_path")) {
		New-Item -ItemType Directory -Force -Path $jpg_path
	}
	
	# For debugging
	# Write-Host "$fn_name() jpg_path = '$jpg_path'"

	# DELETE EXISTING *.JPG FILES ASYNCHRONOUSLY
	get-childitem -path ($jpg_path + "\img*.jpg") | remove-item -force
	
	# USE THE 'DAT' FILE IMPLIED BY THE DATABASE (VIA JSON CLIPSTATS COMMAND)

	$dat_file = "$AlertPath\$AlertFile".Replace('.jpg','.dat')
	if (Test-Path -LiteralPath "$dat_file") {
		$dat_file_found_flg = $true
	}
	
	# For debugging
	# Write-Host "$fn_name() dat_file_found_flg = $dat_file_found_flg"

	# PROCEED IF 'DAT' FILE FOUND
	if ($dat_file_found_flg) {
		
		if ($script:FromConsole -and !$script:Quiet) {
			Write-Host "`nAlert's AI 'DAT' File Processing Summary:"
			Write-Host "-----------------------------------------"
			Write-Host "Found DAT file: '$dat_file'"
		}
	
		# LOAD THE 'DAT' FILE
		if ($PsVersion -ge 7) {
			#$bytes = [System.IO.File]::ReadAllBytes("$dat_file")
			# Ref: https://stackoverflow.com/a/63782608/13499984
			$bytes = Get-Content -LiteralPath "$dat_file" -AsByteStream -Raw
		} else {
			$bytes = Get-Content -LiteralPath "$dat_file" -Encoding Byte -Raw
		}
		$bytes_cnt = $bytes.Count

		# For debugging
		#Write-Host ("$fn_name() bytes len = " + $bytes.Count)
		#Write-Host $bytes
		#$msg = $bytes.Count
		#msg.exe $script:server_username /TIME:60 "$fn_name() bytes len = $msg"

		# //////////////////////////////////////////////////////////////////////////////////
		
		# FIND INDICES FOR ALL IMAGE DELIMITERS

		# Start timer
		# $timer_dat_img_delim_indicies = [Diagnostics.Stopwatch]::StartNew()
	
		# The first 6 bytes of the file looks like @('05', '81', version)
		$search_ary = $bytes[0..5]

		# For debugging
		# Write-Host ("search_ary = " + ($search_ary -Join ' '))

		$img_delim_mrk_idxs = `
			FindBytes `
			-Bytes $bytes `
			-Search $search_ary `
			-All

		$img_delim_mrk_idxs_cnt = $img_delim_mrk_idxs.Count

		# For debugging
		# Write-Host "$fn_name() img_delim_mrk_idxs_cnt = $img_delim_mrk_idxs_cnt"
		# Write-Host "$fn_name() img_delim_mrk_idxs = $img_delim_mrk_idxs"

		# Stop timer
		# $timer_dat_img_delim_indicies.stop()
		# $elapsed_msec = [Math]::Round($timer_dat_img_delim_indicies.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
		# Write-Host ("timer_dat_img_delim_indicies = $elapsed_msec_str msec")

		# Determine the padding level to use for hashtable keys and safe JPG & JSON filenames
		# ... left padding is necessary in hashtable keys to ensure that the output table sorts the images in natural order
		# ... left padding is helpful in filenames to ensure that files appear in order when viewed in the Windows File Explorer
		# $pad_level = ([Math]::log10($img_delim_mrk_idxs_cnt) - ([Math]::log10($img_delim_mrk_idxs_cnt) % 1)) + 1

		# Start timer
		# $timer_dat_jpg_indicies = [Diagnostics.Stopwatch]::StartNew()

		# Build jpg byte start/end markers
		# ... JPG begin marker
		#$search_ary1 = @(255,216,255,224) # hexArray @("FF", "D8", "FF", "E0")
		$hexArray = "FF", "D8", "FF", "E0"
		$search_ary1 = @()
		foreach ($hex in $hexArray) {
			$search_ary1 += [Int32]"0x$hex"
		}
		# ... JPG end marker
		#$search_ary2 = @(255,217) # hexArray @("FF", "D9")
		$hexArray = "FF", "D9"
		$search_ary2 = @()
		foreach ($hex in $hexArray) {
			$search_ary2 += [Int32]"0x$hex"
		}

		$found_ary1 = @()
		$found_ary2 = @()

		for ($i = 0; $i -lt $img_delim_mrk_idxs_cnt ; $i++) {
			
			$img_no = $img_delim_mrk_idxs_cnt - $i 
			# $img_no_key = $img_no.ToString().PadLeft($pad_level)
			
			$img_SOI_search_beg_idx = $img_delim_mrk_idxs[$i] + 6
								
			# GET THE INDEX OF THE **NEXT** SOI MARKER
			$img_SOI_idxs = `
				FindBytes `
				-Bytes $bytes `
				-Search $search_ary1 `
				-StartIndex $img_SOI_search_beg_idx
			
			$found_ary1 += $img_SOI_idxs[0]

			# GET THE INDEX OF THE **PREVIOUS** EOI MARKER
			# ... KEY-POINT: at the last delim markers index we need to search backwards twice
			#     1st, from the delim marker
			#     2nd, from the end of array $bytes
			if ($i -gt 0) {
				if ($i -le $img_delim_mrk_idxs_cnt - 1) {
					# Write-Host "$i : made it1"			
					$img_EOI_idxs = `
						FindBytes `
						-Bytes $bytes `
						-Search $search_ary2 `
						-StartIndex $img_SOI_search_beg_idx `
						-Reverse
						
					$found_ary2 += $img_EOI_idxs[0]

					if ($i -eq $img_delim_mrk_idxs_cnt - 1) {
						# Write-Host "$i : made it2"			
						$end_idx = $bytes.Length - 1
						$img_EOI_idxs = `
							FindBytes `
							-Bytes $bytes `
							-Search $search_ary2 `
							-StartIndex $end_idx `
							-Reverse
							
						$found_ary2 += $img_EOI_idxs[0]
					}
				}
			}
		}

		$found_ary1_cnt = $found_ary1.Count
		$found_ary2_cnt = $found_ary2.Count

		# For debugging
		# Write-Host "$fn_name() jpg data starting byte pattern = $search_ary1"
		# Write-Host "$fn_name() jpg data ending byte byte pattern = $search_ary2"
		# Write-Host "$fn_name() jpg data start indices = $found_ary1"
		# Write-Host "$fn_name() jpg data end indices = $found_ary2"
		# msg.exe $script:server_username /TIME:60 "$fn_name()... `nfound_ary1_cnt = $found_ary1_cnt `nfound_ary2_cnt = $found_ary2_cnt"
		# Write-Host "$fn_name() found_ary1_cnt = $found_ary1_cnt"

		# Stop timer
		# $timer_dat_jpg_indicies.stop()
		# $elapsed_msec = [Math]::Round($timer_dat_jpg_indicies.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
		# Write-Host ("timer_dat_jpg_indicies (SOI & EOI) = $elapsed_msec_str msec")

		# //////////////////////////////////////////////////////////////////////////////////

		# Validate expected DAT images count
		if ($found_ary1_cnt -lt $DatReqdImages) {
			<# if ($script:FromConsole) {
				$bad_DAT_msg = "DAT file has fewer images than required ($found_ary1_cnt/$DatReqdImages)... "
				$bad_DAT_msg += "`n   Please navigate to Blue Iris 'Camera settings' > 'Trigger' tab > 'Artificial Intelligence' dialog"
				$bad_DAT_msg += "`n   - Confirm that setting 'To cancel' has a value like DoNotCancel"
				$bad_DAT_msg += "`n   - Confirm that setting '+ real-time images' has a value = $DatReqdImages (or " + ($DatReqdImages-1) + " if using motion-leading image)"
				$bad_DAT_msg += "`nTrying to extract images from BVR instead..."
				Write-Host $bad_DAT_msg -ForegroundColor red
			} #>
			$rtn_obj = "ISSUE: insufficient JPG images count in DAT file"
			return $rtn_obj
		}
		# Validate expected DAT structure
		# ... expect $found_ary1 like '25 113202'
		# ...        $found_ary2 like '113054 227880'
		$bad_DAT_file_flg = $false
		$i = 0
		foreach ($jpg_idx in $found_ary1) {
			if ($i -gt 0) {
				if ([Int]$jpg_idx -ge [Int]$found_ary2[$i - 1]) {
					$bad_DAT_file_flg = $true
					$bad_DAT_msg = "DAT file has unexpected structure [unbalanced JPG beg/end markers] ($dat_file)"
					break
				}
			}
		}
		
		# For debugging
		# Write-Host "$fn_name() bad_DAT_file_flg = $bad_DAT_file_flg"
		# msg.exe $script:server_username /TIME:60 "$fn_name() bad_DAT_file_flg = $bad_DAT_file_flg"

		$extract_success_flg = $false  # Assume failure

		if (!$bad_DAT_file_flg) {

			# DAT file has at least one JPG file
			if ($found_ary1_cnt -gt 0) {

				# DAT file structure is NOT as expected
				if ($found_ary1_cnt -ne $found_ary2_cnt) {
				
					# Try to repair DAT file marker arrays before giving up
					$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
						RepairDatFile `
						-Bytes $bytes `
						-BegOfImageMarks $found_ary1 `
						-EndOfImageMarks $found_ary2
						
					# For debugging
					#Write-Host "result_flg = $result_flg"

					if ($result_flg) {
						$found_ary1 = $SOI_mrk_idxs_ary
						$found_ary2 = $EOI_mrk_idxs_ary
						$found_ary1_cnt = $SOI_mrk_idxs_ary.Count
						$found_ary2_cnt = $EOI_mrk_idxs_ary.Count

						if ($script:FromConsole) {
							if ($found_ary1_cnt -gt $found_ary2_cnt) {
								Write-Host "REPAIRED 'beg marker' indices: now = $jpg_SOI_mrk_idxs"
							} else {
								Write-Host "REPAIRED 'end marker' indices; now = $jpg_EOI_mrk_idxs"
							}
						}
					}
				}

				# DAT file structure is as expected
				if ($found_ary1_cnt -eq $found_ary2_cnt) {

					if ($found_ary1_cnt -ge $DatReqdImages) {
			
						if ($ImagesCntUsedFlg) {
							if ($script:FromConsole) {
								Write-Warning "Argument '-ImagesCnt' ignored; it applies only to GIFs created from BVR files"
							}
						}

						$jpgs_cnt = $found_ary1_cnt
						$jpg_idx_beg = [Math]::Floor($found_ary1_cnt*$ImagesTrimBegPct/100)
						$jpg_idx_end = $found_ary1_cnt - [Math]::Floor($found_ary1_cnt*$ImagesTrimEndPct/100)
						$jpgs_cnt_adj = $jpg_idx_end - $jpg_idx_beg
						
						# For debugging
						# Write-Host "$fn_name() jpgs_cnt = $jpgs_cnt"
						# Write-Host "$fn_name() jpgs_cnt_adj = $jpgs_cnt_adj"
						# Write-Host "$fn_name() jpg_idx_end = $jpg_idx_end"
						# Write-Host "$fn_name() jpg_idx_beg = $jpg_idx_beg"

						$rtn_obj['source'] = $dat_file
						$rtn_obj['jpgstotal'] = $jpgs_cnt
						$jpgs = @()
						$jpgs_cntr = 0
						#$bi_dat_viewer_time_ms = 0
						
						# EXTRACT THE JPG IMAGES
						
						# Initialize timer collectors
						$jpg_files_names_msec = 0
						$jpg_files_save_msec = 0
						$jpg_files_validate_msec = 0
						
						# KEY POINT: Before 5.9.4.11. JPGs in DAT files were stored newest-to-oldest
						$dat_images_oldest_first_flg = `
							IsBiVersion `
							-CompareToVersion '5.9.4.11' `
							-CurrentVersion $script:BiVersion
						
						# Start timer
						# $timer_jpgs_extract = [Diagnostics.Stopwatch]::StartNew()

						$i = 0
						foreach ($item in $found_ary1) {
							
							# For debugging
							# Write-Host ("$fn_name() item type = " + $item.GetType())
							# Write-Host "$fn_name() item = $item"

							if ($dat_images_oldest_first_flg) {
								$img_no = $found_ary1_cnt - $i
							} else {
								$img_no = $i + 1
							}
							
							# Extract image if in target range
							if ($img_no -ge $jpg_idx_beg -and $img_no -le $jpg_idx_end) {
							
								$bounding_idx1 = $item
								$bounding_idx2 = $found_ary2[$i] + 1  # Include the EOI marker!
								$img_bytes = $bytes[$bounding_idx1..$bounding_idx2]
							
								if ($img_bytes.Count -gt 0) {	

									# Start timer
									$timer_jpg_filename = [Diagnostics.Stopwatch]::StartNew()

									# Build JPG's filespec
									if ($dat_images_oldest_first_flg) {
										$jpg_filename_idx = $jpgs_cntr + 1
									} else {
										$jpg_filename_idx = $jpgs_cnt_adj - $jpgs_cntr 
									}								
									$jpg_filename_sfx = ([String]$jpg_filename_idx).PadLeft(2,'0')
									$jpg_filespec = "$jpg_path\img$jpg_filename_sfx.jpg"

									# Stop timer
									# $timer_jpg_filename.stop()
									# $elapsed_msec = [Math]::Round($timer_jpg_filename.elapsed.totalmilliseconds,1)
									# $jpg_files_names_msec += $elapsed_msec
									# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
									# Write-Host ($img_no.toString() + " timer_jpg_filename = $elapsed_msec_str msec")

									# Start timer
									# $timer_jpg_save = [Diagnostics.Stopwatch]::StartNew()
								
									# Save the JPG file to disk					
									if ($PsVersion -ge 6) {
										# REF: https://stackoverflow.com/a/63782608/13499984
										Set-Content "$jpg_filespec" -Value $img_bytes -AsByteStream 
									} else {
										Set-Content "$jpg_filespec" -Value $img_bytes -Encoding Byte
									}

									# Stop timer
									# $timer_jpg_save.stop()
									# $elapsed_msec = [Math]::Round($timer_jpg_save.elapsed.totalmilliseconds,1)
									# $jpg_files_save_msec += $elapsed_msec
									# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
									# Write-Host ($img_no.toString() + " timer_jpg_save = $elapsed_msec_str msec")

									# Start timer
									# $timer_jpg_validate = [Diagnostics.Stopwatch]::StartNew()

									# Confirm that the extracted data is a valid JPG file
									$jpg_valid_flg = `
										IsJpegFileValid `
										-FilePath $jpg_filespec

									# For debugging
									#msg.exe $script:server_username /TIME:60 "$jpg_filespec is valid = $jpg_valid_flg"

									# Stop timer
									# $timer_jpg_validate.stop()
									# $elapsed_msec = [Math]::Round($timer_jpg_validate.elapsed.totalmilliseconds,1)
									# $jpg_files_validate_msec += $elapsed_msec
									# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
									# Write-Host ($img_no.toString() + " jpg_files_validate_msec = $elapsed_msec_str msec")
									
									if (!$jpg_valid_flg) {
										if ($script:FromConsole) {
											Write-Warning "DAT file export aborted because '$jpg_filespec' is NOT a valid JPG"
										}
										break
									} else {
										$jpgs_cntr++
										$jpgs += $jpg_filespec
									
										# For debugging
										#Write-Host "$fn_name() jpgs = $jpgs"
										#Write-Host "'$jpg_filespec' is a valid JPG"
									}
								}
							}

							$i++
						}

						# Stop timer
						# $timer_jpgs_extract.stop()
						# $elapsed_msec = [Math]::Round($timer_jpgs_extract.elapsed.totalmilliseconds,1)
						# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
						# Write-Host ("timer_jpg_files_names = $jpg_files_names_msec msec")
						# Write-Host ("timer_jpg_files_save = $jpg_files_save_msec msec")
						# Write-Host ("timer_jpg_files_validate = $jpg_files_validate_msec msec")
						# Write-Host ("timer_jpgs_extract (overall) = $elapsed_msec_str msec")

						#Write-Host "bi_dat_viewer_time_ms = $bi_dat_viewer_time_ms" 

						$rtn_obj['jpgsextracted'] = $jpgs
						
						if ($jpgs.Count -gt 0) {
							$extract_success_flg = $true
						}

						if ($script:FromConsole -and !$script:Quiet) {
							Write-Host ("Extracted $jpgs_cntr of $jpgs_cnt total JPG files to folder '$jpg_path'")
						}
						
					} else {
						$rtn_obj = "FAIL: DAT file contains only $found_ary1_cnt/$DatReqdImages required JPGs"							
					}			

				} else {
					$rtn_obj = "FAIL: DAT file has unexpected structure [unbalanced JPG beg/end markers] ($dat_file)"
				}
			} else {
				$rtn_obj = "FAIL: No JPGs found in DAT file '$dat_file'"			
			}		
		} else {
			$rtn_obj = "FAIL: " + $bad_DAT_msg
		}
	} else {
		$rtn_obj = "FAIL: Unable to find DAT file '$dat_file'"
	}
	
	# For debugging
	#msg.exe $script:server_username /TIME:60 "extract_success_flg = $extract_success_flg"
		
	return $rtn_obj
}

function CopyFileAsSequence() {
# PURPOSE:  Copy a source file without over writing; append a sequence number '(N)' 
#           to the end of its basename until it is unique in the destination folder 
# ARGS:     see param lines below
# RETURNS:  (string) new_filespec, or 'fail'
# ATTRIBUTIONS:  https://stackoverflow.com/a/53595350/13499984

	param(
		[String]$SourceFilespec,  # [REQUIRED]
		[String]$TargetFilespec,  # [REQUIRED]
		[String]$DestPath         # [REQUIRED]
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "$fn_name() SourceFilespec = $SourceFilespec"
	#Write-Host "$fn_name() TargetFilespec = $TargetFilespec"
	#Write-Host "$fn_name() DestPath = $DestPath"
	
	$rtn_str = 'fail'  # Assume failure

	# Just in case
	$DestPath = $DestPath.TrimEnd('\')

	if ((Test-Path -LiteralPath "$SourceFilespec")) {

		if ((Test-Path -LiteralPath "$DestPath")) {

			$fileInfo = Get-Item -Path $TargetFilespec
			$base_name = $fileInfo.BaseName
			$extension = $fileInfo.Extension

			# Get an array of all filenames (name only) of the files with a similar name already present in the destination folder
			$allFiles = @(Get-ChildItem $DestPath -File -Filter "$base_name*$extension" | Select-Object -ExpandProperty Name)

			# Construct the new filename
			$new_name = $base_name + $extension  # Initialize
			$count = 1
			while ($allFiles -contains $new_name) {
				# add a sequence number in brackets to the end of the basename until it is unique in the destination folder
				$new_name = "{0}({1}){2}" -f $base_name, $count++, $extension
			}

			# Construct the new filespec
			$new_target_fs = Join-Path -Path $DestPath -ChildPath $new_name

			try {
				Copy-Item -Path $SourceFilespec -Destination $new_target_fs
				$rtn_str = $new_target_fs
			} catch {
				$rtn_str = "ERR/failed to create file '$new_target_fs'"
			}
		} else {
			$rtn_str = "ERR/destination folder does not exist ('$DestPath')"
		}
	} else {
		$rtn_str = "ERR/source file does not exist ('$FileName')"
	}

	return $rtn_str

}

function PurgeOldFiles() {
# PURPOSE:  Delete old files (> N hours) from a filepath like 'D:\myfolder\*.gif'
# ARGS:     see param lines below
# RETURNS:  (array) @(found_files_cnt, purged_files_cnt, oldest_file_hours, oldest_file_path)
#           or (string) 'ERR/' message
# ATTRIBUTIONS:  https://stackoverflow.com/a/46086926
	
    param(
		[String]$Pattern,  # [REQUIRED] full filespec pattern for targeted files; like 'D:\myfolder\*.jpg'
		[Int]$Hours        # [REQUIRED] maximum age for retained files; use 24 for 1 day, 168 for 1 week... 
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	if (!$Pattern -or !$Hours) {
		return "ERR/$fn_name() requires 2 arguments (-Pattern -Hours)"
	}

	# For debugging
	#Write-Host "Entered function: $fn_name()"
	#Write-Host "$fn_name() Pattern = $Pattern"
	#Write-Host "$fn_name() Hours = $Hours"
	
	# GET FILES MATCHING PATTERN
	$cur_ary = get-childitem -Path $Pattern | Select Name
	if ($cur_ary) {
		$cur_ary_cnt = $cur_ary.Count
	} else {
		$cur_ary_cnt = 0
	}

	# For debugging
	#Write-Host "$fn_name() cur_ary cnt = $cur_ary_cnt"
	#Write-Host "$fn_name() cur_ary = $cur_ary"

	# GET FILES MATCHING AGE REQUIREMENT
	$purge_ary = @()
	if ($cur_ary_cnt -gt 0 -and $Hours) {
		
		$limit = (Get-Date).AddHours(-$Hours)
		
		# For debugging
		#Write-Host "$fn_name() limit = $limit"

		$purge_obj = Get-ChildItem -Path $Pattern | 
			Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
			Select Name

		if ($purge_obj) {
			$purge_ary = $purge_obj.Name
		}
	}
	
	# Get count for return array
	$purge_ary_cnt = $purge_ary.Count
	
	# For debugging
	#Write-Host "$fn_name() purge_ary cnt = $purge_ary"
	#Write-Host ("$fn_name() purge_ary = " + ($purge_ary | ConvertTo-Json))

	# REMOVE FILES MATCHING AGE REQUIREMENT
	if ($purge_ary_cnt -gt 0) {
		if ($Hours) {
			$null = Get-ChildItem -Path $Pattern | 
				Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
				Remove-Item -Force
		}
	}

	# GET OLDEST REMAINING FILE
	$oldest_file = dir $Pattern | sort LastWriteTime | select -First 1
	$oldest_file_age_days = ((Get-Date) - $oldest_file.LastWriteTime).Days
	$oldest_file_age_hours = 24 * $oldest_file_age_days
	
	# For debugging
	#Write-Host ("$fn_name() oldest_file = " + $oldest_file.FullName)
	#Write-Host ("$fn_name() oldest_file_age_days = " + $oldest_file_age_days)
	#Write-Host ("$fn_name() oldest_file_age_hours = " + $oldest_file_age_hours)
	
	return  @($cur_ary_cnt, $purge_ary.Count, $oldest_file_age_hours, $oldest_file.Name)
}

function MakeGIF() {
# PURPOSE:  Make GIF image file to be used as attachment for notifications (Pushover, Telegram)
# ARGS:     see param lines below
# RETURNS:  (hashtable) @{'method'    = $gif_method; 
#                         'file'      = $output_fs; 
#                         'delay'     = $delay_ms; 
#                         'cnt'       = $ImagesCnt; 
#                         'datjpgs'   = $dat_jpgs_cnt; 
#                         'height'    = $Height; 
#                         'fps'       = $FPS; 
#                         'kb'        = $gif_kb ; 
#                         'msec'      = $gif_dur_ms; 
#                         'trimbeg'   = $gif_trim_bef_pct; 
#                         'trimend'   = $gif_trim_end_pct; 
#                         'trimcust'  = $gif_trim_end_cust_pct; 
#                         'almsec'    = $al_dur_ms; 
#                         'res'       = $gif_jpg_WxH ; 
#                         'cam'       = $al_cam; 
#                         'memo'      = $al_memo; 
#                         'source'    = $source_fs; 
#                         'archive'   = $archive_gif_fs
#                         'dblocator' = $clip_db_locator
#                        }
#           or (string) if error
# NOTES:    1. Be sure to check if 'gif_size' = 0, as this indicates the GIF file was not actually created
#           2. Pushover API attachment maximum size is 2,621,440 bytes (2621 KB)
# GLOBALS:  Used global vars: $FromConsole, $Quiet, $DevMode 
#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
#           $script:Quiet       = (boolean) $true suppresses verbose output to the console
#           $script:DevMode     = (boolean) $true 
# ATTRIBUTIONS:
#    Thanks to bp2008@ipcamtalk for providing clues to Blue Iris HTTP command to extract JPGs from a BVR
#
# Version 2 (8/29/2024) replaced *Flg args w/ global vars
# Coded by: jaydeel @ ipcamtalk.com
	
    param(
		[String]$RootPath,           # [REQUIRED] root folder used by this function (full path)
		[Array]$BvrPaths,            # [REQUIRED] full folder paths for Blue Iris' 'BVR' files (typ. /New/ folders)
		[String]$AlertPath,          # [REQUIRED] full folder path for Blue Iris' 'DAT' files (typ. /Alerts/ folder)
		[String]$AlertRec,           # Alert record, like '@123456789'
		[Boolean]$LiveFlg,           # create GIF from specified camera's stream
		[String]$Camera,             # camera shortname (required with -Live argument)
		[Int]$Seconds,               # GIF duration (-1 uses alert's duration via JSON cmd 'clipstats'
		[Boolean]$UseDatFileFlg,     # extract JPGs from AI DAT if it exists
		[Boolean]$LatestFlg,         # use DAT file from the most recent alert (that has a DAT file)
		[Int]$DatReqdImages,         # number of images that must be found before extracting any images	
		[Int]$DatImageIntervalMsec,  # number of milliseconds between images in the DAT file
		[Int]$BvrDelayMsec,          # delay fetching JPGs from live BVR for this many milliseconds
		[Boolean]$NoPushFlg,         # true prevents limiting GIF size to below 2.5MB
		[Boolean]$NoArchiveGifFlg,   # true suppresses the archiving of created GIF file
		[Int]$ImagesCnt,             # number of JPG frames to extract from BVR
		[Boolean]$ImagesCntUsedFlg,  # true indicates that the param -ImagesCnt was used on the command line
		[Int]$Height,                # height of image extracted from BVR in pixels
		[Int]$FPS,                   # GIF file frames per second
		[Int]$TrimBegPct,            # trim beginning of alert clip by this percentage
		[Int]$TrimEndPct,            # trim end of alert clip by this percentage
		[Int]$TrimEndCustPct,        # trim end of alert clip by this custom AI-determined percentage
		[Int]$ClipDurAdjPct          # percentage to reduces alert's duration (after applying -TrimBegPct and -TrimEndPct)
)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered function: $fn_name()"	
	# Write-Host "$fn_name() RootPath = '$RootPath'"
	# Write-Host ("$fn_name() BvrPaths = " + ($BvrPaths | ConvertTo-Json))
	# Write-Host "$fn_name() AlertPath = '$AlertPath'"
	# Write-Host "$fn_name() AlertRec = '$AlertRec'"
	# Write-Host "$fn_name() LiveFlg = '$LiveFlg'"
	# Write-Host "$fn_name() Camera = '$Camera'"
	# Write-Host "$fn_name() Seconds = $Seconds"
	# Write-Host ("$fn_name() Seconds = " + $Seconds.GetType())
	# Write-Host "$fn_name() UseDatFileFlg = $UseDatFileFlg"
	# msg.exe $script:server_username /TIME:10 "UseDatFileFlg = $UseDatFileFlg"		
	# Write-Host "$fn_name() LatestFlg = $LatestFlg"
	# Write-Host "$fn_name() DatReqdImages = $DatReqdImages"
	# Write-Host "$fn_name() DatImageIntervalMsec = $DatImageIntervalMsec"
	# Write-Host "$fn_name() BvrDelayMsec = $BvrDelayMsec"
	# Write-Host "$fn_name() NoPushFlg = $NoPushFlg"
	# Write-Host "$fn_name() NoArchiveGifFlg = $NoArchiveGifFlg"
	# Write-Host "$fn_name() ImagesCnt = $ImagesCnt"
	# Write-Host ("$fn_name() ImagesCntUsedFlg type = " + $ImagesCntUsedFlg.GetType())
	# Write-Host "$fn_name() ImagesCntUsedFlg = $ImagesCntUsedFlg"
	# Write-Host "$fn_name() Height = $Height"
	# Write-Host "$fn_name() FPS = $FPS"
	# Write-Host ("$fn_name() TrimBegPct type = " + $TrimBegPct.GetType())
	# Write-Host "$fn_name() TrimBegPct = $TrimBegPct"
	# Write-Host "$fn_name() TrimEndPct = $TrimEndPct"
	# Write-Host "$fn_name() TrimEndCustPct = $TrimEndCustPct"
	# Write-Host "$fn_name() ClipDurAdjPct = $ClipDurAdjPct"
	# Write-Host ("$fn_name() Quiet = " + $script:Quiet)
	# Write-Host ("$fn_name() FromConsole = " + $script:FromConsole"
	# Write-Host ("$fn_name() DevMode = " + $script:DevMode"
	# Write-Host ("$fn_name() cam_rec_is_continuous = " + $script:cam_rec_is_continuous)
	# Write-Host ("$fn_name() cam_pretrig_rectime_msec = " + $script:cam_pretrig_rectime_msec)

	# For debugging
	#$bg_msg =  "Entered function: $fn_name()"
	#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
	
	# Initialize vars
	$alert_rec = $AlertRec
	$db_rec = $DbRec
	$use_dat_file_flg = $UseDatFileFlg
	$trim_end_pct = 0
	
	# HANDLE ARGS
	if (1 -eq 1) {  # Use to collapse this code section while editing

		if ($LiveFlg) {

			# Override default or requested settings
			$use_dat_file_flg = $false
		
			# Get the MOST RECENT BVR FILE for the specified camera
			# NOTE: Even though we use the live video stream to grab JPGs for a 'Live' GIF
			#       we still use the active BVR filename to name the GIF file
			$bvr_timestamp = Get-Date -Format "yyyyMMdd_"
			$pattern = "$Camera.$bvr_timestamp*bvr"  # like, 'DW2.20231103'
			$bvr_file_found = $false
			# ... Find the New folder path
			foreach ($path in $BvrPaths) {
				#$bvr_file_obj = ((Get-ChildItem "$path\*" -Recurse -Include $pattern) | Sort -Descending) | Select -First 1
				# This version ALSO works w/ non-default file naming patterns in the Blue Iris Settings > Record tab
				$bvr_file_obj = ((Get-ChildItem "$path\*" -Recurse -Include $pattern) | Sort LastWriteTime) | Select -Last 1
				if ($bvr_file_obj) {
					$bvr_file_found = $true
					break
				}
			}
			
			# For debugging
			#Write-Host "$fn_name() bvr path = '$bi_new_folder_path\*'"
			#Write-Host "$fn_name() pattern = '$pattern'"
			#Write-Host ("$fn_name() bvr_file_obj type = " + $bvr_file_obj.GetType())
			#Write-Host ("$fn_name() bvr_file_obj = " + ($bvr_file_obj | ConvertTo-Json))
			#Write-Host ("$fn_name() bvr_file_obj.FullName = " + $bvr_file_obj.FullName)
			#Write-Host ("$fn_name() bvr_file_obj.Name = " + $bvr_file_obj.Name)
			#Write-Host "$fn_name() bvr_file = '$bvr_file'"

			if ($bvr_file_found) {
				
				if ($bvr_file_obj.FullName) {
					
					# MAKE ELEMENTS FOR THE ARCHIVED GIF FILE NAME				
					$bvr_filename = $bvr_file_obj.Name  # like, 'DW1.20240302_130000.bvr'
					$arr = ($bvr_filename).Split('.')
					# like, ["DW1", "20240302_130000", "bvr"]
					$bvr_path = $arr[0] + "." + $arr[1] + ".bvr"

					# CALCULATE THE NOW OFFSET TIME - based the most recent BVR file's creation time
					$bvr_filespec = $bvr_file_obj.FullName
					$file_dt = (Get-Item -LiteralPath "$bvr_filespec").CreationTime
					$diff = (Get-Date).Subtract($file_dt)
					$bvr_offset_ms = [Math]::Floor($diff.TotalMilliSeconds)

					# Make the filename for the archived GIF file				
					# ... like, 'DW1.20240302_130000.1464437.gif'
					$archive_gif_path = $arr[0] + "." + $arr[1] + "." + $bvr_offset_ms + ".gif"
					
					# For debugging
					#Write-Host "$fn_name() bvr_filespec = '$bvr_filespec'"
					#Write-Host "$fn_name() bvr_filename = '$bvr_filename'"
					#Write-Host "$fn_name() bvr_path = '$bvr_path'"
					#Write-Host ("$fn_name() bvr_filename arr = " + ($arr | ConvertTo-Json))
					#Write-Host "$fn_name() bvr_offset_ms = '$bvr_offset_ms'"
					#Write-Host "$fn_name() archive_gif_path = '$archive_gif_path'"
					
					<# if ($script:FromConsole -and !$script:Quiet) {
						Write-Host "Active bvr_filespec:    $bvr_filespec"
						Write-Host "Active bvr_file offset: $bvr_offset_ms msec"
					} #>

					# For debugging
					#$bg_msg = "$fn_name() Active bvr_file:`n$bvr_filespec`nbvr_file offset: $bvr_offset_ms ms"
					#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on

				} else {
					$err_msg = "ERR/unable to get active BVR filename: '$bvr_filespec'"
					return $err_msg
				}
			} else {
				$err_msg = "ERR/unable to find existing BVR file"
				return $err_msg
			}
		}

		if ($LatestFlg) {
			
			# GET BLUE IRIS DB ALERTLIST
			# Returns hashtable, like @{ Cam1: @{ prop1:[values], ... , propN:[values] }, ... }
			$response_obj = `
				BIJsonGetAlertList `
				-Cameras $Camera `
				-LatestFlg $LatestFlg `
				-AlertProps 'path','file','memo','date' `
				-ExcludeClones
				
			# For testing - Simulate a bad JSON alertlist response
			#$response_obj = "ERR/test"
			
			if ($response_obj.GetType() -match 'string') {
				if ($response_obj.StartsWith('ERR/')) {
					return $response_obj	
					if ($script:FromConsole) {
						Write-Host "$fn_name() = $err_msg"
					} else {
						return $err_msg
					}
				} else {
					$err_msg = "ERR/unexpected response from BIJsonGetAlertList()"
					if ($script:FromConsole) {
						Write-Host "$fn_name() = $err_msg"
					} else {
						return $err_msg
					}
				}
			}

			$alertlist_hashtable = $response_obj

			# For debugging
			# Write-Host ("$fn_name() hashtable keys = " + $alertlist_hashtable.Keys)
			# Write-Host ("$fn_name() hashtable = " + ($alertlist_hashtable | ConvertTo-Json))

			$al_paths_ary  = [array]$alertlist_hashtable.$Camera.path
			$al_files_ary  = [array]$alertlist_hashtable.$Camera.file
			$al_memos_ary  = [array]$alertlist_hashtable.$Camera.memo
			$al_dates_ary  = [array]$alertlist_hashtable.$Camera.date
			
			$db_rec       = $al_paths_ary[0].Replace('.bvr','')
			$db_file      = $al_files_ary[0]
			$db_ts_sec    = $al_dates_ary[0]
			$al_memo      = $al_memos_ary[0]
			$dat_file     = $db_file.Replace('.jpg','.dat')
			$dat_fs       = $AlertPath + "\" + $dat_file
			$now_ts_sec   = [DateTimeOffset]::UtcNow.ToUnixTimeSeconds()
			$alert_age_sec = [Int64]$now_ts_sec - [Int64]$db_ts_sec
			
			if ($UseDatFileFlg) {
				if (!(Test-Path -LiteralPath "$dat_fs")) {
					$use_dat_file_flg = $false
					
					# For debugging
					Write-Host "$fn_name() unable to find DAT file '$dat_fs'"
				}
			}

			# For debugging
			# Write-Host ("$fn_name() al_paths_ary = " + ($al_paths_ary | ConvertTo-Json))
			# Write-Host ("$fn_name() al_files_ary = " + ($al_files_ary | ConvertTo-Json))
			# Write-Host "$fn_name() latest alert db_rec = $db_rec"
			# Write-Host "$fn_name() latest alert db_file = $db_file"
			# Write-Host "$fn_name() latest alert db_memo = $db_memo"
			# Write-Host "$fn_name() latest alert dat_file (target) = $dat_file"
			# Write-Host "$fn_name() latest alert dat_filespec (target) = $dat_fs"
			# Write-Host ("$fn_name() al_memos_ary = " + ($al_memos_ary | ConvertTo-Json))
			# Write-Host "$fn_name() al_memo = $al_memo"
			# Write-Host ("$fn_name() latest alert unix timestamp type = " + $db_ts_sec.GetType())
			# Write-Host "$fn_name() latest alert unix timestamp (sec) = $db_ts_sec"
			# Write-Host ("$fn_name() now unix timestamp type = " + $now_ts_sec.GetType())
			# Write-Host "$fn_name() now unix timestamp (sec) = $now_ts_sec"
			# Write-Host "$fn_name() unix timestamp delta (sec) = $alert_age_sec"
			
		}
		
		if ($DbRec.Length -gt 0) {
			
			# Also supports UI3 style 'rec='
			# Example: http://192.168.1.3:81/ui3.htm?rec=213168393-3165057&cam=GT
			# UI3Rec = '213168393-3165057'
			
			$hypen_idx = $DbRec.IndexOf('-')
			if ($hypen_idx) {
				$arr = $DbRec.Split('-')
				$db_rec = "@" + ($arr[0]).ToString()
				
				# Currently not used
				if ($arr[1]) {
					$dbrec_offset = [Int]$arr[1]
				} else {
					$dbrec_offset = $null
				}
				
				# For debugging
				#Write-Host "Database record ID = $alert_rec"
				#Write-Host "Database record offset = $dbrec_offset"
				
			} else {
				$db_rec = "@" + $DbRec.ToString()
				$dbrec_offset = $null
			}
		}
	}
	
	# For debugging
	#Write-Host "$fn_name() alert_rec = $alert_rec"
	#msg.exe $script:server_username /TIME:10 "alert_rec = '$alert_rec'`ndb_rec = '$db_rec'"		
	#Write-Host ("$fn_name() db_rec len = " + $db_rec.Length)
	
	# Initialize
	$gif_method = "viaClip"  # Default
	$delay_ms = 0
	$dat_jpgs_cnt = 0
	$gif_images_cnt = $ImagesCnt
	$gif_dur_adj_pct = 0
	$clip_db_locator = $null
				
	# A custom TrimEnd takes precedence
	if ($TrimEndCustPct -gt 0) {
		$trim_end_pct = $TrimEndCustPct
	} else {
		$trim_end_pct = $TrimEndPct
	}
		
	# TRY TO CREATE GIF
	# ... up to 3x before giving up
	for ($i=1; $i -le 3; $i++) {
		
		# For debugging
		#Write-Host "$fn_name() iteration #$i"
		#msg.exe $script:server_username /TIME:5 "trying to create GIF, iteration #$i"		

		if (!$LiveFlg) {

			# GET THE ALERT'S CLIP DETAILS
			
			# Make the DB locator
			if ($alert_rec.Length -gt 0) {
				$clip_db_locator = $alert_rec
			}
			if ($db_rec.Length -gt 0) {
				$clip_db_locator = $db_rec
			}

			# For debugging
			#Write-Host "$fn_name() clip_db_locator = $clip_db_locator"
			#msg.exe $script:server_username /TIME:20 "clip_db_locator = $clip_db_locator"
			
			$clip_stats_obj = `  # A hashtable
				BIJsonClipstats `
				-Clips $clip_db_locator `
				-Props 'camera,path,file,date,offset,triggeroffset,alertmsec,res,memo'
				
			# $response_obj should be like ... 
			#[
			#	{
			#		"result": "success"
			#	},
			#	{
			#		"@152866668.bvr": {
			#			"camera": "DW1",
			#			"path": "@152720932.bvr",
			#			"file": "DW1.20230530_060000.1965063.17-0.jpg",
			#           "date": 123456789
			#			"offset": 2016632,
			#           "triggeroffset": 2018632,
			#			"alertmsec": 27426,
			#			"res": "2688x1520"
			#			"memo": "LD/C;person:91%"
			#		}
			#	}
			#]
			
			# For debugging
			# Write-Host ("$fn_name() clip_stats_obj type = " + $clip_stats_obj.GetType())
			# Write-Host ("$fn_name() clip_stats_obj = " + ($clip_stats_obj | ConvertTo-Json))

			if ($clip_stats_obj.GetType() -match "string") {
				if ($clip_stats_obj.StartsWith("ERR/")) {
					$err_msg = $clip_stats_obj
					return $err_msg
				}
			}

			$clipstats_result = $clip_stats_obj[0]  # A single element array
			$clipstats_data = $clip_stats_obj[1]  # An array of arrays
			$clipstats_success_flg = ($clipstats_result -match "result=success")
			
			# For debugging
			# msg.exe $script:server_username /TIME:20 "clipstats_success_flg = $clipstats_success_flg"	
			# Write-Host "$fn_name() clipstats_result = $clipstats_result"
			# Write-Host ("$fn_name() clipstats_data = " + ($clipstats_data | ConvertTo-Json))
			# $bg_msg = ("clipstats_data = " + ($clipstats_data | ConvertTo-Json))	
			# msg.exe $script:server_username /TIME:20 $bg_msg	
			
			# Show the clipstats object
			# ... tweaked to show the date in readable form that matches the Blue Iris clipslist
			if ($script:FromConsole -and !$script:Quiet) {
				$clonedObject = $clip_stats_obj.PsObject.Copy()
				[Array]$clonedObjectKeys = $clonedObject.Keys
				# Write-Host ("clonedObjectKeys type = " + $clonedObjectKeys.GetType())
				# Write-Host ("clonedObjectKeys len = " + $clonedObjectKeys.length)
				# Write-Host ("clonedObjectKeys = " + ($clonedObjectKeys | ConvertTo-Json))
				$key = $clonedObjectKeys[1]
				$timestamp = $clonedObject.$key.date
				# Write-Host ("clonedObject[$key] timestamp = " + $timestamp.GetType())
				# Write-Host ("clonedObject[$key] timestamp = " + $timestamp)
				$formatteddate = Convert-FromUnixDate $timestamp "M/d/yyyy h:mm:ss tt"
				$clonedObject[1].$key.date = $timestamp.ToString() + " <-- " + $formatteddate
				Write-Host "Alert's JSON 'clipstats' response (selected):"
				Write-Host "---------------------------------------------"
				Write-Host ($clonedObject[1] | ConvertTo-Json)
			}
			
		}
		
		if ($clipstats_success_flg -or $LiveFlg -or $LatestFlg) {

			if ($LiveFlg) {

				# PREP TO MAKE GIF FROM LIVE STREAM

				# For debugging
				#Write-Host "`n$fn_name() Preparing to make LIVE GIF from camera stream..."	
				#msg.exe $script:server_username /TIME:20 "Preparing to make LIVE GIF from camera stream..."	

				$al_cam    = $Camera
				$al_path   = $bvr_path
				$al_file   = $bvr_filename
				$al_offset = $bvr_offset_ms
				$al_dur_ms = $Seconds * 1000
				$al_res    = "WxH"
				$delay_ms  = 0

				# For debugging
				#Write-Host "$fn_name() al_cam = $al_cam"
				#Write-Host "$fn_name() al_path = '$al_path'"
				#Write-Host "$fn_name() al_file = '$al_file'"

				# Adjust source clip offset
				$gif_offset = $al_offset + ($TrimBegPct/100)*$al_dur_ms
				
				# Set GIF duration (round up to nearest whole second)
				$gif_dur_ms = [Math]::Floor($al_dur_ms)
	
				# Calculate time delta between each extracted BVR image (round up to nearest whole second)
				$step_ms = [Math]::ceiling($gif_dur_ms / ($gif_images_cnt - 1))

				# For debugging
				# Write-Host "$fn_name() al_cam = $al_cam"
				# Write-Host "$fn_name() al_path = $al_path"
				# Write-Host "$fn_name() al_offset = $al_offset"
				# Write-Host "$fn_name() gif_offset = $gif_offset"
				# Write-Host "$fn_name() al_dur_ms = $al_dur_ms"
				# Write-Host "$fn_name() gif_dur_ms = $gif_dur_ms"
				# Write-Host "$fn_name() gif_images_cnt = $gif_images_cnt"
				# Write-Host "$fn_name() step_ms = $step_ms"
				
			} else {
				
				# PREP TO MAKE GIF FROM BVR FILE
				
				# For debugging
				#Write-Host "`n$fn_name() Preparing to make GIF from BVR file..."	
				#msg.exe $script:server_username /TIME:20 "Preparing to make GIF from BVR file..."	

				# Because BIJsonClipstats() accepts multiple AlertRec's, we need to use only the first item
				$clipstats_item = $clipstats_data[0]
			
				# For debugging
				#Write-Host ("$fn_name() clipstats_item = " + ($clipstats_item | ConvertTo-Json))

				$pre_trig_playback_msec = 0  # Default
				if ([Int]$clipstats_item.triggeroffset - [Int]$clipstats_item.offset -gt 0) {
					$pre_trig_playback_msec = [Int]$clipstats_item.triggeroffset - [Int]$clipstats_item.offset
				}

				if ($script:FromConsole -and !$script:Quiet) {
					Write-Host ("NOTE: 'alertmsec' includes the pre-trigger playback time (triggeroffset - offset = " + $pre_trig_playback_msec + " msec)")
				}
				
				if ($clipstats_item -match 'NA') {
				
					# GET ALERT'S DETAILS
					# ... $clipstats_item via BIJsonClipstats -Clips $AlertRec -Props 'camera,path,file,msec,alertmsec,res'
					# NOTE: $al_file can be like:
					#       1) DW2.20231027_120000.49763.17-0.jpg 
					#          ... this the typical filespec
					#       2) 2023-10-27\DW2\DW2.20231027_120000.49763.17-0.jpg
					#          ... non-default file naming patterns in the Blue Iris Settings > Record tab
					$al_cam     = $clipstats_item.camera
					$al_path    = $clipstats_item.path
					$al_file    = $clipstats_item.file  
					$al_res     = $clipstats_item.res  # WxH
					$al_offset  = [Int]$clipstats_item.offset
					$al_dur_ms  = [Int]$clipstats_item.alertmsec
					$al_memo    = $clipstats_item.memo
					
					# Change alert duration, if requested
					if ($Seconds -gt 0) {
						$al_dur_ms = $Seconds * 1000
					}
					$al_dur_ms_orig  = $al_dur_ms 

					# For debugging
					#Write-Host "$fn_name() al_dur_ms = $al_dur_ms"
					
					# Confirm BVR exists
					if ($al_file) {
						$al_file_ary = $al_file.Split('.')  
						# Expecting array like @('DW2', '20231027_120000', '49763', '17-0', 'jpg')
						# ...         or like  @('2023-10-27\DW2', '20231027_120000', '49763', '17-0', 'jpg')
						$bvr_file = $al_file_ary[0] + "." + $al_file_ary[1] + ".bvr"  
						# $bvr_file should be like 'DW2.20231027_120000.bvr'
						# ...              or like '2023-10-27\DW2.20231027_120000.bvr'
						$bvr_file_found = $false
						foreach ($path in $BvrPaths) {
							$bvr_filespec = $path + "\" + $bvr_file
							if (Test-Path -Path "$bvr_filespec") {
								$bvr_file_found = $true
								break
							}
						}

					} else {
						$err_msg = "ERR/non-existent database record id"
						return $err_msg									
					}
						
					# For debugging
					#Write-Host "$fn_name() bvr_file_found = $bvr_file_found"
					
					if (!$bvr_file_found) {

						# ALTERNATE CASE ...
						# ... Check if Clipboard clip -OR- or Clipboard clip saved as Alert
						#     Like "Clipboard\\DW2.20240330_175026-175045.378.bvr"
						
						$al_file_ary1 = $al_file.Split('\')  
						# $al_file_ary1 like @('Clipboard', 'DW2.20240330_175026-175045.378.bvr')
						#                 or @('DW2.20240330_175026-175045.378.bvr')
						$al_file_ary2 = ($al_file_ary1[$al_file_ary1.Count - 1]).Split('.')  
						# $al_file_ary2 like @('DW2', '20240330_175026-175045', '378', 'bvr')
						$al_file_ary3 = ($al_file_ary2[1]).Split('_')  
						# $al_file_ary3 like @('20240330', '175026-175045')
						$al_file_ary4 = ($al_file_ary3[1]).Split('-')  
						# $al_file_ary4 like @('175026, '175045')
						$hour_str = $al_file_ary3[1].Substring(0,2)
						# REF: https://devblogs.microsoft.com/scripting/powertip-use-powershell-to-create-repeating-letters/
						$zeros_str = "0"*($al_file_ary4[0].Length - 2)  
						$bvr_file = $al_file_ary2[0] + "." + $al_file_ary3[0] + "_" + $hour_str + $zeros_str + ".bvr"  
						$bvr_file_found = $false
						$bvr_filespec = $path + "\" + $bvr_file
						if (Test-Path -Path "$bvr_filespec") {
							$bvr_file_found = $true
							# Override default or requested settings
							$use_dat_file_flg = $false
						}

						# For debugging
						# Write-Host ("$fn_name() al_file_ary1 = " + ($al_file_ary1 | ConvertTo-Json))
						# Write-Host ("$fn_name() al_file_ary2 = " + ($al_file_ary2 | ConvertTo-Json))
						# Write-Host ("$fn_name() al_file_ary3 = " + ($al_file_ary3 | ConvertTo-Json))
						# Write-Host ("$fn_name() al_file_ary4 = " + ($al_file_ary4 | ConvertTo-Json))
						Write-Host "$fn_name() bvr_filespec = $bvr_filespec"
						
					}
					
					if ($bvr_file_found) {
						
						# For debugging
						#Write-Host "$fn_name() al_file = $al_file"
						#Write-Host "$fn_name() bvr_filespec = $bvr_filespec"
						#Write-Host "$fn_name() bvr_file_path = $bvr_filespec_path"
						#Write-Host "$fn_name() bi_new_folder_path = $bi_new_folder_path" 
						
						# SPECIAL CASE...
						# ... Trap rate instances when 'clipstats' yields a zero value for 'alertmsec'
						if ($DbRec) {
							if ($al_dur_ms -eq 0) {
								$err_msg = "ERR/argument '-AlertRec', '-DbRecNo', or '-UI3Rec' does not point to an EXISTING Alert record (clipstats 'alertmsec' = 0)"
								return $err_msg			
							}
						}

						# Adjust source clip offset
						$gif_offset = $al_offset + ($TrimBegPct/100)*$al_dur_ms

						# Adjust GIF's duration, rounding up to nearest whole second
						# ... NOTE: $trim_end_pct already accounts for $TrimEndCustPct adjustment, which applies only 
						#     if user-setting $gif_trim_end_cust_obj is configured
						if ($TrimBegPct -ge 0) {
							$gif_dur_ms = [Math]::Floor( $al_dur_ms - ($TrimBegPct/100)*$al_dur_ms - ($trim_end_pct/100)*$al_dur_ms )
						} else {
							$gif_dur_ms = [Math]::Floor( $al_dur_ms - ($TrimBegPct/100)*$al_dur_ms - (($trim_end_pct - $TrimBegPct)/100)*$al_dur_ms )
						}
						
						if ($script:DevMode -and $script:FromConsole) {

							Write-Host ("(dev mode) alert's age (sec) = " + $alert_age_sec + " (alert's timestamp - NOW = " + $db_ts_sec + " - " + $now_ts_sec + ")") -ForegroundColor Cyan
							
							Write-Host ("(dev mode) gif offset (msec) BEFORE = " + $al_offset + ", AFTER = " + $gif_offset + ", ADJ = " + ($gif_offset - $al_offset) + " (" + 100*(($gif_offset - $al_offset)/$al_dur_ms) + "% of BEFORE duration)") -ForegroundColor Cyan
							
							Write-Host ("(dev mode) gif duration (msec) BEFORE = " + $al_dur_ms + ", AFTER = " + $gif_dur_ms + ", ADJ = " + ($al_dur_ms - $gif_dur_ms) + " (" + 100*(($al_dur_ms - $gif_dur_ms)/$al_dur_ms) + "% of BEFORE duration)") -ForegroundColor Cyan
						}
						
						# Display adjustment (if any)
						if ($gif_dur_ms -lt $al_dur_ms) {
							if ($script:FromConsole -and !$script:Quiet) {
								Write-Host "`nGIF Duration Adjustment:"
								Write-Host "--------------------------"
								Write-Host "TrimBegPct/TrimEndPct = $TrimBegPct/$trim_end_pct"
								Write-Host "Alert's offset adj = $gif_offset msec"
								Write-Host "Alert's duration = $al_dur_ms msec"
								Write-Host "GIF's duration   = $gif_dur_ms msec"
							}
						}
			
						# Calculate time delta between each extracted BVR image (round up to nearest whole second)
						$step_ms = [Math]::ceiling($gif_dur_ms / ($gif_images_cnt - 1))

						# For debugging
						#Write-Host "$fn_name() al_cam = $al_cam"
						#Write-Host "$fn_name() al_path = $al_path"
						#Write-Host "$fn_name() al_offset = $al_offset"
						#Write-Host "$fn_name() gif_offset = $gif_offset"
						#Write-Host "$fn_name() al_dur_ms = $al_dur_ms"
						#Write-Host "$fn_name() gif_dur_ms = $gif_dur_ms"
						#Write-Host "$fn_name() gif_images_cnt = $gif_images_cnt"
						#Write-Host "$fn_name() step_ms = $step_ms"
						
					} else {
						$err_msg = "ERR/unable to find existing BVR file"
						return $err_msg
					}
					
				} else {
					$err_msg = "ERR/non-existent alert record '$alert_rec'"
					return $err_msg
				}
							
			}				
			
			# MAKE SUBFOLDER FOR THE GIF AND ITS JPGS
			# ... Note: $RootPath is user-defined (_user_settings.ps1 file)
			$path = $RootPath + "\" + $al_cam  # like 'D:\pushover_gifs\DW1' 
			if (!(Test-Path -LiteralPath $path)) {
				New-Item -ItemType Directory -Force -Path $path
			}
		
			# DELETE ANY EXISTING *.JPG AND *.GIF FILES ASYNCHRONOUSLY
			get-childitem -path ($path + "\img*.jpg") | remove-item -force
			get-childitem -path ($path + "\*.gif") | remove-item -force

			# TRY GETTING THE GIF'S JPGs - 2 WAYS
			# ====================================================================
			# 1. GET JPGs FROM ALERT'S DAT FILE
			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			if ($use_dat_file_flg) {

				# For debugging
				#$bg_msg = "$fn_name()`nGetting JPGs from Blue Iris DAT file..."
				#Write-Host ("`n" + $bg_msg)
				#msg.exe $script:server_username /TIME:20 "$bg_msg"	
				#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on

				if ($script:FromConsole -and !$script:Quiet) {
					#Write-Host ""
					#Write-Host "Attempting to extract JPGs from AI Analysis 'DAT' file..."
				}
				
				# Negative left trims are not allowed for DAT file extractions
				if ($TrimBegPct -lt 0) {
					$TrimBegPct = 0
					
					if ($script:FromConsole -and !$script:Quiet) {
						Write-Host ""
						Write-Warning ("Negative -TrimBegPct argument is ignored for DAT file extractions; changed to zero")
					}
				}
				
				# Extract the JPGs
				$extract_obj = `
					ExtractDatImages `
					-AlertFile $al_file `
					-AlertPath $bi_alerts_folder_path `
					-DestPath $RootPath `
					-DatReqdImages $DatReqdImages `
					-LatestFlg $LatestFlg `
					-ImagesCnt $ImagesCnt `
					-ImagesCntUsedFlg $ImagesCntUsedFlg `
					-ImagesTrimBegPct $TrimBegPct `
					-ImagesTrimEndPct $trim_end_pct
					
				# extract_obj is like...
				# {
				#   "source": "D:\\BlueIris\\Alerts\\DW1.20240306_080000.1950096.5-0.dat",
				#   "jpgstotal": 13,
				#   "jpgsextracted": 
				#      [
				#         "D:\\BI_gif_tool\\DW1\\img05.jpg",
				#         "D:\\BI_gif_tool\\DW1\\img04.jpg",
				#         "D:\\BI_gif_tool\\DW1\\img03.jpg",
				#         "D:\\BI_gif_tool\\DW1\\img02.jpg",
				#         "D:\\BI_gif_tool\\DW1\\img01.jpg"
				#      ]
				# }
				
				# For debugging
				#$msg = "$fn_name() extract_obj type = " + $extract_obj.GetType()
				#msg.exe $script:server_username /TIME:60 $msg
				#msg.exe $script:server_username /TIME:60 $extract_obj.Count
								
				# For testing - Simulate a bad DAT JPG extraction
				#$extract_obj = $null
				#$extract_obj = "ERR/ test"
				#$extract_obj = "FAIL: test"
				#$extract_obj = "ISSUE: test"

				if ($extract_obj) {
					
					if ($extract_obj -match 'ERR/' -or $extract_obj -match 'FAIL:' -or $extract_obj -match 'ISSUE:') {
						
						if ($extract_obj -match 'FAIL: Unable to find DAT file') {
							
							$gif_method = 'viaClip b/c DAT file missing'
							
							if ($script:FromConsole -and !$script:Quiet) {
								Write-Host "Unable to find DAT file, trying BVR extraction instead" -ForegroundColor Green
							}
							
						} elseif ($extract_obj -match 'ISSUE: insufficient JPG images count in DAT file') {
							
							$gif_method = 'viaClip b/c too few DAT images'
							
							if ($script:FromConsole -and !$script:Quiet) {
								Write-Host "DAT has an insufficient number of JPG images, trying BVR extraction instead..." -ForegroundColor Green
							}

						} else {
						
							return $extract_obj
						}
						
					} elseif ($extract_obj.Count -gt 0) {

						# SUCCESS

						if ($script:FromConsole -and !$script:Quiet) {
							Write-Host ($extract_obj | ConvertTo-Json)
						}

						$gif_method = "viaDat"
						$dat_file = $extract_obj['source']
						$dat_jpgs_cnt = $extract_obj['jpgstotal']
						$gif_images_cnt = ($extract_obj['jpgsextracted']).Count
						$gif_dur_ms = $DatImageIntervalMsec*($gif_images_cnt  - 1)
						
						# For debugging
						#Write-Host "$fn_name() dat_file = '$dat_file'"
						#Write-Host "$fn_name() dat_jpgs_cnt** = $dat_jpgs_cnt"
						#Write-Host "$fn_name() gif_images_cnt = $gif_images_cnt"
						#Write-Host "$fn_name() gif_dur_ms = $gif_dur_ms"

					} else {
						$err_msg = "ERR/no images extracted from DAT file"
						return $err_msg
					}
				} else {
					$err_msg = "ERR/DAT extraction failed; function ExtractDatImages() returned null"
					return $err_msg
				}
			}
			
			# For debugging
			#msg.exe $script:server_username /TIME:60 "gif_method = $gif_method"

			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			# 2. GET JPGs FROM CAMERA'S STREAM (-LIVE) OR ALERT'S VIDEO CLIP FILE
			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			if (!($gif_method -match "viaDat")) {
				
				if ($LiveFlg) {

					if ($script:FromConsole -and !$script:Quiet) {
						Write-Host "`nGetting JPGs from the Camera's Live Video Stream..."
						Write-Host "---------------------------------------------------"
					}

					# For debugging
					#$bg_msg = "$fn_name()`nGetting JPGs from the Camera's Live Video Stream..."
					#msg.exe $script:server_username /TIME:20 "$bg_msg"	
					#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
					
					$gif_method = "viaLive"

					for ($j=0; $j -lt $gif_images_cnt; $j++) {

						# Build the URL
						$time = ($gif_images_cnt - $j - 1) * $step_ms
						$uri = $script:bi_webserver + "/image/" + $Camera + "&cache=1&h=" + $Height
						#$uri = $script:bi_webserver + "/image/" + $Camera + "&h=" + $Height

						# Build the full filespec for the URL's output
						$jpg_filename_sfx = ([String]($j + 1)).PadLeft(2,'0')
						$filespec = $RootPath + "\" + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"
						
						# GRAB THE IMAGE
						Invoke-RestMethod -UseBasicParsing -Uri $uri -OutFile $filespec | Out-Null

						# Give feedback on progress
						if ($script:FromConsole -and !$script:Quiet) {
							Write-Host ("Initiated request for $uri, after " + $j*$step_ms + " msec")
						}

						Start-Sleep -Milliseconds $step_ms

					}
					
				} 
				
				if (!$LiveFlg) {

					if ($script:FromConsole -and !$script:Quiet) {
						#Write-Host "`nGetting JPGs from Blue Iris Clip..."
						Write-Host "`nBVR Clip Processing Summary:"
						Write-Host "----------------------------"
					}
					
					# DELAY THE FETCHING OF JPGS FROM THE BVR
					if ($LatestFlg) {
						if ($script:cam_rec_is_continuous) {
							$bvr_delay_msec = 0  # Default
							if ($BvrDelayMsec) {
								# Use the custom value provided via cmdline argument -BvrDelayMsec
								$bvr_delay_msec = $BvrDelayMsec  
							} else {
								
								# Determine if alert is 'fresh' (just happened)
								#$alert_is_fresh = ($alert_age_sec -lt $al_dur_ms/1000)
								$alert_is_fresh = ($alert_age_sec -lt $al_dur_ms_orig/1000)
								
								# For testing
								# $alert_is_fresh = $true
								
								# If the latest Alert is 'fresh' (just happened), then delay the JPG fetch
								# ... Why? otherwise, the "/file/clips/" HTTP interface command will grab frames 
								#     from the start of the clip instead of where we want
								# ... The reason for this is that the BVR is still buffering and the HTTP command
								#     is (currently) unable to retrieve JPGs from the buffer.
								#     (Ken at blueirissupport told be once that this is doable, just not implemented yet)
						
								if ($alert_is_fresh) {
									# Delay for a period at least as long as the alert's duration
									#$bvr_delay_msec = [Int](1.10 * $al_dur_ms) + 1
									$bvr_delay_msec = 2 * $script:cam_pretrig_rectime_msec + 1
								}

								# For testing
								#$bvr_delay_msec = 0

								# For debugging
								# Write-Host "$fn_name() bvr_delay_msec = $bvr_delay_msec msec"
								
								# Provide feedback if the delay will be noticeable in the console
								if ($script:FromConsole -and $bvr_delay_msec -gt 1500) {
									Write-Host ("Alert is 'fresh', so delaying JPG fetch for $bvr_delay_msec msec (" + ($bvr_delay_msec/1000).ToString("0.0") + " sec)...") -ForegroundColor green
								}

								Start-Sleep -Milliseconds $bvr_delay_msec
							}
						}
					}

					# For debugging
					#$bg_msg = "$fn_name()`nGetting JPGs from Blue Iris BVR Clip..."
					#msg.exe $script:server_username /TIME:20 "$bg_msg"	
					#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on

					# Build arrays FOR A RUNSPACE that will asynchronously execute many URLs 
					$urls = @()   # Blue Iris HTTP Interface commands - to extract single JPGs from an alert's BVR
					$files = @()  # JPG filenames, serialized like 'img001.jpg' to 'img010.jpg' 
					$offset = $gif_offset
					for ($j=0; $j -lt $gif_images_cnt; $j++) {

						# Build the URL
						$time = $offset + ($j * $step_ms)
						
						#$uri = $script:bi_webserver + "/file/clips/" + $al_path + "?time=" + $time + "&h=" + $Height
						$uri = $script:bi_webserver + "/file/clips/" + $al_path + "?time=" + $time + "&cache=1&h=" + $Height

						# Build the full filespec for the URL's output
						$jpg_filename_sfx = ([String]($j + 1)).PadLeft(2,'0')
						$filespec = $RootPath + "\" + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"

						$urls  += $uri
						$files += $filespec
					}

					# For debugging
					# Write-Host ("urls cnt = " + $urls.Count)
					# Write-Host ("urls = `n" + ($urls | Out-String))
					# Write-Host ("files = `n" + ($files | Out-String))

					# CALL LIST OF URLS ASYNCHRONOUSLY
					# REF: https://stackoverflow.com/a/31525044/13499984

					$maxConcurrentJobs = $gif_images_cnt
					$content = $urls

					# Create a RUNSPACE POOL where $maxConcurrentJobs is the 
					# maximum number of runspaces allowed to run concurrently    
					$RunspacePool = [runspacefactory]::CreateRunspacePool(1,$maxConcurrentJobs)

					# Open the runspace pool (very important)
					$RunspacePool.Open()

					# Provide feedback for a noticeably slow process
					if ($script:FromConsole -and !$script:Quiet) {
						Write-Host "Extracting JPGs from video clip file..."
					}

					# Initialize collectors
					$ps_instances = @()  
					$jobs = New-Object System.Collections.ArrayList
					
					$j = 0
					foreach ($url in $content) {
						
						$file = $files[$j]

						# For debugging
						# Write-Host "url = $url"
						#$bg_msg = "$fn_name()`nfile = $file"
						PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
						
						# Create a new PowerShell instance
						# REF: https://devblogs.microsoft.com/scripting/beginning-use-of-powershell-runspaces
						$ps_instance = [powershell]::Create()
						
						# Add the instance to a collector (to facilitate disposal later)
						$ps_instances += $ps_instance

						# Tell the instance to execute in our runspace pool
						$ps_instance.RunspacePool = $RunspacePool

						# Attach some code to the PS instance
						# ... "Invoke-RestMethod is perfect for quick APIs that have no special response information such as Headers or Status Codes, 
						#     whereas Invoke-WebRequest gives you full access to the Response object and all the details it provides."
						#$ps_instance.AddCommand("Invoke-WebRequest").AddParameter("UseBasicParsing",$true).AddParameter("Uri",$url).AddParameter("OutFile",$file) | Out-Null
						# NOTE: w/o Out-Null, the next line injects "Invoke-RestMethod" objects to the output stream
						$ps_instance.AddCommand("Invoke-RestMethod").AddParameter("UseBasicParsing",$true).AddParameter("Uri",$url).AddParameter("OutFile",$file) | Out-Null
						
						# BEGIN EXECUTION ASYNCHRONOUSLY
						#$ps_instance.BeginInvoke()
						#$ps_instance.BeginInvoke() | Out-Null
						$handle = $ps_instance.BeginInvoke()

						# For debugging
						#Write-Host ("handle = " + ($handle | ConvertTo-Json))

						# Fill collector
						$job_obj = [PSCustomObject]@{
							instance = $null
							data = $null
						}
						$job_obj.instance = $ps_instance
						$job_obj.data = $handle
						# NOTE: w/o Out-Null, the next line injects an integer to the output stream
						$jobs.Add($job_obj) | Out-Null  
		
						# Give feedback on progress
						if ($script:FromConsole -and !$script:Quiet) {
							#Write-Host ("Initiated request for {0}" -f $url)
							Write-Host "Initiated request for $url"
							#Write-Host ("Initiated request for $url`n" + ($handle | ConvertTo-Json))
							#Write-Host ("handle = " + ($handle | ConvertTo-Json -Depth 16))
							#Write-Host ("Available Runspaces in RunspacePool: {0}" -f $RunspacePool.GetAvailableRunspaces())
							<#Write-Host ("Remaining Jobs: {0}" -f @($jobs | Where {
								$_.data.iscompleted -ne 'Completed'
							}).Count) #>
						}
						
						# For debugging
						#$bg_msg = ("$fn_name()`nInitiated request for $url`n" + ($handle | ConvertTo-Json))
						#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
						
						$j++
					}

					# For debugging
					#Write-Host ("Jobs: " + ($jobs | ConvertTo-Json))

					# WAIT UNTIL ALL JPG FILES CREATED
					foreach ($jpg_filespec in $files) {
						$iter = 0
						$max_tries = 20
						$success_flg = $true
						$sleep_ms = 25
						do { 

							# Check status of jobs in runspace pool
							<#Write-Host ("Remaining Jobs: {0}" -f @($jobs | Where {
								$_.data.iscompleted -ne 'Completed'
							}).Count) #>

							if (-not (Test-Path $jpg_filespec)) {
								Start-Sleep -Milliseconds $sleep_ms
								$iter++
							}
							if ($iter -gt $max_tries) {
								$success_flg = $false
								break
							}
						} until (Test-Path $jpg_filespec)
					}
				
					# Feedback async status
					if ($success_flg) {
						$ascync_msg = "SUCCESS: $gif_images_cnt asynchronous web requests; Final JPG ready after $iter $sleep_ms-msec waits"
						if ($script:FromConsole -and !$script:Quiet) {
							Write-Host $ascync_msg
						}
					} else {
						$ascync_msg = "FAIL: $gif_images_cnt asynchronous web requests; JPG NOT ready after $iter $sleep_ms-msec waits ($jpg_filespec)"
					}			
					# For debugging
					#PopupBackgroundMsg -Seconds 30 -Message $ascync_msg -UserName $server_username -InBackground $bg_msgs_on
					
					#Write-Host (“Available Runspaces in RunspacePool: {0}” -f $RunspacePool.GetAvailableRunspaces())

					# Perform a proper cleanup to avoid possible resources issues / memory leaks
					<# $return = $job_obj | ForEach {
						$_.instance.EndInvoke($_.handle)
						$_.instance.Dispose()
					}
					$job_obj.clear() #>
					foreach ($ps_instance in $ps_instances) {
						$ps_instance.Dispose()  # Release resources
					}
					$ps_instances.Clear()

				}

			}
					
			# ==============================================================
			# CREATE THE GIF FILE

			# But first, try to get the JPG file dimensions
			# ... by checking each file as needed, stopping after the first success
			$gif_jpg_WxH = "NA"  # Default
			$needle = $RootPath + "\" + $al_cam + "\img*.jpg"
			$files = Get-ChildItem -Path "$needle"
			foreach ($jpg_file in $files) {
				# Write-Host "jpg_file = $jpg_file"
				$img = [System.Drawing.Image]::FromFile($jpg_file); 
				# Write-Host ("img = " + ($img | ConvertTo-Json))
				$gif_jpg_WxH = "$($img.Width)" + "x" + "$($img.Height)"  # expecting result like '568x320'
				$img.Dispose()  # Release resources
				# Quit once we have a good string
				# ... want to avoid 'x' string when $img object is null, or malformed
				$arr = $gif_jpg_WxH.Split('x')  # creates 2-element array, even if $gif_jpg_WxH = 'x'
				if ($arr[0].length -ne 0 -and $arr[1].length -ne 0) {
					break
				}
			}
			
			# For debugging
			# Write-Host "needle = $needle"
			# Write-Host ("files.FullName = " + ($files.FullName | ConvertTo-Json))
			# Write-Host "gif_jpg_WxH = $gif_jpg_WxH"

			$input_fs  = $RootPath + "\" + $al_cam + "\img%02d.jpg"
			$output_fs = $RootPath + "\" + $al_cam + "\output.gif"
			$jpg_file_idx = $gif_images_cnt
			$i = 0
			do { 
			
				# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
				# This code section is executed only if the initial GIF file is too large
				# for the Pushover API (2.5 MB)
				# --------------------------------------------------------------------------------
				# TRY REMOVING ONE JPG AT A TIME to ensure GIF is below maximum size 
				# accepted by the Pushover API (2.5MB)  ... https://pushover.net/api#attachments
				if ($i -gt 0) {
					$jpg_filename_sfx = ([String]$i).PadLeft(2,'0')	
					$jpg_remove_fs = $RootPath + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"
					if (Test-Path -LiteralPath "$jpg_remove_fs") {
						Remove-Item -LiteralPath "$jpg_remove_fs"
						$gif_images_cnt--
					} else {
						# Give up
						break
					}
				}
				# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
				
				if ($script:FromConsole -and !$script:Quiet) {
					Write-Host "`nExecuting ffmpeg:"
					Write-Host "------------------"
				}

				# For debugging
				#Write-Host "$fn_name() output_fs = '$output_fs'"
				#$bg_msg = "$fn_name()`noutput_fs = $output_fs"
				#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
			
				# MAKE THE GIF
				# optional arguments: -nostats
				ffmpeg `
					-hide_banner `
					-loglevel error `
					-y `
					-f image2 `
					-framerate $FPS `
					-i $input_fs `
					$output_fs
				

				# FFMPEG RESPONSES ENCOUNTERED (SO FAR)
				#   1. [mjpeg @ xxxxxxxx] overread 8
				#       "This is an error that is thrown when ffmpeg tries to read a JPEG image (or a frame from an MJPEG stream) 
				#      block-by-block, but there are still leftover bits."
				#      REF:  https://superuser.com/a/559996
				#   2. "could not open file: I/O Error"
				#      REF:  TBD
				#   3. "Error demuxing input file 0: I/O Error..."
				#      REF:  TBD
				#   4. blah, blah.
				#      REF:  TBD		
				
				if (Test-Path -LiteralPath "$output_fs") {
					if ($script:FromConsole -and !$script:Quiet) {
						Write-Host "success"
					}
				} else {
					$err_msg = "ERR/ffmpeg could not create the GIF file"
					return $err_msg
				}

				# GET SIZE OF GIF FILE
				# ... DEPRECATED: waiting for the file to be created appears to be unnecessary
				<# $gif_kb = 0
				$iter = 0
				$max_tries = 20
				$success_flg = $true
				do { 
					if (-not (Test-Path $output_fs)) {
						Start-Sleep -Milliseconds 50
						$iter++
					}
					if ($iter -gt $max_tries) {
						$success_flg = $false
						break
					}
				} until (Test-Path $output_fs) #>
				
				if (Test-Path -LiteralPath "$output_fs") {
					$sizeKB = (Get-Item -Path $output_fs).Length/1kB
				} else {
					$sizeKB = 'NA'
				}
				
				if ($NoPushFlg) {
					# No need to limit GIF size if we're not using the Pushover API
					break
				}
				
				# Reduce the image count, but not below 1
				if ($i -gt 0) {
					$ImagesCnt--
					if ($ImagesCnt -lt 2) {
						break
					}
				}
				
				$i++
				
			} until ($sizeKB -lt 2500)
			
			if ($sizeKB -ne 'NA') {
				$gif_kb = [double]([Math]::Round($sizeKB,1).ToString("#.0"))
			}
			
			# ARCHIVE THE GIF FILE
			if (!$NoArchiveGifFlg) {
				if ($LiveFlg) {
					$gif_archive_fn = $archive_gif_path			
				} else {
					$gif_archive_fn = (Split-Path $al_file -leaf).Replace(".jpg",".gif")
				}
				$gif_archive_fs = $gif_archive_path + "\" + $gif_archive_fn			
				if (!(Test-Path -LiteralPath "$gif_archive_fs")) {
					Copy-Item -path $output_fs -destination $gif_archive_fs
					$archive_gif_fs = $gif_archive_fs
				} else {
					# Save multiple copies, like 'filename(N).gif'
					$archive_gif_fs = `
						CopyFileAsSequence `
						-SourceFilespec $output_fs `
						-TargetFilespec $gif_archive_fs `
						-DestPath $gif_archive_path
				}
				
				# For debugging
				#Write-Host "gif_archive_fn = $gif_archive_fn"
				#Write-Host "gif_archive_path = $gif_archive_path"
				#Write-Host "gif_archive_fs = $gif_archive_fs"
			}

			# PERFORM MAINTENANCE ON FILE ARCHIVE FOLDERS
			if ($script:FromConsole -and !$script:Quiet) {
				Write-Host "`nGIF Archive Maintenance Summary:"
				Write-Host   "--------------------------------"
			}
			# ... GIF ARCHIVE FOLDER
			if ($gif_archive_lifetime_hours -gt 0) {
				
				$gif_purge_filespec = $gif_archive_path  + "\*.gif"
				
				# like, @(found_files_cnt, purged_files_cnt, oldest_file_hours, oldest_file_path)
				$arr = `
					PurgeOldFiles `
					-Pattern $gif_purge_filespec `
					-Hours $gif_archive_lifetime_hours
					
				# For debugging
				#Write-Host "gif_purge_filespec = '$gif_purge_filespec'"
				#Write-Host ("GIF purge arr = " + ($arr | ConvertTo-Json))
				
				if ($script:FromConsole -and !$script:Quiet) {
					Write-Host ("Count of archived GIF files matching '$gif_purge_filespec' = " + $arr[0])
					if ($arr[1] -gt 0) {			
						Write-Host ("Count of archived GIF files purged (age >= $gif_archive_lifetime_hours hrs) = " + $arr[1])
						Write-Host ("Count of remaining archived GIF files = " + ($arr[0] - $arr[1]))
					} else {
						Write-Host "No old GIF files found to purge (age >= $gif_archive_lifetime_hours hrs)"
						Write-Host ("Oldest GIF file age is " + $arr[2] + " hrs '" + $arr[3] + "'")
					}
				}
			}
			
		} else {
			# TRY AGAIN - UP TO 3X
			if ($i -lt 3) {
				Start-Sleep -Milliseconds 1000
				Continue
			} else {
				return "ERR/unable to get JSON 'clipstats' for $alert_rec"
			}				
		}
		
		# Construct object to be returned
		if ($gif_method -match 'viaDAT') {
			$source_fs = $dat_file
		} else {  # 'viaLive' or 'viaBVR'
			$source_fs = $bvr_filespec
		}
		
		# For debugging
		#Write-Host "$fn_name() gif_method = '$gif_method'"
		#Write-Host "$fn_name() source_fs = '$source_fs'"
		#Write-Host "$fn_name() dat_jpgs_cnt = $dat_jpgs_cnt"
		
		$rtn_obj = [ordered]@{}
		$rtn_obj.Add( 'method', $gif_method )
		$rtn_obj.Add( 'file', $output_fs )
		$rtn_obj.Add( 'delay', $BvrDelayMsec )
		$rtn_obj.Add( 'cnt', $gif_images_cnt )
		$rtn_obj.Add( 'datjpgs', $dat_jpgs_cnt )
		$rtn_obj.Add( 'height', $Height )
		$rtn_obj.Add( 'fps', $FPS )
		$rtn_obj.Add( 'kb', $gif_kb )
		$rtn_obj.Add( 'msec', $gif_dur_ms )
		$rtn_obj.Add( 'trimbeg', $TrimBegPct )
		# A custom TrimEnd takes precedence
		if ($TrimEndCustPct -gt 0) {
			$rtn_obj.Add( 'trimend', 0 )
		} else {
			$rtn_obj.Add( 'trimend', $trim_end_pct )
		}
		$rtn_obj.Add( 'trimcust', $TrimEndCustPct )
		$rtn_obj.Add( 'almsec', $al_dur_ms )
		$rtn_obj.Add( 'res', $gif_jpg_WxH )  # WxH
		$rtn_obj.Add( 'cam', $al_cam )
		$rtn_obj.Add( 'memo', $al_memo )
		$rtn_obj.Add( 'source', $source_fs )
		$rtn_obj.Add( 'archive', $archive_gif_fs )
		$rtn_obj.Add( 'dblocator', $clip_db_locator )
		
		# For debugging
		#Write-Host ("$fn_name() rtn_obj cnt= " + ($rtn_obj.Count))
		#Write-Host ("$fn_name() rtn_obj = " + ($rtn_obj | ConvertTo-Json))
		
		return $rtn_obj
	}

	return "ERR/unable to make GIF"
}

function SendPushover() {
# PURPOSE: Send custom notification via the Pushover API
# ARGS:    $MsgStr          = (string) Pushover notification message; may contain HTML
#          $MsgTitle        = (string) Pushover notification title
#          $PriorityLevel   = (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
#          $Sound           = (string) Pushover sound
#          $AttachPath      = (string) full filespec of resized alert image, or $null
#          $Camera          = (string) Blue Iris camera shortname (replaces '&CAM' in user-setting $po_tags)
#          $QuietFlg        = (boolean) $true prevents background messaging (showing the message sent)
# RETURNS: (string) Pushover API response, or 'ERR/' message
# ATTRIBUTIONS:
#    handling attachments: https://tinyurl.com/mrxp7pn3 (PoshOver Powershell Package)
#
# V4 (2024-07-12)
# V5 (2024-09-21) added -Sound argument
# By jaydeel @ ipcamtalk.com
	
    param(
		[String]$MsgStr,      # [OPTIONAL] default is 'test'
		[String]$MsgTitle,    # [OPTIONAL] default is the App name associaled with your Pushover App token
		[Int]$PriorityLevel,  # [OPTIONAL] default is 0
		[String]$Sound,       # [OPTIONAL] default is ""
		[String]$AttachPath,  # [OPTIONAL] default is null
		[String]$Camera,      # [OPTIONAL] default is ""
		[Boolean]$QuietFlg    # [OPTIONAL] default is $false
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered function: $fn_name()"
	# Write-Host "$fn_name() MsgStr = '$MsgStr'"
	# Write-Host "$fn_name() MsgTitle = '$MsgTitle'"
	# Write-Host "$fn_name() PriorityLevel = $PriorityLevel"
	# Write-Host "$fn_name() Sound = $Sound"
	# Write-Host "$fn_name() AttachPath = '$AttachPath'"	
	# Write-Host "$fn_name() Camera = '$Camera'"	
	# Write-Host "$fn_name() QuietFlg = $QuietFlg"
	# NOTE REGARDING MSG.EXE COMMANDS
	# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
	#msg.exe $script:server_username /TIME:20 "Pushover notification args...`npriority_level = $PriorityLevel`nQuietFlg = $QuietFlg`nAttachPath = $AttachPath`nMsgStr = $MsgStr"

	# HANDLE OPTIONAL COMMANDLINE ARGS
	if (!$MsgStr) {
		$MsgStr = "Alert w/ GIF attachment demo"
	}
	if (!$Camera) {
		$cam = ''
	} else {
		$cam = $Camera
	}
	if (!$Sound) {
		$Sound = ""
	}

	$rtn_str = $null
	
	#Prep the attachment
	if ($AttachPath) {
		#Get-Item -Path $AttachPath
		$attach_obj = (Get-Item -Path $AttachPath)
		$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
	}

	# For debugging
	#Write-Host ("$fn_name() attach_obj = " + $attach_obj)
	#Write-Host ("$fn_name() attach_obj.Name = " + $attach_obj.Name)
	#Write-Host ("$fn_name() po_tags = '" + $script:po_tags + "'")
	#Write-Host ("$fn_name() Camera = " + $Camera)
	
	# PREP CORE PARAMS
	# ... will be modified if sending an attachment
	$uri = 'https://api.pushover.net/1/messages.json'
	if ($script:po_tags) {
		$po_tags = ($script:po_tags).Replace("&CAM", $cam)
	} else {
		$po_tags = ''		
	}
	$parameters = @{
		token      = $script:po_app_token
		user       = $script:po_user_token
		priority   = $PriorityLevel
		retry      = $script:po_retry
		expire     = $script:po_expire
		tags       = $po_tags
		sound      = $Sound
		device     = $script:po_device
		html       = 1
		title      = $MsgTitle
		message    = $MsgStr
	}
	
	# For debugging
	#Write-Host ("parameters.Keys = " + $parameters.Keys)
	#Write-Host "`nPushover parameters..."
	<# [string[]]$arr = $parameters | Out-String -stream
	foreach ($item in $arr) {
		if ($item) {
			Write-Host $item
		}
	} #>
	
	# EXECUTE
	if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {
		
		# SEND BASIC (OTEXT-ONLY) NOTIFICATION

		# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
		#	  do not execute; it doesn't produce anything, no error, warning, or output.
		#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
		#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
		#     all web requests now use basic parsing only.
		# ... attachmnets
		#     https://psg-prod-centralus.cloudapp.net/packages/Poshover/0.1.450600-alpha/Content/Private%5CSend-MessageWithAttachment.ps1
		try {
			$ApiResponse = $parameters | Invoke-RestMethod -Uri $uri -Method Post
			# Alternative syntaxes
			#$parameters | Invoke-RestMethod -Uri $uri -Method Post | Out-Null
			#$ApiResponse = Invoke-RestMethod -Uri $uri -Method Post -Form $parameters
		}
		catch {
			if ($script:FromConsole) {
				Write-Warning "An error was encountered sending the Pushover text message:"
				Write-Error $_
			}
			$rtn_str = "ERR/an error was was encountered sending the Pushover text message"
			return $rtn_str
		}
			
		# Expecting string like "@{status=1; request=21d4f100-3127-4c62-aad9-27058400fd74}"
		$response_str = ($ApiResponse | ConvertTo-Json -Depth 100) | ConvertFrom-Json
				
		# For debugging
		#Write-Host "$fn_name() w/o attachment ApiResponse = $response_str"

	} else {	
	
		# SEND NOTIFICATION WITH ATTACHMENT
		
		# Notes:
		# Modified code from:  https://tinyurl.com/mrxp7pn3
		# More refs: https://gist.github.com/markekraus/a0135d0a5f52b609eda513df4db71cab

		try {

            Add-Type -AssemblyName System.Net.Http
			# NOTE: Originally, the $client= line below was not working when this script was 
			#        executed in the background via a Blue Iris 'run a program' action
			#        SOLUTION:  https://stackoverflow.com/a/70535083
			$client = [system.net.http.httpclient]::new()

			try {
				$content = [system.net.http.multipartformdatacontent]::new()
				foreach ($key in $parameters.Keys) {
					$textContent = [system.net.http.stringcontent]::new($parameters.$key)
					$content.Add($textContent, $key)
				}
				
				# For debugging
				#Write-Host ("content = " + $content | ConvertTo-Json)
				
				$jpegContent = [system.net.http.bytearraycontent]::new($attach_bytes)
				$jpegContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
				$jpegContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
				$jpegContent.Headers.ContentDisposition.Name = 'attachment'
				$jpegContent.Headers.ContentDisposition.FileName = $AttachPath
				$content.Add($jpegContent)

				#For debugging
				# Write-Host ("content = " + ($content | ConvertTo-Json))

				# Not needed.
				# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
				#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

				# Modify the message body
				Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.token, "********").Replace($parameters.user, "********"))"

				# For debugging
				#msg.exe $script:server_username /TIME:10 "executing pushover api cmd w/ attachment..."

				# DO IT!
				$ApiResponse = $client.PostAsync($uri, $content).Result
				
				# For debugging
				#Write-Host ("$fn_name() w/ attachment ApiResponse = " + ($ApiResponse | ConvertTo-Json -Depth 32))
				
				try {
					$response_str = $ApiResponse.Content.ReadAsStringAsync().Result | ConvertFrom-Json
					
					# For debugging
					#Write-Host "$fn_name() ApiResponse.Content = $response_str"
				}
				catch {
					$response_str = "WARNING: Unexpected Pushover API response; attachment may be too large"
				}
				
				# For debugging
				#Write-Host "$fn_name() w attachment ApiResponse = $response_str"
				
			}
			catch {
				$response_str = "ERR/unable to send Pushover notification"				
			}
			finally {
				$content.Dispose()  # Release resources
			}
		}
		finally {
			$client.Dispose()  # Release resources
		}
	}

	# For testing - Simulate a bad pushover response
	#$response_str = 'invalid'
	#$response_str = 'status=0'
	
	if ($response_str -match 'invalid') {
		$rtn_str = "ERR/Pushover API returned 'invalid'"
	} elseif (!($response_str -match 'status=1')) {
		$rtn_str = "ERR/failed to send Pushover notification"				
	} else {
		# SUCCESS!!
		$rtn_str = ("Pushover notification sent`nAPI response = " + $response_str)
	}
	
	return $rtn_str
}

function SendTelegram() {
# PURPOSE: Send custom notification via the Telegram Bot API
# ARGS:    $MsgStr          = (string) pushover notification message; may contain HTML
#          $PriorityLevel   = CURRENTLY UNSPPORTED
#          $AttachPath      = (string) full filespec of resized alert image, or $null
#          $Camera          = (string) Blue Iris camera shortname (replaces '&CAM' in user-setting $po_tags)
#          $QuietFlg        = (boolean) $true prevents background messaging (showing the message sent)
# RETURNS: (string) Telegram API response, or 'ERR/' message
# GLOBALS: Uses global vars: $FromConsole, $Quiet
#          $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
#          $script:Quiet       = (boolean) $true suppresses verbose output to the console
# ATTRIBUTIONS:
#    handling attachments: https://tinyurl.com/mrxp7pn3 (PoshOver Powershell Package)
#
# V1 (2024-07-12)
# By jaydeel @ ipcamtalk.com
	
    param(
		[String]$MsgStr,          # [OPTIONAL] default is 'test'
		[String]$AttachPath,      # [OPTIONAL] default is null
		[String]$Camera,          # [OPTIONAL] default is ""
		[Boolean]$QuietFlg        # [OPTIONAL] default is $false
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered function: $fn_name()"
	# Write-Host "$fn_name() MsgStr = '$MsgStr'"
	# Write-Host "$fn_name() AttachPath = '$AttachPath'"	
	# Write-Host "$fn_name() Camera = '$Camera'"	
	# Write-Host ("$fn_name() FromConsole = " + $script:FromConsoleFlg)
	# Write-Host ("$fn_name() Quiet = " + $script:Quiet)
	# NOTE REGARDING MSG.EXE COMMANDS
	# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
	#msg.exe $script:server_username /TIME:20 "Pushover notification args...``nQuietFlg = $QuietFlg`nAttachPath = $AttachPath`nMsgStr = $MsgStr"

	# HANDLE OPTIONAL COMMANDLINE ARGS
	if (!$MsgStr) {
		$MsgStr = "Alert w/ GIF attachment demo"
	}
	if (!$Camera) {
		$cam = ''
	} else {
		$cam = $Camera
	}

	$PsVersion = $PSVersionTable.PSVersion.Major

	$rtn_str = $null
	
	#Prep the attachment
	if ($AttachPath) {
		try {
			$attach_obj = Get-Item -Path $AttachPath -ErrorAction Stop
		}
		catch {
			if ($script:FromConsole) {
				Write-Warning "The specified attachment file could not be interpreted properly"
			}
			$rtn_str = $false
			return $rtn_flg
		}
		$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
	}

	# For debugging
	# Write-Host ("$fn_name() attach_obj = " + $attach_obj)
	# Write-Host ("$fn_name() attach_obj.Name = " + $attach_obj.Name)
	# Write-Host ("$fn_name() cam = " + $cam)
	
	# PREP CORE PARAMS
	# ... will be modified if sending an attachment
	$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendMessage"

	$parameters = @{
		chat_id     = $script:tg_chat_id
		parse_mode  = 'html'
	}
	
	# EXECUTE
	if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {

		# SEND BASIC NOTIFICATION
		
		$parameters.Add('text', $MsgStr)
	
		# For debugging
		#Write-Host ("parameters.Keys = " + $parameters.Keys)
		<# Write-Host "`nPushover parameters..."
		[string[]]$arr = $parameters | Out-String -stream
		foreach ($item in $arr) {
			if ($item) {
				Write-Host $item
			}
		} #>

		# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
		#	  do not execute; it doesn't produce anything, no error, warning, or output.
		#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
		#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
		#     all web requests now use basic parsing only.

		try {
			if ($PsVersion -gt 5) {
				$ApiResponse = Invoke-RestMethod -Uri $uri -Method Post -Form $parameters -ErrorAction Stop
			} else {
				$ApiResponse = Invoke-RestMethod -Uri $uri -Method Post -Body $parameters -ErrorAction Stop
			}
		}
		catch {
			if ($script:FromConsole) {
				Write-Warning "An error was encountered sending the Telegram text message:"
				Write-Error $_
			}
			$rtn_str = "ERR/an error was was encountered sending the Telegram text message"
			return $rtn_str
		}
		
		# Expecting string like "@{ok=True; result=}"
		$response_str = ($ApiResponse | ConvertTo-Json -Depth 100) | ConvertFrom-Json
				
		# For debugging
		#Write-Host "$fn_name() w/o attachment ApiResponse = $response_str"

	} else {
		
		# SEND NOTIFICATION WITH ATTACHMENT
		
		# Notes:
		# Modified code from:  https://www.powershellgallery.com/packages/PoshGram/0.8.4/Content/PoshGram.psm1

		try {
			$fileObject = Get-Item $AttachPath -ErrorAction Stop
		}
		catch {
			if ($script:FromConsole) {
				Write-Warning "The specified attachment file could not be interpreted properly"
			}
			$rtn_str = "ERR/the specified attachment file could not be interpreted properly"
			return $rtn_str
		}

		# For debugging
		# Write-Host ("$fn_name() fileObject = " + ($fileObject | ConvertTo-Json))

		if ($PsVersion -ge 7) {

			if ($AttachPath.EndsWith('.jpg')) {
				$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendPhoto"
				$parameters.Add('photo', $fileObject)
				$parameters.Add('caption', $MsgStr)
			} elseif ($AttachPath.EndsWith('.gif') -or $AttachPath.EndsWith('.mp4')) {
				$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendVideo"
				$parameters.Add('video', $fileObject)
				$parameters.Add('caption', $MsgStr)			
			} else {
				if ($script:FromConsole) {
					Write-Warning "Unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
				}
				$rtn_str = "ERR/unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
				return $rtn_str			
			}
	
			# For debugging
			# Write-Host "$fn_name() uri = $uri"
			# Write-Host ("parameters.Keys = " + $parameters.Keys)
			# Write-Host "`nTelegram parameters..."
			<# [string[]]$arr = $parameters | Out-String -stream
			foreach ($item in $arr) {
				if ($item) {
					Write-Host $item
				}
			} #>
				
			try {
				$ApiResponse = Invoke-RestMethod -Uri $uri -Method Post -Form $parameters -ErrorAction Stop				
			}
			catch {
				if ($script:FromConsole) {
					Write-Warning "An error was encountered sending the Telegram video message:"
					Write-Error $_
				}
				$rtn_str = "ERR/an error was encountered sending the Telegram video message"
				return $rtn_str			
			}
				
			# Expecting string like "@{ok=True; result=}"
			$response_str = ($ApiResponse | ConvertTo-Json -Depth 100) | ConvertFrom-Json
			# For debugging
			# Write-Host ("$fn_name() ApiResponse = " + ($ApiResponse | ConvertTo-Json -Depth 100))
			# Write-Host ("$fn_name() ApiResponse.ok = " + $ApiResponse.ok)
			# Write-Host ("$fn_name() response_str = " + $response_str)
				
		} else {

			if ($AttachPath.EndsWith('.jpg')) {
				$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendPhoto"
				$parameters.Add('caption', $MsgStr)
			} elseif ($AttachPath.EndsWith('.gif') -or $AttachPath.EndsWith('.mp4')) {
				$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendVideo"
				$parameters.Add('caption', $MsgStr)			
			} else {
				if ($script:FromConsole) {
					Write-Warning "Unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
				}
				$rtn_str = "ERR/unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
				return $rtn_str			
			}
	
			# For debugging
			# Write-Host "$fn_name() uri = $uri"
			# Write-Host ("parameters.Keys = " + $parameters.Keys)
			# Write-Host "`nTelegram parameters..."
			<# [string[]]$arr = $parameters | Out-String -stream
			foreach ($item in $arr) {
				if ($item) {
					Write-Host $item
				}
			} #>
			
			try {
				
				# This command fails in PS5
				#$rtn_flg = Invoke-RestMethod -Uri $uri -Method Post -Body $parameters -ErrorAction Stop

				#Prep the attachment
				if ($AttachPath) {
					#Get-Item -Path $AttachPath
					$attach_obj = Get-Item -Path $AttachPath
					$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
				}

				Add-Type -AssemblyName System.Net.Http
				# NOTE: Originally, the $client= line below was not working when this script was 
				#        executed in the background via a Blue Iris 'run a program' action
				#        SOLUTION:  https://stackoverflow.com/a/70535083
				$client = [system.net.http.httpclient]::new()

				try {
					$content = [system.net.http.multipartformdatacontent]::new()
					foreach ($key in $parameters.Keys) {
						$textContent = [system.net.http.stringcontent]::new($parameters.$key)
						$content.Add($textContent, $key)
					}
					
					# For debugging
					# Write-Host ("content = " + $content | ConvertTo-Json)
					
					$videoContent = [system.net.http.bytearraycontent]::new($attach_bytes)
					$videoContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
					$videoContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
					$videoContent.Headers.ContentDisposition.Name = 'video'  # Telegram API 'video' parameter
					$videoContent.Headers.ContentDisposition.FileName = $AttachPath
					$content.Add($videoContent)

					#For debugging
					# Write-Host ("content = " + ($content | ConvertTo-Json))

					# Not needed.
					# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
					#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

					# Modify the message body
					Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.chat_id, "********"))"

					# For debugging
					#msg.exe $script:server_username /TIME:10 "executing telegram api cmd w/ 'video' parameter..."

					# DO IT!
					$ApiResponse = $client.PostAsync($uri, $content).Result
					
					# For debugging
					# Write-Host ("ApiResponse w/ 'video' parameter = " + ($ApiResponse | ConvertTo-Json -Depth 100))
					
					try {
						$response_str = $ApiResponse.Content.ReadAsStringAsync().Result | ConvertFrom-Json
						
						# For debugging
						# Write-Host "ApiResponse.Content = $response_str"
					}
					catch {
						$response_str = "WARNING: Unexpected Telegram API response; 'video' parameter may be too large"
					}
					
					# For debugging
					# Write-Host "ApiResponse w/ 'video' parameter = $response_str"
					
				}
				catch {
					$response_str = "ERR/unable to send Telegram notification"				
				}
				finally {
					$content.Dispose()  # Release resources
				}
			}
			finally {
				$client.Dispose()  # Release resources
			}
		}
		
	}

	# For testing - Simulate a bad Telegram response
	#$response_str = 'ok=False'
	
	if (!($response_str -match "ok=True")) {
		if ($script:FromConsole) {
			Write-Warning "failed to send Telegram notification"
		}
		$rtn_str = "ERR/failed to send Telegram notification"				
		$rtn_flg = $false
		# Write-Error $_
	} else {
		# SUCCESS!!
		$rtn_str = ("Telegram notification sent`nAPI response = " + $response_str)
	}
	
	return $rtn_str
}

function LogScriptResultsToCsv() {
# PURPOSE:  Log performance data to CSV file.
# ARGS:     see param lines below
# RETURNS:  (string) 'success' or 'ERR/unable to append row to CSV log'
# NOTES:    1. CSV file is created automatically if it does not already exist. 
#           2. CSV file's header row is generated from the object keys in variable $new_row_obj.
# USAGE EXAMPLE:
# 	LogScriptResultsToCsv $csvfilespec $timestamp $gif_source_fs $gif_method $gif_images $gif_height $gif_fps $gif_trim_beg $gif_trim_end $gif_trim_end_cust $gif_dur_sec $gif_size $gif_jpg_wxh $makegif_ms $pusho_ms $telegram_ms $al_res $al_rec
	
    param(
		[String]$csvfilespec,   # [REQUIRED]
		[String]$timestamp,     # [REQUIRED]
		[String]$gif_source_fs,
		[String]$gif_method,
		[String]$gif_images, 
		[String]$gif_height, 
		[String]$gif_fps, 
		[String]$gif_trim_beg, 
		[String]$gif_trim_end, 
		[String]$gif_trim_end_cust, 
		[String]$gif_dur_sec, 
		[String]$gif_size, 
		[String]$gif_jpg_wxh,
		[Double]$makegif_ms,
		[Double]$pusho_ms,
		[Double]$telegram_ms,
		[Double]$total_ms,
		[String]$al_rec
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"
	#Write-Host "$fn_name() csvfilespec = $csvfilespec"
	#Write-Host "$fn_name() timestamp = $timestamp"
	#Write-Host "$fn_name() al_file = $al_file"
	#Write-Host "$fn_name() al_rec = $al_rec"
	#Write-Host "$fn_name() gif_trim_end_cust = $gif_trim_end_cust"

	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$csvfilespec) {
		$err_msg = "ERR/expecting 2 arguments (csvfilespec timestamp)"
		return $err_msg
	}
	if (!$timestamp) {
		$err_msg = "ERR/expecting 2 arguments (csvfilespec timestamp)"
		return $err_msg
	}
	
	$rtn_str = 'success'
		
	# BUILD THE NEW ROW
	# ... Note: this object's keys are the header column labels
	$new_row_obj = [PSCustomObject]@{
		'Timestamp'      = $timestamp
		'SourceFile'     = $gif_source_fs
		'GifMethod'      = $gif_method
		'GifImagesCnt'   = $gif_images
		'GifHeight'      = $gif_height
		'GifFPS'         = $gif_fps
		'GifTrimBegPct'  = $gif_trim_beg
		'GifTrimEndPct'  = $gif_trim_end
		'GifTrimCustPct' = $gif_trim_end_cust
		'GifDurSec'      = $gif_dur_sec
		'GifSizeKb'      = $gif_size
		'GifJpgWxH'      = $gif_jpg_WxH
		'GifMakeMs'      = $makegif_ms
		'PushoverMs'     = $pusho_ms
		'TelegramMs'     = $telegram_ms
		'TotalMs'        = $total_ms
		'AlertRecId'     = $al_rec
	}
	
	# For debugging
	#msg.exe $script:server_username /TIME:10 "new_row_obj... `n$new_row_obj"

	# APPEND NEW ROW TO CSV FILE, creating the file if necessary
	if ($PSVersionTable.PSVersion.Major -ge 7) {
		# REFS: https://tinyurl.com/yvy4mjy6
		#       https://stackoverflow.com/a/70717331
		#       https://stackoverflow.com/a/60681762
		try {
			$new_row_obj | Export-Csv -LiteralPath "$csvfilespec" -Append -Force -UseQuotes AsNeeded
		} catch {
			$err_msg = "ERR/unable to write to CSV log"
			return $err_msg
		}
	} else {
		try {
			$new_row_obj | Export-Csv -LiteralPath "$csvfilespec" -Append -Force -NoTypeInformation
		} catch {
			$err_msg = "ERR/unable to write to CSV log"
			return $err_msg
		}
	}
	
	if (!(Test-Path -LiteralPath "$csvfilespec")) {
		$err_msg = "ERR/unable to write to CSV log ($csvfilespec)"
		return $err_msg
	}
	
	return $rtn_str
}

function PostCustomBiLogEntry() {
# PURPOSE: Add custom message to the Blue Iris logfile
# ARGS:    $MsgStr        = (string) may contain HTML
#          $Level         = (integer) 0|1|2 (0:info, 1:warn, 2:error)
#          $QuietFlg     = (boolean) $true prevents background messaging
# RETURNS: (string) message as posted to the Blue Iris logfile
# KEY POINT: the message string maximum length is 127 characters (BI HTTP interface command '/admin?log=message&level=x')

    param(
		[String]$MsgStr,    # [REQUIRED]
		[String]$Level,     # [OPTIONAL]
		[Boolean]$QuietFlg  # [OPTIONAL]
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"
	#Write-Host "$fn_name() MsgStr = '$MsgStr'"
	#Write-Host "$fn_name() Level = $Level"
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"

	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$MsgStr) {
		$err_msg = "ERR/expecting an argument (MsgStr)"
		return $err_msg
	}
	
	# Remove/replace incompatible substrings from $MsgStr
	$MsgStr = $MsgStr -Replace "`n", "%20" -Replace "\\r",  "%20" -Replace "\\n", "%20" -Replace " ", "%20"
	
	$bi_address = $script:bi_webserver
	if ($script:post_w_creds_flg) {
		$uri = "http://$bi_address/admin?log=$MsgStr&level=$Level&user=$script:bi_user_name&pw=$script:bi_user_pw"
	} else {
		$uri = "http://$bi_address/admin?log=$MsgStr&level=$Level"
	}
	$result = Invoke-WebRequest -Uri $uri
	
	# For debugging
	#Write-Host "uri = $uri"
	#Write-Host "result:`n$result"	

	# HANDLE NOTIFICATIONS
	if (!$QuietFlg) {
		#msg.exe $script:server_username /TIME:10 "Custom BI logfile message posted...`nlevel = $Level`n'$MsgStr'"
	}
	
	return $MsgStr
}

function RegGetValue() {
# PURPOSE:  Get current value of a registry entry
# ARGS:     see param lines below
# RETURNS:  registry value, user-provided default value, or error message
# 
# V2 (2024/07/01)
# Created by jaydeel @ ipcamtalk.com

	param(
		[String]$KeyPath,  # like, 'HKLM:SOFTWARE\Perspective Software\Blue Iris\clips'
		[String]$KeyName,  # like, 'compact'
		$DefaultValue      # value to return if $KeyName does not exist 
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() -KeyPath = '$KeyPath'"
	#Write-Host "$fn_name() -KeyName = '$KeyName'"
	#Write-Host "$fn_name() -DefaultValue = '$DefaultValue'"


	try {
		$reg_obj = Get-ItemProperty -Path $KeyPath -Name $KeyName -ErrorAction Stop
	
		# For debugging
		#Write-Host ("reg_obj = " + ($reg_obj | ConvertTo-Json -Depth 100))
		
		if ($reg_obj) {
			if (!($null -eq $reg_obj.$KeyName)) {
				return $reg_obj.$KeyName
			}
		}
	}
	catch {
		# continue
	}	

	if ($DefaultValue) {
		$rtn_value = $DefaultValue
	} else {
		$err_msg = "ERR/unable to get registry key '$KeyPath\$KeyName'"
		$rtn_value = $err_msg
	}
	
	return $rtn_value
}

function IsCamRecContinuous() {
# PURPOSE:  Get video recording mode registry entry for for specified camera and profile
#           like, HKLM\SOFTWARE\Perspective Software\Blue Iris\Cameras\DriveWay-1\Clips\5\continuous
# ARGS:     see param lines below
# RETURNS:  boolean, or error message string like "ERR/..."
# REQUIRES: RegGetValue()
# 
# V1 (9/7/2024)
# Created by jaydeel @ ipcamtalk.com

	param(
		[String]$CamLongName,   # Blue Iris camera longname
		[String]$CamProfileNo   # Blue Iris camera profile (1-7); caller should validate
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() -CamLongName = '$CamLongName'"
	#Write-Host "$fn_name() -CamProfileNo = $CamProfileNo"

	if (!$CamLongName) {
		$err_msg = "ERR/$fn_name() requires camera shortname and longname"
		return $err_msg
	}

	# GET CAMERA's 'Pre-trigger playback time'
	# ... From the Camera settings > Record tab
	$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\" + $CamLongName + "\Clips\" + $CamProfileNo
	$key_name = "continuous"

	# Get regkey's  current setting's value
	# ... returns value, $null, or string 'ERR/*'
	$reg_get_value = `
		RegGetValue `
		-KeyPath $key_path `
		-KeyName $key_name
		
	# For debugging
	#Write-Host "reg_get_value = $reg_get_value"
	
	if ($null -eq $reg_get_value) {
		$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
		return $err_msg
	} else {
		if ($reg_get_value.GetType() -match 'string') {
			if ($reg_get_value.StartsWith('ERR/')) {
				return $reg_get_value
			}
		}
	}

	# Expecting: 0=When triggered, 2=Continuous, 2=Off, 3=Periodic, each, 4=Triggered+periodic: each, 5=Continuous+Alerts, 6=When alerted, 7=Continuous+Triggered)
	$val = [Int]$reg_get_value
	$rtn_flg = (@(2,5,7) -contains $val)

	# For debugging
	#Write-Host "rtn_val = $rtn_val"
	
	return $rtn_flg
	
}

function GetPreTrigRecMsec() {
# PURPOSE:  Get Pre-trigger playback time 'playback' registry entry for specified camera and profile
#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\FrontEntry\Clips\1\playtime
# ARGS:     see param lines below
# RETURNS:  value, or error message string like "ERR/..."
# REQUIRES: RegGetValue()
# 
# V1 (4/23/2024)
# Created by jaydeel @ ipcamtalk.com

	param(
		[String]$CamLongName,   # Blue Iris camera longname
		[String]$CamProfileNo   # Blue Iris camera profile (1-7); caller should validate
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() -CamLongName = '$CamLongName'"
	#Write-Host "$fn_name() -CamProfileNo = $CamProfileNo"

	if (!$CamLongName) {
		$err_msg = "ERR/$fn_name() requires camera shortname and longname"
		return $err_msg
	}

	# GET CAMERA's 'Pre-trigger playback time'
	# ... From the Camera settings > Record tab
	$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\" + $CamLongName + "\Clips\" + $CamProfileNo
	$key_name = "rectime"

	# Get regkey's  current setting's value
	# ... returns value, $null, or string 'ERR/*'
	$reg_get_value = `
		RegGetValue `
		-KeyPath $key_path `
		-KeyName $key_name
		
	# For debugging
	#Write-Host "reg_get_value = $reg_get_value"
	
	if ($null -eq $reg_get_value) {
		$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
		return $err_msg
	} else {
		if ($reg_get_value.GetType() -match 'string') {
			if ($reg_get_value.StartsWith('ERR/')) {
				return $reg_get_value
			}
		}
	}

	$playtime = $reg_get_value
	$rtn_val = 1000 * ($playtime / 10 )

	# For debugging
	#Write-Host "rtn_val = $rtn_val"
	
	return $rtn_val
	
}

function Convert-FromUnixDate() {
# Returns string like '02/01/2024 21:34:00'

	param (
		[Int]$UnixDate,
		[String]$FormatStr  # like, "yyyy-MM-dd"
	)

	if ($FormatStr) {
		$rtn_str = [timezone]::CurrentTimeZone.ToLocalTime(([datetime]'1/1/1970').AddSeconds($UnixDate)) | Get-Date -f $FormatStr
	} else {
		$rtn_str = [timezone]::CurrentTimeZone.ToLocalTime(([datetime]'1/1/1970').AddSeconds($UnixDate))
	}
   
   return $rtn_str
   
}

<# function SleepWithStatusBar() {
# Source: https://gist.github.com/ctigeek/bd637eeaeeb71c5b17f4

	param (
		[double]$Seconds,
		[String]$Activity,
		[String]$Message
	)
	
    $doneDT = (Get-Date).AddSeconds($Seconds)
    while($doneDT -gt (Get-Date)) {
        $secondsLeft = $doneDT.Subtract((Get-Date)).TotalSeconds
        $percent = ($Seconds - $secondsLeft) / $Seconds * 100
 		Write-Progress -Activity $Activity -Status $Message -SecondsRemaining $secondsLeft -PercentComplete $percent
        [System.Threading.Thread]::Sleep(500)
    }
    Write-Progress -Activity $Activity -Status $Message -SecondsRemaining 0 -Completed
} #>

<# function Convert-ByteArrayToHexString() {
# SOURCE: https://www.sans.org/blog/powershell-byte-array-and-hex-functions/
################################################################
#.Synopsis
# Returns a hex representation of a System.Byte[] array as
# one or more strings. Hex format can be changed.
#.Parameter ByteArray
# System.Byte[] array of bytes to put into the file. If you
# pipe this array in, you must pipe the [Ref] to the array.
# Also accepts a single Byte object instead of Byte[].
#.Parameter Width
# Number of hex characters per line of output.
#.Parameter Delimiter
# How each pair of hex characters (each byte of input) will be
# delimited from the next pair in the output. The default
# looks like "0x41,0xFF,0xB9" but you could specify "\x" if
# you want the output like "\x41\xFF\xB9" instead. You do
# not have to worry about an extra comma, semicolon, colon
# or tab appearing before each line of output. The default
# value is ",0x".
#.Parameter Prepend
# An optional string you can prepend to each line of hex
# output, perhaps like '$x += ' to paste into another
# script, hence the single quotes.
#.Parameter AddQuotes
# A switch which will enclose each line in double-quotes.
#.Example
# [Byte[]] $x = 0x41,0x42,0x43,0x44
# Convert-ByteArrayToHexString -ByteArray $x
#
# 0x41,0x42,0x43,0x44
#.Example
# [Byte[]] $x = 0x41,0x42,0x43,0x44
# Convert-ByteArrayToHexString -ByteArray $x -Width 2 -Delimiter "\x" -AddQuotes
#
# "\x41\x42"
# "\x43\x44"
################################################################

	[CmdletBinding()] Param (
	[Parameter(Mandatory = $True, ValueFromPipeline = $True)] [System.Byte[]] $ByteArray,
	[Parameter()] [Int] $Width = 10,
	[Parameter()] [String] $Delimiter = ",0x",
	[Parameter()] [String] $Prepend = "",
	[Parameter()] [Switch] $AddQuotes )
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	 
	if ($Width -lt 1) { 
		$Width = 1
	}
	if ($ByteArray.Length -eq 0) { 
		Return 
	}
	$FirstDelimiter = $Delimiter -Replace "^[\,\:\t]",""
	$From = 0
	$To = $Width - 1
	
	# For debugging
	#Write-Host ("fn_name() ByteArray cnt = " + $ByteArray.Count)
	#Write-Host ("fn_name() ByteArray[0] = " + $ByteArray[0])
	#Write-Host ("fn_name() ByteArray[1] = " + $ByteArray[1])
	#Write-Host ("fn_name() ByteArray[" + ($ByteArray.Count - 1) + "] = " + $ByteArray[$ByteArray.Count - 1])
	#Write-Host "$fn_name() From = $From, To = $To"
	#Write-Host ("fn_name() ByteArray[" + $From + "] = " + $ByteArray[$From])
	#Write-Host ("fn_name() ByteArray[" + $To + "] = " + $ByteArray[$To])
	
	Do {
		$String = [System.BitConverter]::ToString($ByteArray[$From..$To])
		
		# For debugging
		#Write-Host ("fn_name() String = " + $String)
		
		$String = $FirstDelimiter + ($String -replace "\-",$Delimiter)
		if ($AddQuotes) { 
			$String = '"' + $String + '"' 
		}
		if ($Prepend -ne "") { 
			$String = $Prepend + $String 
		}
		$String
		$From += $Width
		$To += $Width
		
		# [ADDED BY JAYDEEL] added this code to avoid error under Set-StrictMode method 
		if ($To -gt ($ByteArray.Count - 1)) {
			$To = $ByteArray.Count - 1
		}
		
	} While ($From -lt $ByteArray.Length)
} #>

# Start overall performance timer
$timer_total = [Diagnostics.Stopwatch]::StartNew()

# SHOW USAGE GUIDE
if ($Help) {
	if ($FromConsole) {
		Write-Host "================================================================================================================"
		Write-Host "BLUE IRIS GIF TOOL - MAKES GIF FILE & SENDS PUSHOVER/TELEGRAM NOTIFICATION WITH GIF ATTACHMENT"
		Write-Host "----------------------------------------------------------------------------------------------------------------"
		Write-Host "USAGE: $ScriptName -AlertRec [Args]          ... make GIF via BI database record, like '@123456' * ** [1]"
		Write-Host "       $ScriptName -DbRecNo | -UI3Rec [Args] ... make GIF from ANY EXISTING ALERT with a duration ** [2,3]"
		Write-Host "       $ScriptName -Camera -Latest [Args]    ... make GIF from camera's most recent alert ** [4]"
		Write-Host "       $ScriptName -Camera -Live             ... make GIF from camera's active video stream"
		Write-Host "       * ctrl-right-click any Alert in the Blue Iris clipslist to load its DB record to the clipboard"
		Write-Host "       ** script tries 1st to make GIF from a DAT file, then from a BVR clip if a DAT file does not exist"
		Write-Host ""
		Write-Host "ARGS:  -IgnoreDatFile ... make GIF from BVR clip instead of DAT file (DAT is the default)"
		Write-Host "       -DatImageMsec  ... camera-specific msec between DAT images (applies to DAT-derived GIFs only) [5]"
		Write-Host "       -BvrDelayMsec  ... msec to wait before extracting images (applies to BVR-derived GIFs only) [X]"
		Write-Host "       -Seconds       ... set GIF duration (default is 10 seconds); for -Live & BVR-derived GIFs only"
		Write-Host "       -ImagesCnt -Height -FPS ... change GIF's default settings (per the user-settings file)"
		Write-Host "       -TrimBegPct -TrimEndPct ... make GIF from middle (active) portion of the Alert clip's duration [6]"
		Write-Host "       -AlertMemo ... right-trim GIF if fast-moving objects detected via AI (e.g. 'vehicle:91%') [7]"
		#Write-Host "       -NukeDatFile "  # REDACTED
		Write-Host "       -PushService ... override default push application ('Pushover' or 'Telegram')"
		Write-Host "       -Msg      ... notification message (may contain HTML)"
		Write-Host "       -Priority ... Pushover notification priority: -2|-1|0|1|2 = lowest|silent|normal|high|emergency"
		Write-Host "       -Sound    ... Pushover custom sound for notification"
		Write-Host "       -NoPush -NoBiLog -NoArchive ... disable default actions (per the user-settings file) [8]"
		Write-Host "       -Demo  ... skip notification/logfile/delete actions (useful when tuning GIF params)"
		Write-Host "       -Quiet ... suppress verbose output to the Console; show the JSON return object instead"
		Write-Host "       -Help  ... show this usage guide"
		Write-Host "NOTES:"
		Write-Host "[1] [-AlertRec X] value must be prefixed with '@'; intended for use w/ Blue Iris 'On Alert' Actions [A]"
		Write-Host "[2] [-DbRecNo X] is the numeric part of the Alert's database record (must NOT be prefixed with '@') "
		Write-Host "[3] [-UI3Rec X] is the value after '&rec=' in the UI3 address bar (viewing an alert)"
		Write-Host "[4] [-Latest] may be used w/ Blue Iris 'On Alert' Actions [B]; alternative to [A]"
		Write-Host "[5] [-DatImageMsec X] improves estimation of the GIF's duration (applies to DAT-derived GIFs only)"
		Write-Host "    (for X, use value from BI > Camera settings > Alert > AI Configuration > 'Analyze one image each')"
		Write-Host "[6] [-TrimBegPct] may be negative to start GIF before the Alert's offset (BVR-derived GIFs only)"
		Write-Host "[7] [-AlertMemo string] is intended for use w/ Blue Iris 'On Alert' Actions [C]; the string is used"
		Write-Host "     to customize trimming if user-settings file variable '`$gif_trim_end_cust_obj' has been configured"
		#Write-Host "Note:  use [-NukeDatFile] to remove DAT file from Blue Iris storage (always archived first); "  # REDACTED
		#Write-Host "       ALWAYS executes a Database/Regenerate action (required per Blue Iris support)"
		Write-Host "[8] [-NoPush] disables sending a notification; ALSO disables management of GIF size to below 2.5 MB"
		Write-Host ""
		Write-Host "GUIDE TO USING THIS SCRIPT FROM A BLUE IRIS ACTION SET..."
		Write-Host "[A] use [-AlertRec \&ALERT_DB] <-- KEY POINT: a backslash (or backtick '``') prefix is ESSENTIAL"
		Write-Host "[B] use [-Camera '&CAM' -Latest]"
		Write-Host "[C] use [-AlertMemo '&MEMO']"
		Write-Host ""
		Write-Host "RETURNS: (json string) if successful, or 'ERR/*' string if failure"
		Write-Host "(json string) is like {'result: success, data: {gif_method: viaDat, gif_kb: 497.5, ...}}"
		#Write-Host "(string) is like 'ERR/description'"
		Write-Host "----------------------------------------------------------------------------------------------------------------"
		Write-Host ""
		exit
	}
}

# Initialize return object
$rtn_obj = @{}

# CHECK IF SCRIPT'S PATH IS IN ENVIRONMENT 'SYSTEM PATH' VARIABLE
if (!$experienced_user_flg) {
	if ($FromConsole) {
		$EnvPath = [Environment]::GetEnvironmentVariable('Path', 'Machine')
		if ($null -ne $EnvPath) {
			if (!($EnvPath -split ';' -contains $ScriptPath)) {

				# For debugging
				#Write-Host "ScriptPath = $ScriptPath"
				#Write-Host ("EnvPath = " + (($EnvPath -Split ';') | ConvertTo-Json))

				#return "ERR/this script's path '$ScriptPath' does not exist in the Environment System Path variable"
				Write-Host "HEADS-UP: this script's path '$ScriptPath' does not exist in the Environment System Path variable" -ForegroundColor Green
				
				# Attempt repair
				# NOTE: when I tried this it added MANY other/duplicate entries
				#[Environment]::SetEnvironmentVariable("Path", $EnvPath + ";$ScriptPath", "Machine")
				#Write-Warning "Added this script's path '$ScriptPath' to the Environment Path"
			}
		} else {
			#Write-Warning "Unable to inspect Environment variable; be sure script's path is in the Environment Path"
			Write-Host "HEADS-UP: Unable to inspect Environment variable; be sure script's path is in the Environment Path" -ForegroundColor Green
		}
		# For debugging
		#Write-Host ("EnvPath = " + (($EnvPath -Split ';') | ConvertTo-Json))
	}
}

# PARSE COMMANDLINE ARGS
if (1 -eq 1) {  # Use to collapse this code section while editing

	# PERMANENTLY DISABLED
	if ($NukeDatFile) {
		$NukeDatFile = $false
		if ($FromConsole) {
			#Write-Warning "Argument -NukeDatFile performs a DB/Regenerate action, which temporarily disables all camera recording;`nTip: it might be preferable to batch delete no-longer-needed DAT files via my 'BI_DAT_tool.ps1' script"
			Write-Warning "Argument -NukeDatFile has been deprecated (see changelog)"
		}
	}
	if ($NoArchiveDat) {
		$NoArchiveDat = $false
		if ($FromConsole) {
			Write-Warning "Argument -NoArchiveDat has been deprecated (see changelog)"
		}
	}
	if ($Camera -and !($Live -or $Latest)) {
		$err_msg = "ERR/expecting argument -Live or -Latest when using argument -Camera"
		return $err_msg
	}
	if ($Live -and !$Camera) {
		$err_msg = "ERR/expecting argument -Camera when using -Live"
		return $err_msg
	}
	if ($Latest -and !$Camera) {
		$err_msg = "ERR/expecting argument -Camera when using -Latest"
		return $err_msg
	}
	if (!$AlertRec -and !$DbRecNo -and !$UI3Rec -and !$Camera) {
		#$err_msg = "ERR/expecting an argument (-AlertRec, -AlertDB, -DbRecNo, -UI3Rec, or -Camera 'xxx' -Live)"
		$err_msg = "ERR/expecting an argument (-AlertRec, -DbRecNo, -UI3Rec, or -Camera)"
		return $err_msg
	}

	$deb_rec = $null
	$clip_db_locator = $null  # Initialize
	if ($AlertRec -or $DbRecNo -or $UI3Rec) {

		if ($AlertRec -and ($DbRecNo -or $UI3Rec)) {
			$err_msg = "ERR/do not simultaneously use -AlertRec and -DbRecNo or -UI3Rec"
			return $err_msg
		}
		if ($DbRecNo -and ($AlertRec -or $UI3Rec)) {
			$err_msg = "ERR/do not simultaneously use -DbRecNo and -AlertRec or -UI3Rec"
			return $err_msg
		}
		if ($UI3Rec -and ($AlertRec -or $DbRecNo)) {
			$err_msg = "ERR/do not simultaneously use -UI3Rec and -AlertRec or -DbRecNo"
			return $err_msg
		}
		if ($cmdline_args_passed -contains 'DbRecNo' -and $DbRecNo.Length -eq 0) {
			# This happens when using -DbRecNo @123456 (without quotes)
			$err_msg = "ERR/argument -DbRecNo must contain only digits"
			return $err_msg
		}	
		
		if ($AlertRec) {
			$clip_db_locator = $AlertRec.Replace("\","")  # Why? prefix '\' may be used via Blue Iris Action set action
			if (!$clip_db_locator.StartsWith('@')) {
				$err_msg = "ERR/argument -AlertRec is expected to be prefixed with an '@' (like '@123456')"
				return $err_msg
			}
		}
		
		if ($DbRecNo) {
			# Test if argument contains non-numeric characters
			$num_str = ($DbRecNo -replace "[^0-9]" , '').ToString()
			#Write-Host "num_str = $num_str"
			if ($num_str.Length -lt $DbRecNo.Length) {
				if ($DbRecNo.StartsWith('@')) {
					$err_msg = "ERR/argument -DbRecNo must be an integer; it should not be prefixed with '@'"
				} else {
					$err_msg = "ERR/argument -DbRecNo must contain only digits"
				}
				return $err_msg
			} else {
				$clip_db_locator = "@" + $DbRecNo.ToString()
			}
		}
		
		if ($UI3Rec) {
			$ui3_rec = $UI3Rec.Trim()
			if ($ui3_rec.StartsWith('@')) {
				$err_msg = "ERR/argument -UI3Rec should not be prefixed with '@'"
				return $err_msg
			} else {
				$str = $ui3_rec.Replace('-', '')
				# Test if argument w/o hyphen contains any non-numeric characters
				$num_str = $str -replace "[^0-9]" , ''
				#Write-Host "num_str = $num_str"
				if ($num_str.Length -lt $str.Length) {
					$err_msg = "ERR/argument -DbRecNo must contain only digits, or digits separated by a hyphen"
					return $err_msg
				} else {
					# Use string like '103356469-2981206'
					$hypen_idx = $ui3_rec.IndexOf('-')
					if ($hypen_idx) {
						$arr = $ui3_rec.Split('-')
						$clip_db_locator = "@" + ($arr[0]).ToString()
					} else {
						$clip_db_locator = "@" + $DbRec.ToString()
					}
				}			
			}
		}
	}
	# Write-Host "clip_db_locator = $clip_db_locator"

	if ($DatImageMsec) {
		if ($DatImageMsec -lt 0) {
			$err_msg = "ERR/argument '-DatImageMsec' must be greater than zero"
			return $err_msg
		}
		if ($DatImageMsec -gt 0) {
			$dat_image_interval_msec = $DatImageMsec
		}
	}

	$duration_sec = 10  # Default - for -Live & BVR-derived GIFs obly
	#if ($cmdline_args_passed -contains 'Seconds') {
	if ($Seconds) {
		if ($Seconds -gt 0) {
			$duration_sec = $Seconds
		} else {
			$err_msg = "ERR/Argument -Seconds must be greater than zero"
			return $err_msg
		}
	}

	$bvr_delay_msec = 0  # Defaults
	if ($BvrDelayMsec) {
		if ($BvrDelayMsec -lt 0) {
			$err_msg = "ERR/argument '-BvrDelayMsec' must be greater than zero"
			return $err_msg
		}
		if ($BvrDelayMsec -gt 0) {
			$bvr_delay_msec = $BvrDelayMsec
		}
	}
	# Write-Host "bvr_delay_msec = $bvr_delay_msec"
	
	if ($Height) {
		if ($Height -lt 0) {
			$err_msg = "ERR/argument '-Height'  must be greater than zero"
			return $err_msg
		}
	}
	if ($FPS) {
		if ($FPS -lt 0) {
			$err_msg = "ERR/argument '-FPS' must be greater than zero"
			return $err_msg
		}
	}
	if ($TrimBegPct) {
		if (($TrimBegPct -gt 99)) {
			$err_msg = "ERR/argument '-TrimBegPct' must less than 99"
			return $err_msg
		}
		$gif_trim_beg = $TrimBegPct
	}
	if ($TrimEndPct) {
		if (($TrimEndPct -lt 0) -or ($TrimEndPct -ge 99)) {
			$err_msg = "ERR/argument '-TrimEndPct' must be between 0 and 99"
			return $err_msg
		}
		$gif_trim_end = $TrimEndPct
	}

	# Change GIF's END TRIM if using AI label testing
	$clip_dur_adj = 0
	$cust_lbl_disp = ''
	$gif_trim_end_cust = 0	
	
	if ($AlertMemo) {
	
		$arr = $AlertMemo.Split(';')

		# For debugging
		#Write-Host ("AlertMemo arr = " + ($arr | ConvertTo-Json))

		$al_memo_ai_str = ''
		foreach ($item in $arr) {
			if ($item.IndexOf('%') -gt 0) {
				$al_memo_ai_str = $item.Trim()
				break
			}
		}

		# For debugging
		#Write-Host "al_memo_ai_str = '$al_memo_ai_str'"

		if ($al_memo_ai_str.length -gt 0) {
			
			$al_memo_ai_str_pcs = $al_memo_ai_str.Split(',')
			$j = 0
			foreach ($al_memo_ai_str_pc in $al_memo_ai_str_pcs) {
				$arr = $al_memo_ai_str_pc.Split(':')
				$al_memo_ai_str_pcs[$j] = $arr[0]
				$j++
			}
			
			# We use just the first label. It is always the highest confidence.
			$first_lbl = $al_memo_ai_str_pcs[0]

			if ($gif_trim_end_cust_obj[$first_lbl]) {
				$cust_lbl_disp = $first_lbl  # Used for error messages
				$gif_trim_end_cust = $gif_trim_end_cust_obj[$first_lbl]
			}

			# For debugging
			#Write-Host ("al_memo_ai_str_pcs = " + ($al_memo_ai_str_pcs | ConvertTo-Json))
			#Write-Host "cust_lbl_disp = $cust_lbl_disp"

			$clip_dur_adj = $TrimBegPct + $gif_trim_end_cust
			
			if ($clip_dur_adj -gt 100) {
				$err_msg = "ERR/sum of clip duration adjustments cannot exceed <100% ('TrimBegPct/TrimEndPct/TrimEndCustPct = $TrimBegPct/$TrimEndPct/$gif_trim_end_cust)"
				return $err_msg
			} 
			<# if ($clip_dur_adj -lt 0) {
				$err_msg = "ERR/sum of clip duration adjustments cannot be <0% ('TrimBegPct/TrimEndPct/TrimEndCustPct = $TrimBegPct/$TrimEndPct/$gif_trim_end_cust)"
				return $err_msg
			} #> 

		}
		
	} else {
		
		$clip_dur_adj = $TrimBegPct + $TrimEndPct
		
		if ($clip_dur_adj -gt 100) {
			$err_msg = "ERR/sum of clip duration adjustments cannot exceed <100% ('TrimBegPct/TrimEndPct = $TrimBegPct/$TrimEndPct)"
			return $err_msg
		} 
		<# if ($clip_dur_adj -lt 0) {
			$err_msg = "ERR/sum of clip duration adjustments cannot be <0% ('TrimBegPct/TrimEndPct = $TrimBegPct/$TrimEndPct)"
			return $err_msg
		} #> 
	}
	
	if ($gif_trim_end_cust -gt 0) {
		$gif_trim_end = 0
	}

	# For debugging
	#Write-Host "clip_dur_adj = $clip_dur_adj"
	#Write-Host "gif_trim_beg = $gif_trim_beg"
	#Write-Host "gif_trim_end = $gif_trim_end"
	#Write-Host "gif_trim_end_cust = $gif_trim_end_cust"

	if ($Priority -or ($Priority -eq 0)) {
		if (($Priority -ge -2) -and ($Priority -le 2)) {
			$po_priority = $Priority
		}
	}

	$gif_via_DAT_flg = $true  # Default
	if($IgnoreDatFile) {
		$gif_via_DAT_flg = $false
	}
	# Write-Host "gif_via_DAT_flg = $gif_via_DAT_flg"
	
	if ($IgnoreDatFile -and $Live) {
		$err_msg = "ERR/argument -IgnoreDatFile should not be used with argument -Live"
		return $err_msg
	}
	
	if ($ImagesCnt) {
		$gif_images_cnt = $ImagesCnt
	}
	#Write-Host "gif_images_cnt = $gif_images_cnt"
	if ($Height) {
		$gif_height = $Height
	}
	if ($FPS) {
		$gif_fps = $FPS
	}
	
	if ($PushService) {
		if (!(@("pushover", "telegram") -contains $PushService.ToLower())) {  # -contains operator is case-insensitive by default
			$err_msg = "ERR/argmument -PushService must be 'Pushover' or 'Telegram'"
			return $err_msg
		} else {
			$push_service = $PushService.ToLower()
		}
	}

	$nopush_flg = $false  # Default
	if($NoPush) {
		$nopush_flg = $true
	}

	$nobilog_flg = $false  # Default
	if ($NoBiLog) {
		$nobilog_flg = $true
	}

	$noarchivegif_flg = $false  # Default
	if ($NoArchive) {
		$noarchivegif_flg = $true
	}

	$quiet_flg = $false  # Default
	if ($Quiet) {
		$quiet_flg = $true
	}
	if ($quiet_flg) {
		$bg_msgs_on = $false
	}

	$demo_flg = $false  # Default
	if ($Demo) {
		$demo_flg = $true
	}
	
	# Simulator Args
	#$Sim1_arg_exists = ($cmdline_args_passed -contains 'Sim1')
	#$Sim2_arg_exists = ($cmdline_args_passed -contains 'Sim2')
	#$Sim3_arg_exists = ($cmdline_args_passed -contains 'Sim3')
	#$Sim4_arg_exists = ($cmdline_args_passed -contains 'Sim4')
	#$Sim5_arg_exists = ($cmdline_args_passed -contains 'Sim5')
}

# ABORT IF USER SETTINGS PATHS NOT FOUND
if (1 -eq 1) {  # Use to collapse this code section while editing
	if (!$app_path) {
		$err_msg = "ERR/setting '`$app_path' missing in user-settings file`n"
		if ($post_to_BI_logfile_flg) {
			$null = PostCustomBiLogEntry `
				-MsgStr "$err_msg ($ScriptName)" `
				-Level 2 `
				-QuietFlg ($bg_msgs_on -eq $false)
		}
		if ($FromConsole) {
			return $err_msg
		} else {
			msg.exe $server_username /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
			exit
		}	
	}
	if (!$bi_alerts_folder_path) {
		$err_msg = "ERR/setting '`$bi_alerts_folder_path' missing in user-settings file`n"
		if ($post_to_BI_logfile_flg) {
			$null = PostCustomBiLogEntry `
				-MsgStr "$err_msg ($ScriptName)" `
				-Level 2 `
				-QuietFlg ($bg_msgs_on -eq $false)
		}
		if ($FromConsole) {
			return $err_msg
		} else {
			msg.exe $server_username /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
			exit
		}	
	}
	if (!$bi_new_folder_paths) {
		if (!$bi_new_folder_path) {
			$err_msg = "ERR/setting '`$nbi_new_folder_paths' missing in user-settings file`n"
			if ($post_to_BI_logfile_flg) {
				$null = PostCustomBiLogEntry `
					-MsgStr "$err_msg ($ScriptName)" `
					-Level 2 `
					-QuietFlg ($bg_msgs_on -eq $false)
			}
			if ($FromConsole) {
				return $err_msg
			} else {
				msg.exe $server_username /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
				exit
			}	
		}
	}
}

# -----------------------------------------------------------
# MANAGE APP'S SUBFOLDERS
if (1 -eq 1) {  # Use to collapse this code section while editing
	# Note: $app_path is user-defined (_user_settings.ps1 file)

	# 1. Home folder
	if (!(Test-Path -LiteralPath "$app_path")) {
		New-Item -ItemType Directory -Force -Path $app_path | Out-Null
	}

	# 2. Create subfolder for archiving GIF files created by this script.
	$gif_archive_path = ($app_path + "\_gif_archive")  # like 'D:\pushover_gifs\_gif_archive'
	if (!(Test-Path -LiteralPath "$gif_archive_path")) {
		New-Item -ItemType Directory -Force -Path $gif_archive_path
	}
	# ... And add/refresh a README to the folder
	$readme_path = "$gif_archive_path\_README.txt"
	if (!(Test-Path -LiteralPath "$readme_path")) {
		New-Item -Path "$readme_path" | Out-Null
		$msg = "GIF Files Archive Folder" +
			"`n------------------------" +
			"`n`nWHAT? This folder archives every GIF file created by script 'BI_gif_tool.ps1'." + 
			"`n`nTIP: This folder may be managed by the script to automatically prune old files." + 
			"`n     To implement this feature, see setting '`$gif_archive_lifetime_hours' in the user-settings file" +
			"`n`nNote: This folder and this README.txt file are programmatically created."
		$msg | Out-File -FilePath "$readme_path" -Append
	}
		
	# 3. Create the Script's DevMode folder if necessary
	$DevmodePath = ($app_path + "\_devmode")  # like 'D:\pushover_gifs\_devmode'
	if (!(Test-Path -LiteralPath "$DevmodePath")) {
		New-Item -ItemType Directory -Force -Path $DevmodePath | Out-Null
	}
	# ... and add/refresh a README to the folder
	if ($DevMode) {
		$readme_path = "$DevmodePath\_README.txt"
		if (!(Test-Path -LiteralPath "$readme_path")) {
			New-Item -Path "$readme_path" | Out-Null
			$msg = "Script 'DevMode' Folder" +
				"`n----------------------------" +
				"`n`nWHAT? This folder holds output created when using the -DevMode argument" + 
				"`n`nNote: This folder and this README.txt file are programmatically created."
			$msg | Out-File -FilePath "$readme_path" -Append
		}
	}
}
						
# Get current version of Blue Iris
# ... Note: this function call relatively time consuming (~100 msec), so we avoid using in a loop
$BiVersion = GetBiVersion
# Write-Host "BiVersion = $BiVersion"

# GET SHORT- & LONG-NAMES, AND PROFILES for ALL BLUE IRIS CAMERAS
# ... via Blue Iris JSON interface
if (1 -eq 1) {  # Use to collapse this code section while editing

	$cliplist_hashtable = `
		BIJsonCamListParse `
		-CamProps 'optionDisplay,isEnabled,profile' `
		-QuietFlg $Quiet `
		-FromConsoleFlg $fromConsole

	# $cliplist_hashtable is like...
	# {
	# 	"DW1": {
	# 		"optionDisplay": "DriveWay-1",
	# 		"isEnabled": true,
	#       "profile": 5
	# 	},
	# 	"DW2": {
	# 		"optionDisplay": "DriveWay-2",
	# 		"isEnabled": true
	#       "profile": 5
	# 	},
	# 	---
	# }
			
	# For debugging
	#Write-Host ("hashtable keys = " + ([String[]]$cliplist_hashtable.Keys | ConvertTo-Json))
	#Write-Host ("hashtable = " + ($cliplist_hashtable | ConvertTo-Json))

	$cam_shortnames_ary = [String[]]$cliplist_hashtable.Keys
	$cam_lognnames_ht = @{}
	$cam_profiles_ht = @{}
	foreach ($item in $cam_shortnames_ary) {
		if ($cliplist_hashtable[$item]['isEnabled']) {
			$cam_lognnames_ht.Add($item, $cliplist_hashtable[$item]['optionDisplay'])
			$cam_profiles_ht.Add($item, $cliplist_hashtable[$item]['profile'])
		}
	}
			
	# For debugging
	# Write-Host ("cam_shortnames_ary = " + ($cam_shortnames_ary | ConvertTo-Json))
	# Write-Host ("cam_lognnames_ht = " + ($cam_lognnames_ht | ConvertTo-Json))
	# Write-Host ("cam_profiles_ht = " + ($cam_profiles_ht | ConvertTo-Json))
}

# GET CAMERA's 'Video recording method' and 'Pre-trigger record time'
# ... from Blue Iris registry
if (1 -eq 1) {  # Use to collapse this code section while editing

	$cam_longname = $cam_lognnames_ht[$Camera]
	$cam_profile = $cam_profiles_ht[$Camera]

	$cam_rec_is_continuous = $false  # Default
	$response = `
		IsCamRecContinuous `
		-CamLongName $cam_longname `
		-CamProfileNo $cam_profile			

	if ($response.GetType() -match 'bool') { 
		$cam_rec_is_continuous = $response
	}

	# For debugging
	# Write-Host "Camera = '$Camera'"
	# Write-Host "cam_longname = '$cam_longname'"
	# Write-Host ("response type = " + $response.GetType())
	# Write-Host "response = $response"
	# Write-Host "cam_rec_is_continuous = $cam_rec_is_continuous"

	$cam_pretrig_rectime_msec = 0  # Default
	$response = `
		GetPreTrigRecMsec `
		-CamLongName $cam_longname `
		-CamProfileNo $cam_profile			

	if ($response.GetType() -match 'int') { 
		$cam_pretrig_rectime_msec = $response
	}

	# For debugging
	# Write-Host "Camera = '$Camera'"
	# Write-Host "cam_longname = '$cam_longname'"
	# Write-Host ("response type = " + $response.GetType())
	# Write-Host "response = $response"
	# Write-Host "cam_pretrig_rectime_msec = $cam_pretrig_rectime_msec"
}

# TRY TO MAKE GIF
if ($AlertRec -or $DbRecNo -or $UI3Rec -or $Live -or $Latest) {

	# Start resize performance timer
	$timer_makegif = [Diagnostics.Stopwatch]::StartNew()
	
	# Handle Blue Iris installations using multiple 'New' folder paths
	if ($bi_new_folder_paths_obj.Count -gt 0) {
		$arr = $bi_new_folder_paths_obj.Keys
		if ($FromConsole -and !$Quiet) {
			Write-Host "Assuming Blue Iris is using multiple 'New' folders:"
			Write-Host ($arr | ConvertTo-Json)
		}
		# Fix REQUIRED param $bi_new_folder_path if unset by user
		if ($bi_new_folder_path.StartsWith('YOUR')) {
			$bi_new_folder_path = $arr[0]
		}
	}

	# For debugging
	# Write-Host "AlertRec = $AlertRec"
	# Write-Host "RootPath = $app_path"
	# Write-Host "New = $app_path"
	# Write-Host "UseDatFile = $gif_via_DAT_flg"
	# msg.exe $server_username /TIME:10 "UseDatFile = $gif_via_DAT_flg"		
	# Write-Host "DatReqdImages = $gif_DAT_reqd_images_cnt"
	# Write-Host "ImagesCnt = $gif_images_cnt"
	# Write-Host "TrimEndCust = $gif_trim_end_cust_obj"
	# Write-Host "Height = $gif_height"
	# Write-Host "FPS = $gif_fps"
	# Write-Host "QuietFlg = $quiet_flg"

	$makegif_obj = `
		MakeGIF `
		-RootPath $app_path `
		-BvrPaths $bi_new_folder_paths `
		-AlertPath $bi_alerts_folder_path `
		-AlertRec $clip_db_locator `
		-LiveFlg $Live `
		-Camera $Camera `
		-Seconds $duration_sec `
		-UseDatFileFlg $gif_via_DAT_flg `
		-LatestFlg $Latest `
		-DatReqdImages $gif_DAT_reqd_images_cnt `
		-DatImageIntervalMsec $dat_image_interval_msec `
		-BvrDelayMsec $bvr_delay_msec `
		-NoPushFlg $nopush_flg `
		-NoArchiveGifFlg $noarchivegif_flg `
		-ImagesCnt $gif_images_cnt `
		-ImagesCntUsedFlg $PSBoundParameters.ContainsKey('ImagesCnt') `
		-Height $gif_height `
		-FPS $gif_fps `
		-TrimBegPct $gif_trim_beg `
		-TrimEndPct $gif_trim_end `
		-TrimEndCustPct $gif_trim_end_cust `
		-ClipDurAdjPct $clip_dur_adj
	
	# MakeGIF() RETURNS: 
	# {
	#	"method": "via Clip",
	#	"file": "D:\\pushover_gifs\\GT\\output.gif",
	#	"delay": 52612,
	#	"cnt": 10,
	#	"datjpgs": 10,
	#	"height": 320,
	#	"fps": 2,
	#	"kb": 226.2,
	#	"msec": 2625,
	#	"trimbeg": 0,
	#	"trimend": 0,
	#	"trimcust": 30,
	#	"almsec": 52612,
	#	"res": "1920x1080",
	#	"cam": "GT",
	#	"memo": "person:86%",
	#	"source": "GT.20230609_160000.bvr"
	#	"archive": "D:\pushover_gifs\_archive\GT.20230609_160000.971414.17-0.gif"
	#   "dblocator": "@123456789"
	# }
	# or "ERR/message"
	
	# Trap error conditions
	if ($makegif_obj.GetType() -match 'string') {
		#if ($makegif_obj.StartsWith("ERR/")) {
		#if ($makegif_obj.StartsWith("ERR/") -or $makegif_obj.StartsWith("WARNING:")) {
		if ($makegif_obj.StartsWith("ERR/") -or $makegif_obj.StartsWith("FAIL")) {
			if ($FromConsole) {
				Write-Host ("ERR/unexpected response from MakeGIF(): '" + $makegif_obj + "'") -ForegroundColor Red
				Write-Host ""
				exit
			} else {
				return $makegif_obj
			}
		}
	}

	# For debugging
	#Write-Host ("MakeGIF() response type = " + $makegif_obj.GetType())
	#Write-Host ("MakeGIF() response = " + ($makegif_obj | ConvertTo-Json))
	
	# Parse output for items to be displayed now, and logged later
	$gif_method        = $makegif_obj.method
	$gif_filespec      = $makegif_obj.file
	$delay_ms          = $makegif_obj.delay     # type is [Int]
	$gif_imgs_cnt      = $makegif_obj.cnt       # type is [Int]
	$dat_jpgs_cnt      = $makegif_obj.datjpgs   # type is [Int]
	$gif_height        = $makegif_obj.height    # type is [Int]
	$gif_fps           = $makegif_obj.fps       # type is [Int]
	$gif_ms            = $makegif_obj.msec      # type is [Int]
	$gif_trim_beg      = $makegif_obj.trimbeg   # type is [Int]
	$gif_trim_end      = $makegif_obj.trimend   # type is [Int]
	$gif_trim_end_cust = $makegif_obj.trimcust  # type is [Int]
	$gif_kb            = $makegif_obj.kb        # type is [double]
	$gif_jpg_wxh       = $makegif_obj.res
	$al_cam            = $makegif_obj.cam
	$al_memo           = $makegif_obj.memo
	$gif_source_fs     = $makegif_obj.source
	$archive_fs        = $makegif_obj.archive
	$db_locator        = $makegif_obj.dblocator

	$gif_kb2 = [Math]::Round($gif_kb,0).ToString("0")
	$gif_sec = [Math]::Round($gif_ms/1000,1).ToString("#.0")
	$gif_source_fn = (Split-Path $gif_source_fs -leaf).TrimStart('\')
	
	# For debugging
	#Write-Host "gif_method = $gif_method"
	#msg.exe $server_username /TIME:10 "gif_method = $gif_method"
	#Write-Host "delay_ms = $delay_ms"
	#Write-Host "gif_imgs_cnt = $gif_imgs_cnt"
	#Write-Host "dat_jpgs_cnt = $dat_jpgs_cnt"
	#Write-Host "gif_height = $gif_height"
	#Write-Host "gif_fps = $gif_fps"
	#Write-Host "gif_ms = $gif_ms"
	#Write-Host "gif_trim_beg = $gif_trim_beg"
	#Write-Host "gif_trim_end = $gif_trim_end"
	#Write-Host "gif_trim_end_cust = $gif_trim_end_cust"
	#Write-Host "gif_kb = $gif_kb"
	#Write-Host "gif_jpg_wxh = $gif_jpg_wxh"
	#Write-Host "al_cam = $al_cam"
	#Write-Host "al_memo = $al_memo"
	#Write-Host "gif_source_fs = '$gif_source_fs'"
	#Write-Host "gif archive_fs = '$archive_fs'"
	#Write-Host "gif_source_fn = '$gif_source_fn'"

	if ($FromConsole -and !$Quiet) {
		Write-Host "`nGIF File Summary:"
		Write-Host "-----------------"
		Write-Host "GIF created: '$gif_filespec'"
		Write-Host "GIF method:  $gif_method"
		Write-Host "GIF specs:   ImageCnt,Height,Dur(msec),FPS,Size(KB) = $gif_imgs_cnt,$gif_height,$gif_ms,$gif_fps,$gif_kb"
		#Write-Host "GIF archived as: '$archive_fs'"
		#Write-Host "GIF source:  '$gif_source_fs'"
		
	}
	if ($gif_kb -gt 2500) {
		if ($FromConsole) {
			Write-Warning "GIF file may exceed limit (2500 KB) for Pushover API"
		}
	}
	if (!$noarchivegif_flg) {
		if ($archive_fs -match 'fail') {
			if ($FromConsole) {
				Write-Warning "unable to archive the GIF file (reason unknown)"
			}
		} elseif ($archive_fs.StartsWith("ERR/")) {
			if ($FromConsole) {
				Write-Warning "unable to archive the GIF file ($archive_fs)"
			}
		} else {
			if ($FromConsole -and !$quiet_flg) {
				Write-Host "GIF archived as: '$archive_fs'"
			}
		}
	}

	if ($FromConsole -and !$quiet_flg) {
		Write-Host "GIF source:  '$gif_source_fs'"
	}

	# Stop resize performance timer
	$timer_makegif.stop()
	#$makegif_ms = [Math]::Round(($timer_makegif.elapsed.totalmilliseconds - $delay_ms),1).ToString("#.0")
	$makegif_ms = [Math]::Round(($timer_makegif.elapsed.totalmilliseconds),1).ToString("#.0")
	$makegif_ms2 = [Math]::Round(($timer_makegif.elapsed.totalmilliseconds),0).ToString("0")
	
	# Fill return object
	#$rtn_obj = @{}
	$rtn_data_obj = [ordered]@{}
	$rtn_obj['result'] = 'success'
	$rtn_data_obj['gif_method'] = $gif_method
	$rtn_data_obj['gif_make_ms'] = $makegif_ms
	$rtn_data_obj['gif_images'] = $gif_imgs_cnt
	$rtn_data_obj['dat_images'] = $dat_jpgs_cnt
	$rtn_data_obj['gif_height'] = $gif_height
	$rtn_data_obj['gif_fps'] = $gif_fps
	$rtn_data_obj['gif_ms'] = $gif_ms
	$rtn_data_obj['gif_trim_pct_beg'] = $gif_trim_beg
	$rtn_data_obj['gif_trim_pct_end'] = $gif_trim_end
	$rtn_data_obj['gif_trim_pct_cust'] = $gif_trim_end_cust
	$rtn_data_obj['gif_kb'] = $gif_kb
	$rtn_data_obj['gif_jpg_wxh'] = $gif_jpg_wxh
	$rtn_data_obj['gif_fs'] = $archive_fs
	#$rtn_data_obj['gif_source_fs'] = $bi_new_folder_path + "\" + $gif_source_fs.Name
	$rtn_data_obj['gif_source_fs'] = $gif_source_fs
	$rtn_obj['data'] = $rtn_data_obj
}

# Start notification service performance timer
$timer_notification = [Diagnostics.Stopwatch]::StartNew()

# SEND NOTIFICATION
if (!$demo_flg -and !$nopush_flg) {
	
	if ($push_service -eq "pushover") {
		
		if ($Title) {
			$po_title = $Title
		} else {
			$po_title = "via $ScriptName"
		}
		
		if ($Sound) {
			$po_sound = $Sound
		}
		
		if ($gif_kb -gt 0) {
			
			$gif_image_path = $app_path + "\" + $al_cam + "\output.gif"
			
			if (!$Msg) {
				# Default notification message
				$notification_msg = "GIF: imgs=$gif_imgs_cnt, H=$gif_height, sec=$gif_sec, FPS=$gif_fps, KB=$gif_kb2"
				$notification_msg +="`nTrimBeg/End/EndCust: $gif_trim_beg/$gif_trim_end/$gif_trim_end_cust"
				$notification_msg +="`nSourceJpgSize: $gif_jpg_wxh"
				if ($al_memo) {
					# If argument -AlertMemo was used, check if it different from the alertlist memo
					# If so, wrap the -AlertMemo value in slashes as a 'heads-up' in the notification message
					# (an example of a situation where this can occur is in testing)
					if ($AlertMemo -and $al_memo) {
						if ($al_memo -ne $AlertMemo) {
							$al_memo = "/" + $AlertMemo + "/"
						}
					}
					$notification_msg +="`nAlertMemo: " + $al_memo
				} else {
					$notification_msg +="`nAlertMemo: (none)"
				}
			} else {
				$notification_msg = $Msg		
			}

			# Append optional 'DAT tag' to notification message string
			if ($append_gif_method_tag_flg) {
				if ($gif_method -match 'viaDAT') {
					$method_str = "viaDAT:" + $dat_jpgs_cnt
					#$method_str = $notification_msg + " viaDAT:" + $gif_imgs_cnt + "/" + $dat_jpgs_cnt
				} else {
					$method_str = "$gif_method"
				}
				$notification_msg += "`nMethod: ($method_str)"
			}
			
			# Append optional script performance info to notification message string
			if ($append_performance_tag_flg) {
				$notification_msg += "`nGifMakeMsec: " + $makegif_ms2
			}
			
			# For debugging
			#Write-Host "notification_msg AFTER = '$notification_msg'"
			#msg.exe $server_username /TIME:10 "gif_method = '$gif_method'"		

			# Returns string - either API response, or "ERR/*"|"WARNING:*" message
			# ... returns 'ERR/*' or 
			#     API response is like '@{status=1; request=28f81c23-dc07-4a36-8523-f6428159284c}'
			$reponse_str = `
				SendPushover `
				-MsgStr $notification_msg `
				-MsgTitle $po_title `
				-PriorityLevel $po_priority `
				-Sound $po_sound `
				-AttachPath $gif_image_path `
				-Camera $al_cam `
				-QuietFlg ($bg_msgs_on -eq $false)

			if ($FromConsole -and !$Quiet) {
				Write-Host "`nPushover Summary:"
				Write-Host "-----------------"
			}

			# For testing - Simulate a bad pushover response
			#$reponse_str = "ERR/test"
			#$reponse_str = "WARNING:test"
			
			# for debugging
			#Write-Host "SendPushover() response_str = $reponse_str"
			
			# Trap error conditions
			if ($reponse_str.StartsWith("ERR/") -or $reponse_str.StartsWith("WARNING")) {

				# POST MESSAGE TO BI LOGFILE
				if ($post_to_BI_logfile_flg) {
					$null = PostCustomBiLogEntry `
						-MsgStr "$ScriptName - unexpected response trying to send Pushover notification: $reponse_str" `
						-Level 2 `
						-QuietFlg $quiet_flg
				}

				if ($FromConsole) {
					Write-Host ("ERR/unexpected response from SendPushover(): '" + $reponse_str + "'") -ForegroundColor Red
					Write-Warning "Notification not sent"
				}		
				
			} else {
					
				if ($FromConsole -and !$Quiet) {
					Write-Host $reponse_str
				}
				
				<# if ($bg_msgs_on) {
					$bg_msg = "$reponse_str ($script:ScriptName)"
					PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on
				} #>
			}
		}
	}
	
	if ($push_service -eq "telegram") {
		
		if ($gif_kb -gt 0) {
			
			$gif_image_path = $app_path + "\" + $al_cam + "\output.gif"
			
			if (!$Msg) {
				# Default notification message
				$notification_msg = "GIF: imgs=$gif_imgs_cnt, H=$gif_height, sec=$gif_sec, FPS=$gif_fps, KB=$gif_kb2"
				$notification_msg +="`nTrimBeg/End/EndCust: $gif_trim_beg/$gif_trim_end/$gif_trim_end_cust"
				$notification_msg +="`nSourceJpgSize: $gif_jpg_wxh"
				if ($al_memo) {
					# If argument -AlertMemo was used, check if it different from the alertlist memo
					# If so, wrap the -AlertMemo value in slashes as a 'heads-up' in the notification message
					# (an example of a situation where this can occur is in testing)
					if ($AlertMemo -and $al_memo) {
						if ($al_memo -ne $AlertMemo) {
							$al_memo = "/" + $AlertMemo + "/"
						}
					}
					$notification_msg = $notification_msg + "`nAlertMemo: " + $al_memo
				} else {
					$notification_msg = $notification_msg + "`nAlertMemo: (none)"
				}
			} else {
				$notification_msg = $Msg		
			}

			# Append optional 'DAT tag' to message string
			if ($append_gif_method_tag_flg) {
				if ($gif_method -match 'viaDAT') {
					$method_str = "viaDAT:" + $dat_jpgs_cnt
					#$method_str = $notification_msg + " viaDAT:" + $gif_imgs_cnt + "/" + $dat_jpgs_cnt
				} else {
					$method_str = "$gif_method"
				}
				$notification_msg += "`nMethod: ($method_str)"
			}
			
			# Append optional script performance info to Telegram message string
			if ($append_performance_tag_flg) {
				$notification_msg += "`nGifMakeMsec: " + $makegif_ms2
			}
			
			# For debugging
			#Write-Host "notification_msg AFTER = '$notification_msg'"
			#notification_msg.exe $server_username /TIME:10 "gif_method = '$gif_method'"		

			# Returns string - either API response, or "ERR/*"|"WARNING:*" message
			# ... returns 'ERR/*' or 
			#     API response is like '@{status=1; request=28f81c23-dc07-4a36-8523-f6428159284c}'
			$reponse_str = `
				SendTelegram `
				-MsgStr $notification_msg `
				-AttachPath $gif_image_path `
				-Camera $al_cam `
				-QuietFlg ($bg_msgs_on -eq $false)

			if ($FromConsole -and !$Quiet) {
				Write-Host "`nTelegram Summary:"
				Write-Host "-----------------"
			}

			# For testing - Simulate a bad API response
			#$reponse_str = "ERR/test"
			#$reponse_str = "WARNING:test"
			
			# for debugging
			# Write-Host "SendTelegram() response_str = $reponse_str"
			
			# Trap error conditions
			if ($reponse_str.StartsWith("ERR/") -or $reponse_str.StartsWith("WARNING")) {

				# POST MESSAGE TO BI LOGFILE
				if ($post_to_BI_logfile_flg) {
					$null = PostCustomBiLogEntry `
						-MsgStr "$ScriptName - unexpected response trying to send Telegram notification: $reponse_str" `
						-Level 2 `
						-QuietFlg $quiet_flg
				}

				if ($FromConsole) {
					Write-Host ("ERR/unexpected response from SendTelegram(): '" + $reponse_str + "'") -ForegroundColor Red
					Write-Warning "Notification not sent"
				}		
				
			} else {
					
				if ($FromConsole -and !$Quiet) {
					Write-Host $reponse_str
				}
				
				<# if ($bg_msgs_on) {
					$bg_msg = "$reponse_str ($script:ScriptName)"
					PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on
				} #>
			}
		}
	}

}

# Stop notification service performance timer
$timer_notification.stop()
if ($push_service -eq "telegram") {
	$telegram_ms = [Math]::Round($timer_notification.elapsed.totalmilliseconds,1).ToString("0.0")
	$pusho_ms = "0.0"
} else {
	$telegram_ms = "0.0"
	$pusho_ms = [Math]::Round($timer_notification.elapsed.totalmilliseconds,1).ToString("0.0")
}

# Stop overall performance timer
$timer_total.stop()
$total_ms = [Math]::Round($timer_total.elapsed.totalmilliseconds,1) + $BvrDelayMsec
$total_ms_disp = $total_ms.ToString("0.0")

# LOG SCRIPT PERFORMANCE INFO TO CSV FILE
# ... Prep values for logging
if (1 -eq 1) {  # Use to collapse this code section while editing

	$csvfilespec = $app_path + "\" + $ScriptBaseName + "_" + $ScriptVersion + "_log.csv"
	$timestamp   = Get-Date -Format "yyyy/MM/dd_HH:mm:ss"
	$total_ms    = [decimal]$total_ms 
	$pusho_ms    = [decimal]$pusho_ms 
	#$gif_ms      = [decimal]$gif_ms 
	#$gif_sec     = [decimal]$gif_sec

	if ($gif_kb -gt 0) {
		$images_cnt    = [Int]$gif_imgs_cnt
		$gif_height    = [Int]$gif_height
		$gif_fps       = [Int]$gif_fps
		$gif_trim_beg  = [Int]$gif_trim_beg
		$gif_trim_end  = [Int]$gif_trim_end
		$gif_trim_cust = [Int]$gif_trim_end_cust
		$gif_sec       = [decimal]$gif_sec
		$gif_size      = [Int]$gif_kb
		$makegif_ms    = [decimal]$makegif_ms
	} else {
		$images_cnt    = 0
		$gif_height    = 0
		$gif_fps       = 0
		$gif_trim_beg  = 0
		$gif_trim_end  = 0
		$gif_trim_cust = 0
		$gif_sec       = 0
		$gif_size      = 0
		$makegif_ms    = [decimal]$makegif_ms 
	}

	$disp_total_ms    = [Math]::Round($total_ms,0)
	$disp_delay_ms    = [Math]::Round($delay_ms,0)
	$disp_makegif_ms  = [Math]::Round($makegif_ms,0)
	$disp_pusho_ms    = [Math]::Round($pusho_ms,0)
	$disp_telegram_ms = [Math]::Round($telegram_ms,0)

	# SHOW PERFORMANCE NOTICE
	# NOTE (2024-02-15): The 'JPG fetch delay' has been rendered irrelevant as a result of
	# changing the source of JPGs for -Live GIFs from the active BVR file to the live stream.
	# 'JPG fetch delay' is now ALWAYS ZERO. Despite this, we are continuing to report it in 
	# Blue Iris logfile entries. WHY? to not break other utilities that extract/use this script's
	# performance data from the BI logfile).
	if ($FromConsole -and !$quiet_flg) {
		if ($push_service -eq "telegram") {
			$perf_msg = (`
				"Performance Summary:" `
				+ "`n--------------------" `
				+ "`nJPG fetch delay:         $disp_delay_ms msec" `
				+ "`nmakegif execution time:  $disp_makegif_ms msec" `
				+ "`ntelegram execution time: $disp_telegram_ms msec" `
				+ "`ntotal execution time:    $disp_total_ms msec"`
				)
		}
		if ($push_service -eq "pushover") {
			$perf_msg = (`
				"Performance Summary:" `
				+ "`n--------------------" `
				+ "`nJPG fetch delay:         $disp_delay_ms msec" `
				+ "`nmakegif execution time:  $disp_makegif_ms msec" `
				+ "`npushover execution time: $disp_pusho_ms msec" `
				+ "`ntotal execution time:    $disp_total_ms msec"`
				)
		}
		Write-Host "`n$perf_msg"
	}

	# ... Log the values
	#     RETURNS:  (array) @(images_cnt, gif_height, gif_fps, gif_ms, gif_size, gif_filespec)
	$log_response = `
		LogScriptResultsToCsv `
		$csvfilespec `
		$timestamp `
		$gif_source_fn `
		$gif_method `
		$images_cnt `
		$gif_height `
		$gif_fps `
		$gif_trim_beg `
		$gif_trim_end `
		$gif_trim_end_cust `
		$gif_sec `
		$gif_size `
		$gif_jpg_wxh `
		$makegif_ms `
		$pusho_ms `
		$telegram_ms `
		$total_ms `
		$db_locator

	#Write-Host ("LogScriptResultsToCsv response = " + ($log_response | ConvertTo-Json))

	if ($FromConsole -and !$Quiet) {
		Write-Host "`nCSV Log Summary:"
		Write-Host "-----------------"
	}

	if ($log_response.StartsWith("ERR/")) {
		Write-Host $log_response
		if (!$FromConsole) {
			$bg_msg = $log_response
			PopupBackgroundMsg -Seconds 10 -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
		}
	} else {
		if ($FromConsole -and !$Quiet) {
			Write-Host "Performance data logged to '$csvfilespec'"
		}
	}
}

#Write-Host "bg_msgs_on = $bg_msgs_on"

# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
if ($bg_msgs_on -and !$FromConsole) {
	if ($alert_rec) {
		$alert_rec_str = $alert_rec
	} elseif ($AlertRec) {
		$alert_rec_str = $AlertRec
	} elseif ($DbRecNo) {
		$alert_rec_str = $DbRecNo
	} elseif ($UI3Rec) {
		$alert_rec_str = $UI3Rec
	} elseif ($Camera -and $Latest) {
		$alert_rec_str = 'LATEST ALERT'
	} else {
		$alert_rec_str = '<unspecified>'
	}
	$msg_sec = 30
	$bg_msg = "Alert record: $alert_rec_str" + 
		"`n`nSource file: $gif_source_fn" +
		"`n`nGIF Height/FPS/Dur(s)/Size(KB): $gif_height / $gif_fps / $gif_sec / $gif_kb" + 
		"`n`nTimes (delay,makegif,pusho,total) ... $disp_delay_ms, $disp_makegif_ms, $disp_pusho_ms, $disp_total_ms msec" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"
		
	# KEY POINT: message length limit is 255 characters
	# For debugging.
	#Write-Host ("bg_msg len = " + $bg_msg.length)
	#Write-Host ("bg_msg:`n" + $bg_msg)
	
	PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
}

# POST MESSAGE TO BLUE IRIS LOGFILE
# KEY POINT: the message string maximum length is 127 characters (BI HTTP interface command '/admin?log=message&level=x')
if (1 -eq 1) {  # Use to collapse this code section while editing

	if ($FromConsole -and !$Quiet) {
		Write-Host "`nBlue Iris Logfile:"
		Write-Host "------------------"
	}
	$bilogfile_msg = "No custom BI logfile message posted"
	if (!$nobilog_flg) {
		if ($post_all_events) {
			if (!$demo_flg) {
				if ($rtn_obj -match "ERR/") {
					$log_level = 2
					$log_msg = "$scriptName - $rtn_obj"
				} elseif ($rtn_obj -match "WARNING:") {
					$log_level = 1
					$log_msg = "$scriptName - $rtn_obj"
				} else {
					$log_level = 0
					if ($push_service -eq "telegram") {
						$disp_notify_ms = $disp_telegram_ms
					} else {
						$disp_notify_ms = $disp_pusho_ms
					}
					# $log_msg = "gif_tool [$gif_height H, $gif_fps FPS, $gif_sec s, $gif_kb KB] [$disp_delay_ms, $disp_makegif_ms, $disp_notify_ms, $disp_total_ms ms] $gif_method $push_service rec=$alert_rec $gif_source_fn"
					$log_msg = "gif_tool [$gif_height H, $gif_fps FPS, $gif_sec s, $gif_kb KB] [$disp_delay_ms, $disp_makegif_ms, $disp_notify_ms, $disp_total_ms ms] $gif_method $push_service $gif_source_fn"
				}
				$response_str = `
					PostCustomBiLogEntry `
					-MsgStr "$log_msg" `
					-Level $log_level `
					-QuietFlg $true
					
				if ($response_str) {
					$bilogfile_msg = "Custom BI logfile message posted"
				}
			}
		}
	}
	if ($FromConsole -and !$Quiet) {
		Write-Host $bilogfile_msg
	}
}

if ($FromConsole -and $Quiet) {
	#Write-Host "return object ="
}

if (!$FromConsole -or $Quiet) {
	if ($RtnHashtable) {
		return $rtn_obj
	} else {
		Write-Host "CLI return object`n-----------------"
		return ($rtn_obj | ConvertTo-Json)
	}
}

if ($FromConsole -and !$Quiet) {
	Write-Host ""
}

#Stop-Transcript